//************************************************************
//           Math unit input and output functions
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter6\Mu Input
// Entry point: Mu Input
// Other files: Un32_4.lib
//              Un32_4.h
// Development: Visual C++ 6.0 Console Application
// Date:        January 12/2001
//
// Description:
//              Header file for conversion functions in 
//              the Un32_4 module of math32.lib
//************************************************************
// Library functions
extern "C" FPU_INPUT();
extern "C" ASCII_TO_EXP();
extern "C" FPU_OUTPUT();
extern "C" GET_TAG();

//**************************************************************
//                       Prototypes
//**************************************************************
int FpuInput(char[]);
void AsciiToExp(char[], char[]);
int FpuOutput(char[]);
int GetTag(int, char[]);

//**************************************************************
//            C++ interface functions to Un32_4 module
//**************************************************************
int FpuInput(char number[])
{
// On entry:
//     parameter number[] is a user-defined array holding an
//     ASCII decimal number formatted in exponential or
//     decimal notation.
//
//     Exponential notation as follows:
//
//                 sm.mmmmmmmmmmmmmmmmm ESeeee
// where
// s = sign of number (blank = +)
// m = up to 18 significand digits. Extra digits are ignored
// . = decimal point following the first significand digit.
//     If no decimal point in string, it is assumed
// E = explicit letter E (or e) to signal start of exponent
// S = + or - sign of exponent. If no sign, positive is assumed
// e = up to four exponent digits. The numerical value of the
//     exponent cannot exceed +/- 4932
// One or more spaces can be used to separate the last digit of
// the significand and the start of the exponent
//
// Examples of input:
//                    1.781252345E-1
//                    -3.14163397  E+0
//                    1.2233445566778899 e1387
//                    11.2233
//                   -0.004455
// On exit:
//           User's value is loaded into math unit ST(0) register
//           Previous values in the stack are pushed down one
//           register
//           returned value = 0 (no error)
//
//        b. If returned value not zero, returned value indicates
//           the following errors
//           1    if no E or e symbol in the caller's string
//                NAN encoding is loaded into ST(0)
//           2    if exponent exceeds valid range
//                Overflow - positive infinity is loaded into ST(0)
//                Underflow - negative infinity is loaded into ST(0)
//           3    if there is an invalid character in caller's
//                string. NAN encoding is loaded into ST(0)
//
//        AX is destroyed. All other registers are preserved.
//        The caller's NDP environment is preserved, except
//        the stack top register.
//********************************************************************
// CPU compatibility chart:
//                       | 8086/8088/8087  | NO  |
//                       | 80286/80287     | NO  |
//                       | 80386/80387     | YES |
//                       | 486 DX          | YES |
//                       | PENTIUM         | YES|
//********************************************************************

int rtnError = 0;
_asm
{
	MOV     EDX,number
	CALL    FPU_INPUT
	; Capture error code
	MOV     rtnError,EAX
}
  return rtnError;  
}


void AsciiToExp(char ascStr[], char expStr[])
{
// Checks for the ASCII input string and converts it to
// exponential form
// On entry:
//        ascStr[]  is an ASCII decimal string, terminated in a NULL
//                  byte, which holds the caller's ASCII decimal
//                  input
//             Note: The caller's buffer may contain the $ sign
//                   and commas as formatting symbols. The string
//                   terminator (00H) may be preceded by a CR
//                   (0DH) or LF (0AH) control codes. If the
//                   characters E or e are in the string, the
//                   procedure assumes that the string is already
//                   in exponential notation
//
//        expStr[]  is a 27-byte unformatted output buffer which will
//                  be filled with the output string formatted in
//                  exponential form
//
// On exit:
//        expStr[]  is the caller's output buffer with the input number
//                  reformatted in exponential form, as follows:
//
//                 sm.mmmmmmmmmmmmmmmmm ESeeee
// where
// s = sign of number (blank = +)
// m = 18 significand digits
// . = decimal point following the first significand digit
// E = letter E signals the start of the exponent field
// S = + or - sign of exponent
// e = up to four exponent digits
//
// NOTE:
//      This function can be used to make sure that the input string
//      passed to the FpuInput() function is formatted in exponential
//      notation. 
//
//
_asm
{
	MOV     ESI,ascStr
    MOV     EDI,expStr 
	CALL    ASCII_TO_EXP
 
}
  return;  
}

int FpuOutput(char userBuf[])
{
//          Converts the value in the top register of the FPU
//          stack into an ASCII decimal number in scientific
//          notation
//
// On entry:
//         userBuf[] is the unformatted caller's buffer area
//         with minimum space for 25 bytes of storage
//
//         ST(0) holds value to be converted
//
// On exit:
//         Return value = 0
//         ST(0) holds value to converted
//         The caller's buffer will contain the ASCII decimal
//         representation of the number at the FPU stack top
//         register, formatted in scientific notation, as follows:
//
//                 sm.mmmmmmmmmmmmmmmmm ESeeee
// where
// s = sign of number (blank = +)
// m = 18 significand digits
// . = decimal point following the first significand digit
// E = explicit letter E to signal start of exponent
// S = + or - sign of exponent
// e = up to four exponent digits
//
// Examples of output:
//                    1.78125234500000000 E-12
//                   -3.14163397000000000 E+0
//                    1.22334455667788998 E+1388
//
//       Return value = 1
//	     Invalid value in top of stack
//
// Other number types return the following 12-character text srings
// in the caller's buffer
//                        + denormal
//                        - denormal
//                        + NAN
//                        - NAN
//                        + infinity
//                        - infinity
//                        unsupported
//                         EMPTY
//
//      Signed zeros are reported as follows:
//                    0.00000000000000000 E+0 (positive zero)
//                   -0.00000000000000000 E+0 (negative zero)
//
//       The caller's math unit environment is preserved

int rtnError = 0;	
_asm
{
	MOV     EDX,userBuf
	CALL    FPU_OUTPUT
    JNC     FPU_EXIT
    // Error
	MOV     rtnError,1
FPU_EXIT:
}
    return rtnError; 
}

int TagCode(int reg, char userBuf[])
{
// Function to obtain tag code corresponding to the requested FPU
// register
// A 7-character buffer with the tag text is returned in the buffer
// passed by the caller
// On entry:
//         userBuf[] is caller's storage area with a minimum of 7 
//                   free bytes for message storage
//         reg hols register stack desired (range 0 to 7)
//
// On exit:
//         userBuf[] contains tag string
//         returned value is tag code for math unit register
//
int value;
_asm
{
     MOV    EDI,userBuf
     MOV    EAX,reg
	 CALL   GET_TAG
     AND    EBX,0FFFFH
	 MOV    value,EBX
}
  return value;
}	 
	 
