; Module name: AUXSERV.ASM                   Library:   SOLUTION.LIB
; Created:
; Update:
;
;*******************************************************************
;*******************************************************************
;                            AUXSERV.ASM
;*******************************************************************
;*******************************************************************
;
; Library module for auxiliary routines
;
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  TIMER_ON
PUBLIC  TIMER_OFF
PUBLIC  SHOW_TIMER
PUBLIC  SYS_PARAMS
PUBLIC  EQUIPMENT
PUBLIC  SOUND
PUBLIC  MILLI_TIME
PUBLIC  RANDOM_NUM
;
;
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;******************************************************************
; IMPORTANT:
;          In order to avoid affecting the system, the timing
; routines do not reset the clock. This may cause a reading to be
; invalid due to counter overflow. For this reason, timing operations
; should be repeated at least twice, so as to discard the possibility
; of a timer overflow error.
;
;********************** program timing routines ******************
;
; This code section contains routines that can be used to time
; program execution in clock cycles. Timing is measured from the
; call to the procedure TIMER_ON to the call to the procedure
; TIMER_OFF. The auxiliary procedure SHOW_TIMER displays, at the
; cursor position, the number of clock cycles elapsed.
;
; Note: All internal data is stored in the code segment so as to
;       simplify interfacing with user software
;
START_COUNT     DW      0H      ; Counter value at start
END_COUNT       DW      0H      ; Counter at end
;
DEC_TIME        DB      0DH,0AH,'Clock Cycles:'
TIME_DIGITS     DB      '     $'
;
;
;******************************************************************
TIMER_ON        PROC    FAR
; Read system timer and store low word of timer count
; On entry:
;         nothing
; On exit:
;        CS:START_COUNT = low order word of system timer counter
;
        MOV     AH,0
        INT     1AH
        MOV     CS:START_COUNT,DX
        RET
TIMER_ON        ENDP
;******************************************************************
;
TIMER_OFF       PROC    FAR
; Read system timer and store low word of timer count
; On entry:
;         nothing
; On exit:
;        CS:END_COUNT = low order word of system timer counter
;
        MOV     AH,0            ; Read clock service request
        INT     1AH             ; To time-of-day service
; Save counter low word in program storage
        MOV     CS:END_COUNT,DX
        RET
TIMER_OFF       ENDP
;******************************************************************
;
SHOW_TIMER      PROC    FAR
;
; This procedure subtracts the timer counts stored by the
; TIMER_ON and TIMER_OFF procedures, converts the binary result
; to ASCII decimal and displays it at the cursor position
;
; On entry:
;         nothing
; On exit:
;         elapsed clock cycles between the previous TIMER-ON and
;         TIMER-OFF calls is displayed at the current cursor
;         position.
;
        MOV     DX,CS:END_COUNT         ; Count at end
        MOV     AX,CS:START_COUNT       ; At start
        SUB     DX,AX                   ; End minus start
        LEA     DI,CS:TIME_DIGITS       ; ASCII output buffer
; Use BIN_TO_ASC local procedure for conversion
        PUSH    DS              ; Save entry DS
        PUSH    CS
        POP     DS              ; Pass code segment as data
        CALL    BIN_TO_ASC
; Use DOS service number 9 to display time string
; Note: DS is still set to CS
        MOV     AH,9            ; DOS service request code
        LEA     DX,CS:DEC_TIME  ; Buffer
        INT     21H
        POP     DS              ; Restore DS
        RET
SHOW_TIMER      ENDP
;
;*************************************|
;  routine to obtain system hardware  |
;*************************************|
; Hardware elements identified:
;       1. Systems
;               a. PC line: PC, XT, AT, PC Convertible or PCjr
;               b. PS/2 line: Model 25/30, Model 50/60 or,
;                             Model 70/80
;
;       2. PC line video display card or system:
;               a. Monochrome card
;               b. Color graphics adapter
;               c. Enhanced graphics adapter
;               d. No display card
;               e. PCjr
;
;       3. PS/2 line video display options:
;               a. MCGA with analog monochrome display
;               b. MCGA with analog color display
;               c. VGA with analog monochrome display
;               d. VGA with analog color display
;
;       4. Active display page
;
;       5. Active video mode
;
;       6. Number of screen columns in alphanumeric modes
;
; Code segment temporary storage for system identification codes
SYS_CODE        DB      0
;
SYS_PARAMS      PROC    FAR
; Procedure to obtain system identification parameters
; On entry:
;         nothing
; On exit:
;        CL = Machine line and type, as follows:
;             PC line (low bit is clear):
;             0 = Personal Computer    2 = PC XT
;             4 = PCjr                 6 = PC AT
;             8 = PC convertible
;             PS/2 line (low bit is set):
;             1 = Model 25/30          3 = Model 50/60
;             5 = Model 70/80
;        CH = Video hardware
;             PC line (high bit clear):
;             0 = MDA                  1 = CGA
;             2 = EGA                  3 = invalid code
;             PS/2 line (high bit set):
;             128 = VGA monochrome     129 = VGA color
;             130 = MCGA monochrome    131 = MCGA color
;        AH = number of screen columns in alphanumeric modes
;        AL = current video mode
;        BH = active display page (0 based)
;
;********************|
;   clear storage    |
;********************|
        MOV     CS:SYS_CODE,0   ; Could have garbage
;*****************|
;  test for PS/2  |
;    systems      |
;*****************|
; Use service number 26 of INT 10H to determine if system is a
; PC or PS/2 line machine
        MOV     AH,26           ; BIOS service request number
        MOV     AL,0            ; Read display combinations
        INT     10H
; AL = 1AH if system is a machine of the PS/2 line
; AL not equal 1AH if system is a machine of the PC line
        CMP     AL,26           ; Is it a PS/2 system?
        JNE     MACHINE_TYPE    ; Hardware is not PS/2
                                ; bit 0 is clear on default
;******************|
;     PS/2 line    |
;******************|
; Test for a value in BL less than 7 to discard non PS/2 options
        CMP     BL,7
        JL      MACHINE_TYPE    ; Non-PS/2 display system
        OR      CS:SYS_CODE,00000001B   ; Set bit number 0
;*****************|
;  machine type   |
;*****************|
MACHINE_TYPE:
;        7  6  5  4  3  2  1  0  SYS_CODE bit map
;        |  |  |  |  |  |  |  |
;        |  |  |  |  |  |  |  |____ MACHINE LINE
;        |  |  |  |  |  |  |        0 = PC line   1 = PS/2 line
;        |  |  |  |  |__|__|___ MACHINE TYPE
;        |  |  |  |         PC line               PS/2 line
;        |  |  |  |         000 = PC              000 = Model 25/30
;        |  |  |  |         001 = PC XT           001 = Model 50/60
;        |  |  |  |         010 = PCjr            010 = Model 70/80
;        |  |  |  |         011 = PC AT
;        |  |  |  |         100 = PC Convertible
;        |  |  |__|____ VIDEO HARDWARE
;        |  |               PC line               PS/2 line
;        |  |               00 = MDA              00 = VGA monochrome
;        |  |               01 = CGA              01 = VGA color
;        |  |               10 = EGA              10 = MCGA monochrome
;        |  |               11 = invalid code     11 = MCGA color
;        |__|____ NOT USED
;
; Obtain Machine identification code from address F000:FFFEH
; PC systems:
;         FFH ...... Personal Computer
;         FEH ...... PC XT
;         FBH ...... PC XT with BIOS dated 01/10/86 or later
;         FDH ...... PCjr
;         FCH ...... PC AT or Model 50/60
;         F9H ...... PC Convertible
; PS/2 systems:
;         FAH ...... Model 25/30
;         F8H ...... Model 70/80
;
        MOV     AX,0F000H       ; Base
        MOV     ES,AX           ; To extra segment
        MOV     AL,ES:[0FFFEH]  ; Machine ID byte to AL
; Select bit mask for bits 1, 2, and 3 according to machine ID code
        CMP     AL,0FFH         ; Personal computer
        JNE     TEST_FE
        JMP     VIDEO_HW        ; xxxx000x code is default setting
                                ; for bits 1, 2, and 3
TEST_FE:
        CMP     AL,0FEH         ; PC XT
        JNE     TEST_FB
        OR      CS:SYS_CODE,00000010B
        JMP     VIDEO_HW        ; Machine code is xxxx001x
TEST_FB:
        CMP     AL,0FBH         ; PC XT with BIOS dated 01/10/86
        JNE     TEST_FD
        OR      CS:SYS_CODE,00000010B
        JMP     VIDEO_HW        ; Machine code is xxxx001x
TEST_FD:
        CMP     AL,0FDH         ; PCjr
        JNE     TEST_FC
        OR      CS:SYS_CODE,00000100B
        JMP     VIDEO_HW        ; Machine code is xxxx010x
TEST_FC:
        CMP     AL,0FCH         ; PC AT
        JNE     TEST_F9
        OR      CS:SYS_CODE,00000110B
        JMP     VIDEO_HW        ; Machine code is xxxx011x
TEST_F9:
        CMP     AL,0F9H         ; PC Convertible
        JNE     TEST_FA
        OR      CS:SYS_CODE,00001000B
        JMP     VIDEO_HW        ; Machine code is xxxx100x
TEST_FA:
        CMP     AL,0FAH         ; Model 25/30
        JNE     TEST_F8
        JMP     VIDEO_HW        ; Machine code is xxxx000x
TEST_F8:
        CMP     AL,0F8H         ; Model 70/80
        JNE     BAD_TYPE        ; Machine type cannot be determined
        OR      CS:SYS_CODE,00000100B
        JMP     VIDEO_HW        ; Machine code is xxxx010x
;********************|
; machine type error |
;********************|
BAD_TYPE:
        STC                     ; Carry flag to signal error
        RET
;********************|
;        PS/2        |
;  video hardware    |
;********************|
VIDEO_HW:
; Value returned by service number 26, INT 10H
;              BL = 6H if professional graphics system
;              BL = 7H if VGA monochrome
;              BL = 8H if VGA color
;              BL = 0BH if MCGA monochrome
;              BL = 0CH if MCGA color
;              BL = any other value if a non-standard PS/2 system
;
        CMP     BL,06H          ; Test for professional graphics
                                ; system
        JNE     TEST_VGA
        OR      CS:SYS_CODE,00110000B
        JMP     VIDEO_PAGE      ; Machine code is xx11xxxx
TEST_VGA:
        CMP     BL,07H          ; Test for VGA monochrome
        JNE     TEST_VGA_C
        JMP     VIDEO_PAGE      ; Machine code is xx00xxxx
                                ; default setting
TEST_VGA_C:
        CMP     BL,08H          ; Test for VGA color
        JNE     TEST_MCGA
        OR      CS:SYS_CODE,00010000B
        JMP     VIDEO_PAGE      ; Machine code is xx01xxxx
TEST_MCGA:
        CMP     BL,0BH          ; Test for MCGA monochrome
        JNE     TEST_MCGA_C
        OR      CS:SYS_CODE,00100000B
        JMP     VIDEO_PAGE      ; Machine code is xx10xxxx
TEST_MCGA_C:
        CMP     BL,0CH          ; Test for MCGA color
        JNE     PC_VIDEO
        OR      CS:SYS_CODE,00110000B
        JMP     VIDEO_PAGE      ; Machine code is xx11xxxx
;*********************|
;  PC video hardware  |
;*********************|
PC_VIDEO:
; Obtain IBM equipment list word from address 0040:0010H
; Codes:
; Bit:  7  6  5  4  3  2  1  0
;       x  x  1  1  x  x  x  x___ Monochrome card
;       x  x  1  0  x  x  x  x___ CGA (80 x 25 mode)
;       x  x  0  1  x  x  x  x___ CGA (40 x 25 mode)
;
; Note: The data output by service request number 15 of BIOS
;       INT 10 does not differentiate between the CGA and the
;       EGA. Neither does the equipment list word. The presence
;       of a EGA card must be determined using BIOS service
;       number 18 of INT 10H
;
; Set ES to base address of ROM BIOS data area
        MOV     AX,0040H        ; Base
        MOV     ES,AX
        MOV     AX,ES:[0010H]   ; Equipment word to AX
        AND     AX,00110000B    ; Mask off all other bits
;********************|
;         MDA        |
;********************|
        CMP     AL,00110000B    ; Test for MDA card
        JNE     TEST_CGA
        JMP     VIDEO_PAGE      ; Machine code is xx00xxxx
                                ; default setting
;********************|
;        CGA         |
;********************|
TEST_CGA:
        CMP     AL,00100000B    ; First CGA code
        JNE     TEST_CGA2
IS_CGA:
        OR      CS:SYS_CODE,00010000B
        JMP     VIDEO_PAGE      ; Machine code is xx01xxxx
TEST_CGA2:
        CMP     AL,00010000B    ; Second CGA code
        JE      IS_CGA
;********************|
;         EGA        |
;********************|
        MOV     AH,18           ; BIOS service request number
        MOV     BL,16           ; Return EGA information
        INT     10H
; Test for no EGA in system
        CMP     BL,16
        JE      NOT_EGA
; Must be the Enhanced Graphics Adapter
        OR      CS:SYS_CODE,00100000B
        JMP     VIDEO_PAGE      ; Machine code is xx10xxxx
; No video card identified in system
NOT_EGA:
        OR      CS:SYS_CODE,00110000B
                                ; Machine code is xx11xxxx
;********************|
;     video mode     |
;  page and columns  |
;********************|
VIDEO_PAGE:
; Use service request number 15 of BIOS INT 10H to obtain
; current display page, mode, and screen columns, as follows:
;  AH = number of screen columns in alphanumeric modes
;  AL = current video mode
;  BH = active display page (0 based)
        MOV     AH,15           ; Request code number
        INT     10H
;********************|
; set return codes   |
;********************|
        MOV     CL,CS:SYS_CODE  ; Bit map to CL
        AND     CL,00001111B    ; Clear 4 high bits
        MOV     CH,CS:SYS_CODE  ; Bit map to CH
        SHR     CH,1            ; Shift right four times
        SHR     CH,1
        SHR     CH,1
        SHR     CH,1
; Test low bit in CL for PC or PS/2 lines
        TEST    CL,00000001B    ; Set for PS/2
        JZ      EXIT_SYS_P      ; Not PS/2. End processing
; System is a PS/2 machine, set high bit in CH
        OR      CH,10000000B
EXIT_SYS_P:
        CLC                     ; No error detection
        RET
SYS_PARAMS      ENDP
;*******************************************************************
;
EQUIPMENT       PROC    FAR
; Determine optional and standard devices present in system
; This procedure uses the BIOS service provided by INT 11H
; and INT 12H
; On entry:
;         Nothing
; On exit:
;        AH = number of printer ports
;        AL = number of serial ports
;        BH = number of diskette drives
;        BL = 1 if math coprocessor installed
;        CH = 1 if pointing device (mouse) intalled
;        DX = number of 64K memory blocks
;
        INT     11H     ; BIOS service request
; INT 11H returns AX configured as follows:
                        ; 7 6 5 4 3 2 1 0  AH bit map
                        ; | | | | | | | |__ RESERVED
                        ; | | | | |_|_|____ serial ports
                        ; | | | |__________ RESERVED
                        ; | | |____________ internal modem
                        ; |_|______________ printers
                        ; 7 6 5 4 3 2 1 0  AL bit map
                        ; | | | | | | | |__ IPL diskette
                        ; | | | | | | |____ math coprocessor
                        ; | | | | | |______ PS/2 pointing device
                        ; | | | | |________ RESERVED
                        ; | | |_|__________ Video hardware
                        ; | |               01 = 40 x 25 color
                        ; | |               10 = 80 x 25 color
                        ; | |               11 = 80 x 25 monochrome
                        ; |_|______________ diskette drives
                        ;                   00 = 1 drive
                        ;                   01 = 2 drives
;**********************|
;   test for printers  |
;**********************|
        MOV     DX,AX           ; Bit pattern to DX
        PUSH    DX              ; Save bit map
        AND     DH,11000000B    ; Mask off unwanted bits
        MOV     CL,6            ; Positions to shift right
        SHR     DH,CL           ; Values is now in binary
        MOV     AH,DH           ; And move to return register
        POP     DX              ; Restore bit map
;**********************|
; test for serial ports|
;**********************|
        AND     DH,00001110B    ; Mask off unwanted bits
        SHR     DH,1            ; Values is now in binary
        MOV     AL,DH           ; And move to return register
;**********************|
;  test for diskettes  |
;**********************|
        PUSH    DX              ; Save bit map
        AND     DL,11000000B    ; Mask off unwanted bits
        MOV     CL,6            ; Positions to shift right
        SHR     DL,CL           ; Values is now in binary
        MOV     BH,DL           ; And move to return register
        INC     BH              ; Adjust to range 1 to 4
        POP     DX              ; Restore bit map
;**********************|
;     test for 8087    |
;**********************|
        PUSH    DX              ; Save bit map
        AND     DL,00000010B    ; Mask off unwanted bits
        SHR     DL,1            ; Values is now in binary
        MOV     BL,DL           ; And move to return register
        POP     DX              ; Restore bit map
;**********************|
;    test for mouse    |
;**********************|
        AND     DL,00000100B    ; Mask off unwanted bits
        SHR     DL,1
        SHR     DL,1            ; Values is now in binary
        MOV     CH,DL           ; And move to return register
;***********************|
;    get memory size    |
;***********************|
; BIOS INT 12H returns the number of contiguos 1K memory blokcs
        PUSH    AX              ; Save return registers
        PUSH    BX
        PUSH    CX
        INT     12H
        MOV     DX,AX           ; Move memory to return register
        POP     CX              ; Restore registers
        POP     BX
        POP     AX
        MOV     CL,0            ; Clear non-meaningful register
        CLC                     ; No errors detected
        RET
EQUIPMENT       ENDP
;******************************************************************
;
MILLI_TIME      PROC     FAR
; Timer in 1/1000 second intervals
; On entry:
;         BX = time delay in milliseconds
;
READ_COUNTER:
; Read timer channel 0 at port 40H
        MOV     AL,00000110B    ; 0 0 0 0 0 1 1 0
                                ; ___ ___ _____ _
                                ;  |   |    |   |__ binary mode
                                ;  |   |    |______ mode 3
                                ;  |   |___________ latch for read
                                ;  |_______________ channel 0
        OUT     43H,AL          ; To counter command port
        JMP     SHORT $+2       ; I/O delay
; Read LSB then MSB
        IN      AL,40H
        JMP     SHORT $+2
        MOV     DL,AL
        IN      AL,40H
        JMP     SHORT $+2
        MOV     DH,AL
; DX holds timer counter value
;*******************|
;    count 1 ms     |
;*******************|
        SUB     DX,1190         ; 1 millisecond delay
ONE_MS:
        MOV     AL,06H          ; Latch for read code
        OUT     43H,AL          ; To counter command port
        JMP     SHORT $+2       ; I/O delay
; Read LSB then MSB
        IN      AL,40H          ; Port for channel 0
        JMP     SHORT $+2
        MOV     CL,AL           ; LOB to DL
        IN      AL,40H          ; Read HOB
        JMP     SHORT $+2
        MOV     CH,AL           ; HOB to DH
; CX holds timer counter value
; DX holds terminal count for 1 millisecond delay
        CMP     CX,DX           ; Compare counts
        JA      ONE_MS          ; Wait until DX < CX
;*******************|
;    repeat BX ms   |
;*******************|
        DEC     BX              ; Milliseconds counter
        NOP
        JNZ     READ_COUNTER    ; Continue if count not finished
        RET
MILLI_TIME      ENDP
;
;******************************************************************
;
SOUND           PROC    FAR
;
; On entry:
;          BX = Frequency of tone (in Hz)
;               Human audible range is 20 to 20,000 Hz
;          CX = duration of sound in milliseconds
;               1000 milliseconds = 1 second
;
;**********************|
;  calculate divisor   |
;**********************|
; Formula for calculating divisor word
;            1,193,180
;            ----------  =       divisor word
;         desired frequency
; Note: 1,193,180 = 1234DCH
;
        MOV     AX,34DCH        ; High part of numerator
        MOV     DX,12H          ; Low part of numerator
        DIV     BX              ; Divide numerator by frequency
                                ; Quotient is in AX
        MOV     BX,AX           ; Divisor word to BX
; Set port 61H (8255 port PB) for timer 2 gate and for speaker
; direct control
        IN      AL,61H          ; Read current port status
        JMP     SHORT $+2
        OR      AL,00000011B    ; xxxx xx11 bit pattern
        OUT     61H,AL          ; To port
        JMP     SHORT $+2
; Program timer channel 2
        MOV     AL,10110110B    ; 1011 0110
                                ; xxxx xxx0 = binary
                                ; xxxx 011x = mode 3 (IBM mode)
                                ; xx11 xxxx = LSB and MSB
                                ; 10xx xxxx = timer channel 2
        OUT     43H,AL          ; To 8253 command register
        JMP     SHORT $+2
; Place frequency of BX Hz as channel 2 frequency divisor
; Address of timer channel 2 = 42H
        MOV     AL,BL           ; LSB
        OUT     42H,AL          ; To timer channel 2
        JMP     SHORT $+2
        MOV     AL,BH           ; MSB
        OUT     42H,AL
        JMP     SHORT $+2
; Use TIME_DELAY procedure to wait the number of 1/18th second
; intervals passed in CX
        MOV     BX,CX           ; DX has delay in 18th second units
        CALL    MILLI_TIME
; Turn off speaker
; Reset bits 0 and 1 of port 61H to turn off the speaker
        IN      AL,61H
        JMP     SHORT $+2
        AND     AL,0FCH         ; xxxx xx00 bit pattern
        OUT     61H,AL
        RET
SOUND           ENDP
;******************************************************************
;
RANDOM_NUM      PROC    FAR
; Random number generator using channel 0 of the system timer
; On entry:
;         BX = high value of range desired, for example, to
;              obtain a randaom number in the range 0 to 100
;              BX will hold the value 100
;              Lowest allowable range is 0 to 4
; On exit:
;         DX = random number in range
;
; Note: this random number generator will give a valid random seed
;       if it is not read a regular intervals. Otherwise the
;       values may be non-random
;       The random number is generated using the shift register
;       algorithm
;
; Read timer channel 0 at port 40H
        MOV     AL,00000110B    ; 0 0 0 0 0 1 1 0
                                ; ___ ___ _____ _
                                ;  |   |    |   |__ binary mode
                                ;  |   |    |______ mode 3
                                ;  |   |___________ latch for read
                                ;  |_______________ channel 0
        OUT     43H,AL          ; To counter command port
        JMP     SHORT $+2       ; I/O delay
; Read LSB then MSB
        IN      AL,40H          ; Read LSB
        JMP     SHORT $+2
        MOV     DL,AL
        IN      AL,40H          ; Read MSB
        JMP     SHORT $+2
        MOV     DH,AL           ; Save in DH
; DX = value in timer latch
;**********************|
;   manipulate seed    |
;**********************|
; Shift register algorithm
RESCALE:
        MOV     AX,DX           ; Copy value in AX
        MOV     CL,3            ; Shift right 3 digits
        SAR     AX,CL
        ADD     AX,DX           ; Add with no carry
        MOV     CL,4            ; Value to shift left
        SAL     AX,CL           ; Shift left
        ADD     AX,DX           ; Add again
;**********************|
;    scale to range    |
;**********************|
; Dividing by the upper limit of the range will generate a
; remainder less than this limit
        MOV     DX,0            ; Use value in AX
        DIV     BX              ; Remainder is in DX
        RET
RANDOM_NUM      ENDP
;
;******************************************************************
;                         local procedures
;*******************************************************************
BIN_TO_ASC      PROC    NEAR
; Convert a 16-bit binary in the DX register to an ASCII decimal
; number
; On entry:
;         DX = binary source
;         DI --> 5-byte output buffer
; On exit:
;         DI --> start of 5-byte buffer holding the ASCII decimal
;                number
;
        PUSH    DI              ; Save start of buffer
        CALL    CLEAR_5         ; Clear user's buffer
; DI --> last byte in user's buffer
        MOV     CX,0            ; Clear counter
BINA0:  PUSH    CX              ; Save count
        MOV     AX,DX           ; Add in numerator
        MOV     DX,0            ; Clear top half
        MOV     CX,10           ; Enter decimal divisor
        DIV     CX              ; Perform division AX/CX
        XCHG    AX,DX           ; Get quotient
        ADD     AL,30H          ; Make digit ASCII
        MOV     [DI],AL         ; Store digit in buffer
        DEC     DI              ; Bump destination pointer
        POP     CX              ; Restore counter
        INC     CX              ; Count the digit
        CMP     DX,0            ; Test for end of binary
        JNZ     BINA0           ; Continue if not end
;
        POP     DI              ; Restore pointer to start of
                                ; buffer
        RET
BIN_TO_ASC      ENDP
;******************************************************************
CLEAR_5         PROC    NEAR
; On entry:
;         DS:DI --> start of a 5-byte user buffer
; On exit:
;         DS:DI --> last byte in the 5-byte buffer
        MOV     CX,5            ; Five digits to clear
CLEAR_BUF:
        MOV     BYTE PTR [DI],20H       ; Clear digit
        INC     DI              ; Bump pointer
        LOOP    CLEAR_BUF       ; Repeat for 5 digits
        DEC     DI              ; Adjust buffer pointer to last
                                ; digit
        RET
CLEAR_5         ENDP
;
P_CODE  ENDS
        END
