;*********************************|
;  OUTPUT ASCII STRING FROM FPU   |
;*********************************|
FPU_OUTPUT      PROC    FAR
; Operation:
;          Converts the value in the top register of the FPU
;          stack into an ASCII decimal number in scientific
;          notation
;
; On entry:
;         DS:DX points to an unformatted caller's buffer area
;         with minimum space for 25 bytes of storage
;
; On exit:
;         a. Carry flag clear
;         The caller's buffer will contain the ASCII decimal
;         representation of the number at the FPU stack top
;         register, formatted in scientific notation, as follows:
;
;                 sm.mmmmmmmmmmmmmmmmm ESeeee
; where
; s = sign of number (blank = +)
; m = 18 significand digits
; . = decimal point following the first significand digit
; E = explicit letter E to signal start of exponent
; S = + or - sign of exponent
; e = up to four exponent digits
;
; Examples of output:
;                    1.78125234500000000 E-12
;                   -3.14163397000000000 E+0
;                    1.22334455667788998 E+1388
;
;       b. Carry flag set (invalid value in top of stack)
;
; Other number types return the following 12-character text srings
; in the caller's buffer
;                        + denormal
;                        - denormal
;                        + NAN
;                        - NAN
;                        + infinity
;                        - infinity
;                        EMPTY
;                        unsupported
;
;      Signed zeros are reported as follows:
;                    0.00000000000000000 E+0 (positive zero)
;                   -0.00000000000000000 E+0 (negative zero)
;
;      AX is destroyed. All other registers and the caller's
;      environment are preserved
;********************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;********************************************************************
;*******************|
;   examine ST(0)   |
;*******************|
; Get condition codes for stack top register and store for later
; reference
        FXAM                    ; Examine condition codes
        FWAIT
        FSTSW   CS:STATUS_WORD  ; Store status word in memory
        FWAIT
;*******************|
; save caller's FPU |
;  environment and  |
;     registers     |
;*******************|
        FSTP    CS:TEMP_REAL    ; Save stack top in memory
        FWAIT
        FLD     CS:TEMP_REAL    ; Restore stack top value
        FSAVE   CS:STATE_80X87
        FWAIT
        FLD     CS:TEMP_REAL    ; Restore stack top value
        PUSH    CX              ; Save registers used by routine
        PUSH    DX
        PUSH    SI
        PUSH    DI
;***************************|
;  get number type at ST(0) |
;***************************|
; Code uses the procedure named NUM_AT_ST0 in the UNIT_2 module
; of MATHLIB to identify the number at ST(0)
; DS:DX --> caller's buffer for number string or error message
        PUSH    DX              ; Save caller's buffer
        XCHG    DX,DI           ; Transfer pointer to DI
        CALL    NUM_AT_ST0      ; Library procedure
        POP     DX              ; Restore caller's pointer
;*******************|
; test for invalid  |
;     values        |
;*******************|
; Code is returned in AL. Caller's buffer contains text string
        CMP     AL,1            ; Test for + or - normal
        JA      NOT_1OR2        ; Go if not
        JMP     NORMAL_NUMBER   ; Go to processing routine
; Test for zeros
NOT_1OR2:
        CMP     AL,4            ; Is it positive zero ?
        JNE     TRY_NEG_ZERO    ; If not keep testing
        JMP     IS_POS_ZERO     ; Processing for positive zero
TRY_NEG_ZERO:
        CMP     AL,5            ; Is it negative zero ?
        JNE     INVALID_EXIT    ; Go if not
        JMP     IS_NEG_ZERO     ; Processing for negative zero
;***************************|
;  exit for invalid types   |
;***************************|
INVALID_EXIT:
        FRSTOR  CS:STATE_80X87  ; Restore original FPU environment
        FWAIT
; Restore registers
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        STC                             ; Set carry flag
        RET                             ; to flag error to caller
;*******************|
;   positive or     |
;  negative zero    |
;*******************|
; Zero can be signed positive or negative. ANSI/IEEE 764 requires
; preserving the sign of zero and infinity
IS_POS_ZERO:
        LEA     SI,CS:POS_ZERO          ; Error message string
        JMP     EITHER_ZERO
IS_NEG_ZERO:
        LEA     SI,CS:NEG_ZERO          ; Error message string
EITHER_ZERO:
        CALL    FILL_CALLERS
        MOV     AL,0                    ; No error
        FRSTOR  CS:STATE_80X87  ; Restore original FPU environment
        FWAIT
; Restore registers
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        CLC                             ; No error
        RET
;*******************|
;  extract exponent |
;*******************|
; The binary exponent must be converted into a decimal exponent
NORMAL_NUMBER:

        int     03h

        FSTCW  CS:USERS_CW      ; Save caller's control word
        FWAIT
        FLDCW   CS:ROUND_UP     ; Force rounding up with reduced
                                ; precision to avoid spurious
                                ; rounding of results
        FWAIT
;                               |  ST(0)  |  ST(1)  |  ST(2)  |
        FLD     ST(0)           ;    #    |    #    |  EMPTY  |
        FABS                    ;   |#|   |    #    |  EMPTY  |
        FLD1                    ;    1    |   |#|   |   #     |
        FXCH    ST(1)           ;   |#|   |    1    |   #     |
; Make |#| = x
        FYL2X                   ; Log2-x  |    #    |
        FLDL2T                  ; Log2-10 | Log2-x  |   #     |
        FXCH    ST(1)           ; Log2-x  | Log2-10 |   #     |
        FDIV    ST,ST(1)        ;  exp    | Log2-10 |   #     |
        FXCH    ST(1)           ; Log2-10 |  exp    |   #     |
        FSTP    ST(0)           ;  exp    |   #     |  EMPTY  |
; Change control word to round down
        FLDCW   CS:ROUND_DOWN   ; Force rounding control
        FWAIT
        FRNDINT                 ;  EXP    |   #     |  EMPTY  |
TEST_OVER:
        FLD     ST(0)           ;  EXP    |  EXP    |   #     |
; Change again to normal rounding control
        FBSTP   CS:PACKED_EXP   ;  EXP    |   #     |  EMPTY  |
        FISTP   CS:BINARY_EXP   ;   #     |  EMPTY  |
;
        FLDCW   CS:USERS_CW     ; Restore original control word
;*********************|
; extract significand |
;*********************|
; The validity and required precision of the significand field
; is achieved by testing the 18th significand bit of the BCD
; result. If this bit is not a BCD digit, the significand is
; divided by 10. If this bit is 0, the significand is multiplied
; by 10. Conversion is repeated until the 18-th digit is valid
; and not zero
;                               |  ST(0)  |  ST(1) |  ST(2)  |
;                               |   #     |  EMPTY |  EMPTY  |
        FLD     ST(0)           ;   #     |    #   |  EMPTY  |
; Round and store significand, just in case number it is integer
MANTISSA:
        FRNDINT
        FBSTP   CS:PACKED_SIG   ;   #     |    #   |  EMPTY  |
; Test 18-th BCD for invalid or not zero
        FWAIT
        LEA     SI,CS:PACKED_SIG
        ADD     SI,8            ; To most significand BCD byte
        MOV     AL,CS:[SI]
        CMP     AL,0FFH         ; Integer is too large to test
        JE      TOO_LARGE
        AND     AL,11110000B    ; Mask to clear bits 0 to 3
        JNZ     END_MANTISSA
; The required precision is not yet achieved
; Multiply by 10 and repeat
        FILD    CS:TEN          ; Load multiplier 
        FMULP   ST(1),ST        ; Multiply by 10 
        FLD     ST(0)           ; Adjust stack 
        JMP     MANTISSA        ; Repeat conversion
; The significand contains invalid BCD digits
; Divide by 10 and repeat
TOO_LARGE:
        FILD    CS:TEN          ; Load divisor
        FDIVP   ST(1),ST        ; Divide by 10 
        FLD     ST(0)           ; Adjust stack
        JMP     MANTISSA        ; Repeat conversion 
; The significand is valid and to sufficient precision
;*******************|
;    restore FPU    |
;    environment    |
;*******************|
END_MANTISSA:
        FRSTOR  CS:STATE_80X87  ; Restore original FPU environment
        FWAIT
;*******************|
; format caller's   |
;      buffer       |
;*******************|
        LEA     SI,CS:FPD_MES           ; Blank message
        CALL    FILL_CALLERS
;*******************|
;   build decimal   |
;     number        |
;*******************|
; The decimal number is built from the data stored in the buffers
; PACKED_SIG and PACKED_EXP. The output number is formatted in
; exponential form and stored in the caller's buffer
        LEA     SI,CS:PACKED_SIG; Significand source
        ADD     SI,9            ; To sign byte of BCD
        MOV     DI,DX           ; User's buffer area
        MOV     AL,CS:[SI]      ; Get sign byte from BCD
        CMP     AL,0            ; Zero is positive
        JE      POS_SIGNIF      ; Significand is positive
        MOV     BYTE PTR [DI],'-'    ; Set negative sign
POS_SIGNIF:
        INC     DI              ; Bump decimal buffer pointer
        DEC     SI              ; and BCD significand pointer
; Move digit preceding the significand's decimal point
        MOV     AL,CS:[SI]      ; Get packed BCD digits into AL
        CALL    NIBBS_2_DEC     ; Decimal digits in AH and AL
        MOV     [DI],AH         ; Set first decimal digit
        INC     DI              ; Skip decimal point
        INC     DI
        MOV     [DI],AL         ; Set second digit
        INC     DI              ; Bump pointer
; Move next eight BCD digit pairs as a block
        DEC     SI              ; Point to next BCD pair
        MOV     CX,8            ; Digits to move
MOVE_8_DIGITS:
        MOV     AL,CS:[SI]      ; Get packed BCD digits into AL
        CALL    NIBBS_2_DEC     ; Decimal digits in AH and AL
        MOV     [DI],AH         ; Set first digit
        INC     DI              ; Bump decimal digits pointer
        MOV     [DI],AL         ; Place second digit
        INC     DI              ; Bump pointer
        DEC     SI              ; Source pointer to next BCD pair
        LOOP    MOVE_8_DIGITS
        ADD     DI,2            ; Bump to exponent sign position
;*******************|
;  sign of exponent |
;*******************|
        LEA     SI,CS:PACKED_EXP; Exponent BCD storage
        ADD     SI,9            ; To sign byte of BCD
        MOV     BYTE PTR [DI],'+'       ; Assume positive exponent
        MOV     AL,CS:[SI]      ; Get sign byte from BCD
        CMP     AL,0            ; Zero is positive
        JE      POS_EXPONENT    ; Significand is positive
        MOV     BYTE PTR [DI],'-'    ; Set negative sign
POS_EXPONENT:
        INC     DI              ; Bump decimal buffer pointer
;*******************|
; 4 exponent digits |
;*******************|
        LEA     SI,CS:PACKED_EXP; Reset pointer
        INC     SI              ; to first BCD digit pair of
                                ; exponent
; Move 4 exponent digits, skipping leading zeros
; BL = 0 if first non-zero digit not yet found
        MOV     BL,0            ; Switch for first non-zero
        MOV     CX,2            ; Counter for 2 packed BCDs
MOVE_2_DIGITS:
        MOV     AL,CS:[SI]      ; Get packed BCD digits into AL
        CALL    NIBBS_2_DEC     ; Decimal digits in AH and AL
        CMP     BL,0            ; Execute leading zero test ?
        JNE     NO_LEAD_TEST1   ; Switch OFF, skip test
; Test for leading zero
        CMP     AH,'0'          ; Zero ASCII digit ?
        JE      SKIP_LEAD0_1    ; Skip leading zero
; Valid no zero digit, reset lead zero test switch
        MOV     BL,0FFH         ; BL not zero
NO_LEAD_TEST1:
        MOV     [DI],AH         ; Set first decimal digit
        INC     DI              ; Bump buffer pointer
SKIP_LEAD0_1:
; Second test for leading zeros
        CMP     BL,0            ; Execute test ?
        JNE     NO_LEAD_TEST2   ; Switch OFF, skip test
        CMP     AL,'0'          ; Zero ASCII digit ?
        JE      SKIP_LEAD0_2
NO_LEAD_TEST2:
        MOV     [DI],AL         ; Set second digit
; Valid no zero digit, reset lead zero test switch
        MOV     BL,0FFH         ; BL not zero
        INC     DI              ; Bump buffer pointer
SKIP_LEAD0_2:
        DEC     SI              ; To next BCD digit pair
        LOOP    MOVE_2_DIGITS

;*******************|
;  test for all 0s  |
;*******************|
; If exponent is all zeros, the leading zero switch will be set
        CMP     BL,0            ; Is exponent all zeros?
        JNE     EXIT_OUTPUT     ; No, exit
; Set a zero digit in the exponent
        MOV     BYTE PTR [DI],'0'
;*******************|
; restore and exit  |
;*******************|
EXIT_OUTPUT:
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        CLC                     ; No error exit
        RET
FPU_OUTPUT      ENDP
;
     