;*****************************************************************
;*****************************************************************
;                           UNIT_9.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  April 6, 1994
; Updated:
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
; PROCEDURES IN THIS MODULE
PUBLIC  CALCULATE_Y     ; .. calculate value or dependent variable
PUBLIC  EVALUATE        ; .. evaluate non-parenthetical element of
                        ;    equation string
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATHUNIT library
EXTRN   ASCII_TO_EXP:   FAR
EXTRN   FPU_INPUT:      FAR
EXTRN   X_TO_Y_BYLOG:   FAR
EXTRN   SINE:           FAR
EXTRN   COSINE:         FAR
EXTRN   TANGENT:        FAR
EXTRN   ARC_SINE:       FAR
EXTRN   ARC_COSINE:     FAR
EXTRN   ARC_TANGENT:    FAR
EXTRN   DEG_TO_RAD:     FAR
EXTRN   RAD_TO_DEG:     FAR
;
INCLUDELIB MATHUNIT
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
;****************************************************************
;         procedure to solve and algebraic equation which
;               can include parenthetical elements
;****************************************************************
;****************************|
;    code segment data for   |
;    CALCULATE_Y procedure   |
;****************************|
EQUATION_BUF    DB      100 DUP (0) ; Buffer for equation text
WORK_BUF        DB      80 DUP (0)  ; Work buffer of equation
                                    ; elements
LEFT_PARS       DW      0       ; Offset of left parenthesis
X_VARIABLE      DT      0       ; Storage for caller's x
CURRENT_K       DB      0       ; Current value of Ck
                                ; in high range of CONST_ARRAY


CALCULATE_Y     PROC    FAR
; Procedure to calculate the dependent variable in a mathematical
; expression that includes parenthetical terms, possibly nested
; This procedure uses the EVALUATE procedure to solve the
; expressions enclosed in parenthesis, or with no parenthesis
;
; On entry:
;         DS:SI --> ASCII equation text in the format:
;                   y = w
;                   where w is an expression that may include
;                   parenthetical terms, the independent
;                   variable, numeric constants, and the keywords
;                   and operators listed in the header of the
;                   EVALUATE procedure
;                   The equation must terminate in 0H or 0DH
;         ST(0) = value of the independent variable (x)
;
; On exit:
;         ST(0) = principal value of the dependent variable (y)
;                 for the entry value of x
;
; PROCESSING LOGIC:
; A. Preliminary steps:
;        1. Copy equation string from = sign to terminator into
;           EQUATION_BUF
;        2. Reset pointers and clear work buffers
;
; B. Processing steps
;        1. Search EQUATION_BUF for ( symbol
;           If no ( symbol then STEP 6
;        2. Store offset of ( symbol in EQUATION_BUF as
;           LEFT_PARS
;        3. Search from LEFT_PARS for ( or ) symbols
;        4. If ( symbol found then STEP 2
;        5. If ) symbol found then move text from LEFT_PARS
;           to ) symbol into WORK_BUF and passed to EVALUATE
;           Replace ( symbol in EQUATION_BUF with result of
;           EVALUATE procedure (excess C0H) and erase text
;           in EQUATION_BUF to next ) symbol
;           Continue at STEP 1
;        6. Pass EQUATION_BUF to EVALUATE procedure
;           Apply exit logic to determine final result
;           EXIT with y = ST(0)
;
;***************************|
;      preliminary steps    |
;***************************|
        MOV     CS:CURRENT_K,0      ; Reset Ck storage address
        FSTP    CS:X_VARIABLE       ; Store x
; DS:SI --> caller's equation string
        LEA     DI,CS:EQUATION_BUF      ; Pointer to destination
; Search for = sign or for first non-ASCII in caller's buffer
CALLERS_CHAR:
        MOV     AL,[SI]         ; Get caller's character
        CMP     AL,'='          ; Test for = sign
        JE      CALLERS_START   ; Go if = sign
        CMP     AL,20H          ; Smallest ASCII code
        JB      CALLERS_START   ; Go if less than 20H
        INC     SI              ; Bump pointer
        JMP     CALLERS_CHAR    ; Continue
; At this point start of the caller's equation string is found
; Copy caller's right expression to EQUATION_BUF
CALLERS_START:
        INC     SI              ; Skip start character
MOVE_CALLERS:
        MOV     AL,[SI]         ; Get character
        CMP     AL,20H          ; Start of ASCII range
        JB      CALLERS_END     ; Less than 20H is end of string
; At this point AL holds a valid equation character
        MOV     CS:[DI],AL      ; Store in buffer
        INC     SI              ; Bump pointers
        INC     DI
        JMP     MOVE_CALLERS    ; Continue
CALLERS_END:
        MOV     BYTE PTR CS:[DI],0H ; Terminator to buffer
;***************************|
;    processing step B1:    |
;    search for ( symbol    |
;***************************|
; At this point the right element of the caller's equation is
; stored in CS:EQUATION_BUF
RESOLVE_Y:
        LEA     SI,CS:EQUATION_BUF ; Pointer to equation buffer
SEARCH_FOR_LP:
        MOV     AL,CS:[SI]      ; Get character
        CMP     AL,0H           ; Test for terminator
        JNE     VALID_CHAR      ; Go if not terminator
; At this point end of equation is found
; No parenthesis symbol in equation string
        JMP     STEP_B6         ; Go to final parse
VALID_CHAR:
        CMP     AL,'('          ; Test for left parenthesis
        JE      LEFT_PAR_FND    ; Go if found
        INC     SI              ; Bump pointer
        JMP     SEARCH_FOR_LP   ; Continue searching
;***************************|
;    processing step B2:    |
;  store offset of ( symbol |
;***************************|
; At this point a ( symbol has been found in equation string
; SI holds offset of ( symbol in EQUATION_BUF
LEFT_PAR_FND:
        MOV     CS:LEFT_PARS,SI ; Store offset of ( symbol
;***************************|
;    processing step B3:    |
; search for ( or ) symbols |
;***************************|
        INC     SI              ; Skip ( symbol
SEARCH_FOR_LR:
        MOV     AL,CS:[SI]      ; Get character
        CMP     AL,0H           ; Test for terminator
        JNE     VALID_LR        ; Go if not terminator
; At this point end of equation is found before a left or
; right parenthesis symbol
;***************************|
;    SYNTAX ERROR EXIT      |
;***************************|
SYNTAX_ERROR:
        STC                     ; Carry is syntax error flag
        RET
;***************************|
;    processing step B4:    |
;    nested parenthesis     |
;***************************|
VALID_LR:
        CMP     AL,'('          ; Test for nested parenthesis
        JE      LEFT_PAR_FND    ; Back to step 2
; Continue testing
        CMP     AL,')'          ; Test for parsable expression
        JE      PARSABLE        ; Go if not ) symbol
        INC     SI              ; Bump pointer
        JMP     SEARCH_FOR_LR   ; Continue searching
;***************************|
;    processing step B5:    |
;  parsable element found   |
;***************************|
; At this point an equation element located between ( and )
; symbols has been identified. This element can be passed to the
; EVALUALTE procedure
PARSABLE:
        MOV     SI,CS:LEFT_PARS   ; Pointer to start of parsable
                                  ; expression
        INC     SI                ; Skip ( symbol
        LEA     DI,CS:WORK_BUF    ; Pointer to buffer
COPY_TERM:
        MOV     AL,CS:[SI]        ; Character from source
        CMP     AL,')'            ; End of string ?
        JE      TERM_END          ; Go if end of string
        MOV     CS:[DI],AL        ; Store character in WORK_BUF
        INC     SI                ; Bump buffer pointers
        INC     DI
        JMP     COPY_TERM         ; Continue copying
; At this point the parsable expression is in WORK_BUF
TERM_END:
        MOV     BYTE PTR CS:[DI],0H  ; Terminator to WORK_BUF
; Set segments for EVALUATE procedure
        PUSH    DS              ; Save caller's DS
        PUSH    CS
        POP     DS              ; DS = CS
        LEA     SI,CS:WORK_BUF  ; Pointer to equation
; CS:X_VARIABLE holds value of x
        FLD     CS:X_VARIABLE   ; x to ST(0)
        CALL    FAR PTR EVALUATE   ; Equation solving procedure
        POP     DS              ; Restore caller's DS
; ST(0) holds result
; CS:CURRENT_K holds offset of next available storage
; in high area of CONST_ARRAY
        MOV     AL,CS:CURRENT_K ; Number of next storage
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage area
        MOV     CL,10           ; 10 bytes per element
        MOV     AH,0            ; Clear high byte
        MUL     CL              ; AX now holds offset
        ADD     DI,AX           ; Pointer plus offset
        ADD     DI,640          ; Skip low range
        FSTP    TBYTE PTR CS:[DI]   ; Store result in CONST_ARRAY
                                ; high area
; Store Ck reference in text string
        MOV     AL,CS:CURRENT_K ; Number of next storage
        ADD     AL,0C0H         ; Convert to Cj format
                                ; in high part of buffer
        MOV     DI,CS:LEFT_PARS ; SI --> start of expression
        MOV     BYTE PTR CS:[DI],AL ; Store result in Cj form
;
        INC     CS:CURRENT_K    ; Next storage location
; Erase text from this point to first ) symbol
CLEAR_STRING:
        INC     DI              ; Bump pointer
        CMP     BYTE PTR CS:[DI],')' ; Test for ) symbol
        JE      END_OF_PARS     ; Go if found
        MOV     BYTE PTR CS:[DI],' ' ; Place space in buffer
        JMP     CLEAR_STRING    ; Continue clearing
END_OF_PARS:
        MOV     BYTE PTR CS:[DI],' ' ; Clear ) symbol also
; Restart from start of buffer
        JMP     RESOLVE_Y
;***************************|
;    processing step B6:    |
;      final solution       |
;***************************|
; At this point the parenthetical elements in the caller's
; equation (if any) have been resolved
; EQUATION_BUF holds a parsable expression
STEP_B6:
; Set segments for EVALUATE procedure
        PUSH    DS              ; Save caller's DS
        PUSH    CS
        POP     DS              ; DS = CS
        LEA     SI,CS:EQUATION_BUF ; Pointer to equation
; CS:X_VARIABLE holds value of x
        FLD     CS:X_VARIABLE   ; x to ST(0)
        CALL    FAR PTR EVALUATE   ; Equation solving procedure
        POP     DS              ; Restore caller's DS
; Exit logic:
; If CS:CURRENT_J > 1 then the last iteration of EVALUATE
; performed calculations. Final result is in ST(0)
; If CS:CURRENT_J = 1 then the last iteration of EVALUATE
; performed no calculations. Final result as follows:
;   If CS:CURRENT_K > 0 then Ck is final result
;   If CS:CURRENT_K = 0 then CS:X_VARIABLE is final result
;***************************|
;     execute exit logic    |
;***************************|
        CMP     CS:CURRENT_J,1  ; Test result of last iteration
        JA      RES_IS_ST0      ; Result in ST(0) exit
; At this point final result is either the result of the last
; iteration that performed calculations or the independent
; variable
        CMP     CS:CURRENT_K,0  ; Test for no calculations from
                                ; previous iterations
        JE      RES_IS_X        ; Result is x exit
; At this point final result is found in Ck
        MOV     AL,CS:CURRENT_K ; Number of next storage
        DEC     AL              ; Find last operation performed
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage area
        MOV     CL,10           ; 10 bytes per element
        MOV     AH,0            ; Clear high byte
        MUL     CL              ; AX now holds offset
        ADD     DI,AX           ; Pointer plus offset
        FLD     TBYTE PTR CS:[DI]   ; Load result to ST(0)
        CLC                     ; No syntax error
        RET
; Final result is x
RES_IS_X:
        FSTP    ST(0)           ; Clear the stack
        FLD     CS:X_VARIABLE   ; Load x
        CLC                     ; No syntax error
        RET
; Final result in ST(0)
RES_IS_ST0:
; ST(0) already holds value of the dependent variable (y)
        CLC                     ; No syntax error
        RET
CALCULATE_Y     ENDP
;
;****************************************************************
;    procedure to evaluate an algebraic expression with no
;                parenthetical elements
;****************************************************************
;****************************|
;    code segment data for   |
;      EVALUATE procedure    |
;****************************|
CONST_ARRAY     DB      1280 DUP (0H); Storage for 128 constants
                                    ; in Cj format
; The first 64 locations (from j = 0 to j = 63) are for values
; generated during the execution of EVALUATE (low range)
; The locations from j = 64 to j = 127 are for constants passed
; by the caller (high range)
; Cj represents a data item in the low range
; Ck represents a data item in the high range
;
CURRENT_J       DB      0           ; Current value of Cj
                                    ; in low range
I_LEVEL         DB      0  ; Level of parenthetical indentation
SOLUTION        DT      0  ; Storage for solution of equation or
                           ; of parenthetical expression
EQ_START        DW      0  ; Entry offset of equation string
NUMERIC_BUF     DB      20 DUP (0) ; Buffer for numeric data
TEMP_BUF        DB      28 DUP (0) ; Temporary buffer
KWORD_LIST      DB      'SI',3 ; sine
                DB      'CO',3 ; cosine
                DB      'TA',3 ; tangent
                DB      'AS',4 ; arcsine
                DB      'AC',4 ; arcosine
                DB      'AT',4 ; arctangent
                DW      0      ; Terminator
; Each 2-byte ASCII keyword code is followed by the length of the
; full keyword in the equation string
THIS_TOKEN      DB      0       ; Current token control
;
EVALUATE        PROC    FAR
; Support procedure for evaluating an equation or element of
; equation that contains no parenthetical elements, according to
; the following rules of equation grammar:
;
; Recognized keywords:
;           keyword           token
;          SIN (sine)         01H
;          COS (cosine)       02H
;          TAN (tangent)      03H
;          ASIN (arcsine)     04H
;          ACOS (arcosine)    05H
;          ATAN (arctangent)  06H
;
; Recognized symbols:
;                   ^ = exponentiation
;                   * = multiplication
;                   / = division
;                   + = addition (or unary +)
;                   - = subtraction (or unary -)
;
; Other elements of equation grammar:
;            1. ASCII constants, for example:
;                   1, 2.33, or -1.2E+233
;            2. The independent variable, represented by the
;               letter x or X. On entry, ST(0) = x
;            3. Binary constants, represented by a single digit
;               entry in excess 80H format
;
; Rules of precedence:
;           1. Equation is evaluated left-to-right
;           2. Trigonometric functions are evaluated first
;           3. then exponentiation operations
;           4. then multiplication and division
;           5. addition and subtraction are performed last
;
; On entry:
;          DS:SI --> ASCII equation string terminated in 0H
;                    Equal sign and elements to its left are
;                    removed by the calling routine
;          ST(0) = x (independent variable)
;
; On exit:
;          ST(0) = solution of equation or mathematical
;                  expression passed by the caller
;
; SCANNING STAGE OPERATIONS:
;         STEP A
;            1. Store independent variable at CONST_ARRAY,
;               offset 0. Code Cj is 80H + 0(offset)
;            2. Replace 'x' and 'X' in equation string with 80H
;            3. Replace valid keywords with tokens
;               For example: SIN = 01H, COS = 02H, etc.
;
;         STEP B
;            1. Convert constants to extended precision format
;               and store constant data in CONST_ARRAY
;            2. Replace numerical data in equation string with
;               Cj, where j is the offset of the constant term
;               in CONST_ARRAY (0 based), encoded as follows:
;                     Cj = B0H + j
;               Note: single-byte storage is required in case
;                     a constant in the equation string is
;                     encoded in a single digit
;
; Scanning example:
;               Equation string before scanning:
;               x^2 + 7 * x - 257.4
;
;               Equation string after scanning:
;               x^C0 + C1 * x - C2
;               Note: after scanning all numerical data will be
;                     represented in Cj format
;
;               At this point:
;               CONST_ARRAY = 2.0, 7.0, 257.4, ------
;
; SOLVING STAGE OPERATIONS: (string is read left-to-right)
;         STEP A
;               Perform trigonometric functions (tokenized)
;               Store results in CONST_ARRAY(j)
;               Replace equation string element with Cj
;         STEP B
;               Perform al exponentiation operations
;               Store results in CONST_ARRAY(j)
;               Replace equation string element with Cj
;         STEP C
;               Perform all multiplications and divisions
;               Store results in CONST_ARRAY(j)
;               Replace equation string element with Cj
;         STEP D
;               Perform all additions and subtractions
;               Store final results in CS:SOLUTION

;***************************|
;  reset program controls   |
;***************************|
        MOV     CS:CURRENT_J,1  ; Initial value of Cj
                                ; (j = 0 is reserved for x)
;***************************|
;      scanning step A      |
;     store value of x      |
;***************************|
; DS:SI --> equation string
        MOV     CS:EQ_START,SI          ; Store offset of string
; Store x at offset 0 of CONST_ARRAY
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage
        FSTP    TBYTE PTR CS:[DI]       ; Store x at this address
;***************************|
;   replace x's in string   |
;***************************|
; Locate all incidence of x or X in the equation string and
; replace with storage reference 80H
; SI --> start of equation string
LOOK_4_X:
        CMP     BYTE PTR [SI],'X'  ; Test for X symbol
        JE      X_SYMBOL        ; Go if X
        CMP     BYTE PTR [SI],'x'  ; Test for x symbol
        JE      X_SYMBOL        ; Go if X
        CMP     BYTE PTR [SI],0H   ; Test for end of string
        JE      TOKENIZE        ; Go if terminator
; Neither x nor string end. Keep searching
        INC     SI              ; Bump pointer
        JMP     LOOK_4_X
; Incidence of x or X found in equation string
X_SYMBOL:
        MOV     BYTE PTR [SI],80H  ; Replace x with reference
        INC     SI              ; Bump pointer
        JMP     LOOK_4_X        ; Continue search
;***************************|
;   replace keywords with   |
;          tokens           |
;***************************|
; Valid keywords are:
;           SIN, COS, TAN, ASIN, ACOS, ATAN
; tokens:   01H  02H  03H  04H   05H   06H
; Note: string operations are not used due to the program setting
;       of the segment registers
TOKENIZE:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
NEW_KWORD:
        MOV     CS:THIS_TOKEN,01H  ; Initialize for first token
        LEA     DI,CS:KWORD_LIST ; List of word-size token codes
        MOV     AX,WORD PTR [SI]  ; Equation word to AX
; Test for end of equation string
        CMP     AX,0            ; 00H is end of equation string
        JE      SCAN_STEP_B     ; Go if at end
TEST_KWORD:
        CMP     WORD PTR CS:[DI],AX ; Compare with token
        JE      KWORD_FOUND      ; Go if match in keyword
; First work in token list does not match
; Test next word in token list
        ADD     DI,3            ; Each word is 2 bytes + length
        INC     CS:THIS_TOKEN   ; Next token code
; Test for end of token code lists
        CMP     WORD PTR CS:[DI],0H  ; 00H is end of token list
        JNE     TEST_KWORD      ; Go if not at end of list
; At this point all token codes have been tested
        INC     SI              ; Bump text string pointer
        JMP     NEW_KWORD
;***************************|
;      keyword in list      |
;***************************|
; At this point a token keyword has been located in the equation
; string. DS:SI --> start of keyword in string
; Token code is stored in CS:THIS_TOKEN
; Get length of keyword in equation string
KWORD_FOUND:
        MOV     CL,BYTE PTR CS:[DI+2]  ; Length follows word code
        DEC     CL               ; Allow for token
        MOV     CH,0             ; Clear high of counter
        MOV     BL,CS:THIS_TOKEN ; Get token code
        MOV     BYTE PTR [SI],BL ; Place in equation string
        INC     SI               ; Skip token position
ERASE_KWORD:
        MOV     BYTE PTR [SI],' ' ; Blank to equation text
        INC     SI                ; Bump pointer
        LOOP    ERASE_KWORD       ; Continue for CX characters
; At this point tokenization is complete for this keyword
        JMP     NEW_KWORD       ; Continue
;***************************|
;      scanning step B      |
;      store constants      |
;***************************|
; Proceed to locate constants in equation string, convert to
; extended format, store in low range of CONST_ARRAY, and replace
; numerical data in equation string with array index Cj
; (j is offset in CONST_ARRAY)
SCAN_STEP_B:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
NUMBER_AGAIN:
        CALL    SEARCH_4_NUM    ; Auxiliary procedure to search
                                ; for numeric data in string
; Carry set if no numeric data in equation string
        JNC     NUMBER_FOUND    ; Number processing if no carry
        JMP     SOLVE_STEP_A    ; Go if no other number in string
;***************************|
;   convert number and      |
;       store code          |
;***************************|
; At this point numeric data has been found in equation string
; SI --> decimal point or first ASCII-numeric digit
; Clear numeric buffer
NUMBER_FOUND:
        LEA     DI,CS:NUMERIC_BUF       ; Pointer
        MOV     CX,20           ; Buffer length
CLEAR_NUM_BUF:
        MOV     BYTE PTR CS:[DI],0      ; 0 to buffer
        INC     DI              ; Bump pointer
        LOOP    CLEAR_NUM_BUF
; Set pointer to destination buffer
        LEA     DI,CS:NUMERIC_BUF  ; Buffer for numeric string
; Move numeric data to CS:NUMERIC_BUF for conversion
;****************************|
;   check for signed number  |
;****************************|
; First check if number is preceded by sign
        MOV     AL,BYTE PTR [SI-1]
        CMP     AL,'-'         ; Is it a negative signed number ?
        JE      POSSIBLE_SIGN  ; Go if negative
        CMP     AL,'+'         ; Is it a positive signed number ?
        JE      POSSIBLE_SIGN  ; Go if positive
        JMP     SIGNED_NUM_END ; Go if not signed
; At this point the numeric digit is preceded by a + or - sign
; Code must scan next left non-space in equation string
; If next left non-space is a Cj code, then + or - symbol is
; an operation sign. If not, then - or - symbol indicates a 
; signed value
POSSIBLE_SIGN:
        PUSH    SI              ; Save pointer
        DEC     SI              ; Point to - or + symbol
SKIP_TO_LEFT:
        DEC     SI              ; Index left once
        CMP     BYTE PTR [SI],' ' ; Test for space
        JE      SKIP_TO_LEFT    ; Ignore space
        CMP     BYTE PTR [SI],80H  ; Test for Cj code
        JAE     IS_OPERATION    ; Go if Cj code found
; At this point sign is interpreted as representing a signed value
        CMP     AL,'-'         ; Is it a negative signed number ?
        JE      NEG_NUM_SIGN   ; Go if negative
        CMP     AL,'+'         ; Is it a positive signed number ?
        JE      POS_NUM_SIGN   ; Go if positive
NEG_NUM_SIGN:
        MOV     BYTE PTR CS:[DI],'-'    ; - sign to buffer
        INC     DI
; Clear sign from equation buffer
POS_NUM_SIGN:
        POP     SI              ; Restore pointer
        MOV     BYTE PTR [SI-1],' '       ; Blank character
        JMP     SIGNED_NUM_END  ; Exit this routine
IS_OPERATION:
        POP     SI              ; Restore pointer
; According to rules of equation grammar, numeric data terminates
; with any non-ASCII character, except ., e, and E
; Move data to NUMERIC_BUF until terminator
SIGNED_NUM_END:
        PUSH    SI              ; Save pointer to string start
MOVE_2_SPACE:
        MOV     AL,BYTE PTR [SI]        ; Get digit
        CALL    NUMERIC_RANGE   ; Support procedure
        JC      END_OF_NUMDATA  ; Go if non-numeric
; Character is part of a numeric sub-string
        MOV     BYTE PTR CS:[DI],AL      ; Store digit in buffer
        INC     DI              ; Bump buffer pointers
        INC     SI
        JMP     MOVE_2_SPACE    ; Continue moving digits
; At this point all numeric data is in CS:NUMERIC_BUF
; Set segments and pointers for conversion routine
END_OF_NUMDATA:
        PUSH    DS              ; Save data segment
        PUSH    CS
        POP     DS              ; DS = CS
        LEA     SI,CS:NUMERIC_BUF       ; Pointer
        LEA     DI,CS:TEMP_BUF  ; Temporary buffer for
                                ; exponential conversion
        CALL    ASCII_TO_EXP    ; Convert to exponential format
; String is now in exponential notation
        LEA     DX,CS:TEMP_BUF  ; Pointer to buffer
        CALL    FPU_INPUT       ; String to floating-point
; ST(0) now holds numeric data in FPU format
        POP     DS              ; Restore DS segment
        POP     SI              ; Pointer to start of numeric
                                ; string in equation
;***************************|
;  replace numeric data in  |
;     equation string       |
;    with Cj reference      |
;***************************|
; SI --> first numeric character
; Cj is stored in a single character by adding j to 80H
        MOV     AL,CS:CURRENT_J ; Current j in Cj
        ADD     AL,80H          ; Add in 80H
        MOV     BYTE PTR [SI],AL     ; Store in string
; Erase all remaining numeric data from equation string
ERASE_NUMERIC:
        INC     SI              ; Bump pointer
        MOV     AL,BYTE PTR [SI]     ; Read buffer character
        CALL    NUMERIC_RANGE   ; Test digit
        JC      STORE_NUMBER    ; Go if end of numeric string
        MOV     BYTE PTR [SI],' '    ; Clear character
        JMP     ERASE_NUMERIC   ; Continue
; All numeric characters erased in equation string and replace
; by the single-digit reference in Cj format
; CS:CURRENT_J holds binary number of next available storage in 
; CONST_ARRAY. Each element takes up 10 bytes
STORE_NUMBER:
        CALL    STORE_ONLY      ; Auxiliary procedure to store
                                ; numeric data in CONST_ARRAY
        JMP     NUMBER_AGAIN    ; Continue searching for
                                ; numeric data
;****************************************************************
;                    end of scanning stage
;****************************************************************
;***************************|
;  solving stage, step A:   |
;  trigonometric functions  |
;***************************|
; At this point equation numeric data is represented in Cj form
;        Cj (80H+j) = data stored in CONST_ARRAY, term j
SOLVE_STEP_A:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
TRIG_OPERATIONS:
        CALL    SEARCH_4_TOKEN  ; Support procedure
        JNC     GOT_TOKEN       ; Go if a token
        JMP     SOLVE_STEP_B    ; Exit routine if no token
; At this point DS:SI --> token in equation string
; AH = token
; Clear token from equation string
; Test token and branch to processing routine
GOT_TOKEN:
        MOV     BYTE PTR [SI],' '
        CMP     AH,01H          ; Test for SINE
        JNE     TEST_TK_02      ; Go if not this token
;***************************|
;           sine            |
;***************************|
; Processing operations for sine function
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = operand for trigonometric function
; processing assumes this operand is an angle in degrees
        CALL    DEG_TO_RAD      ; Convert to radian
        CALL    SINE            ; Calculate sine
        DEC     SI              ; Adjust pointer left
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     TRIG_OPERATIONS ; Test for another token
TEST_TK_02:
        CMP     AH,02H          ; Test for COSINE
        JNE     TEST_TK_03      ; Go if not this token
;***************************|
;         cosine            |
;***************************|
; Processing operations for cosine function
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = operand for trigonometric function
; processing assumes this operand is an angle in degrees
        CALL    DEG_TO_RAD      ; Convert to radian
        CALL    COSINE          ; Calculate cosine
        DEC     SI              ; Adjust pointer left
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     TRIG_OPERATIONS ; Test for another token
TEST_TK_03:
        CMP     AH,03H          ; Test for TANGENT
        JNE     TEST_TK_04      ; Go if not this token
;***************************|
;         tangent           |
;***************************|
; Processing operations for tangent function
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = operand for trigonometric function
; processing assumes this operand is an angle in degrees
        CALL    DEG_TO_RAD      ; Convert to radian
        CALL    TANGENT         ; Calculate tangent
        DEC     SI              ; Adjust pointer left
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     TRIG_OPERATIONS ; Test for another token
TEST_TK_04:
        CMP     AH,04H          ; Test for ARCSINE
        JNE     TEST_TK_05      ; Go if not this token
;***************************|
;        arcsine            |
;***************************|
; Processing operations for arcsine function
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = operand for trigonometric function
; processing assumes this operand refers to the arc-function of
; an angle in radian
        CALL    ARC_SINE        ; Calculate arc function
        CALL    RAD_TO_DEG      ; Convert to degrees
        DEC     SI              ; Adjust pointer left
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     TRIG_OPERATIONS ; Test for another token
TEST_TK_05:
        CMP     AH,05H          ; Test for ARCOSINE
        JNE     TEST_TK_06      ; Go if not this token
;***************************|
;        arcosine           |
;***************************|
; Processing operations for arcosine function
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = operand for trigonometric function
; processing assumes this operand refers to the arc-function of
; an angle in radian
        CALL    ARC_COSINE      ; Calculate arc function
        CALL    RAD_TO_DEG      ; Convert to degrees
        DEC     SI              ; Adjust pointer left
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     TRIG_OPERATIONS ; Test for another token
;***************************|
;        arctangent         |
;***************************|
; Operation must be arctangent
; Processing operations for arctangent function
TEST_TK_06:
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = operand for trigonometric function
; processing assumes this operand refers to the arc-function of
; an angle in radian
        CALL    ARC_TANGENT     ; Calculate arc function
        CALL    RAD_TO_DEG      ; Convert to degrees
        DEC     SI              ; Adjust pointer left
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     TRIG_OPERATIONS ; Test for another token
;***************************|
;  solving stage, step B:   |
;       exponentiation      |
;***************************|
SOLVE_STEP_B:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation

EXPONENTIATE:
        MOV     AL,'^'          ; Exponentiation symbol
        CALL    SEARCH_EQ_STR   ; Search for ^ symbol
        JC      SOLVE_STEP_C    ; Continue if no exponentiation
; At this point DS:SI --> ^ symbol
; Clear exponentiation symbol from equation string
        MOV     BYTE PTR [SI],' '
; Get exponentiation operands
        CALL    GET_LEFT_OPRND  ; Auxiliary routine to get base
; Base is now in ST(0)
        CALL    GET_RIGHT_OPRND ; Exponent
; At this point:
;     ST(0) = exponent
;     ST(1) = base
        CALL    X_TO_Y_BYLOG    ; Exponentiation routine
; ST(0) = x^y
;***************************|
;  replace numeric data in  |
;     equation string       |
;    with Cj reference      |
;***************************|
        DEC     SI              ; Point to left element
; SI --> old position of ^ symbol
        CALL    REF_AND_STORE   ; Auxiliary procedure to insert
                                ; reference and store data
        JMP     EXPONENTIATE    ; Test for another exponential
;*****************************|
;   solving stage, step C:    |
; multiplication and division |
;*****************************|
; Multiplication and division operations are performed
; left-to-right, as found in the equation string
; First do multiplications
;****************************|
;      multiplication        |
;****************************|
SOLVE_STEP_C:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
MULTIPLY:
        MOV     AL,'*'          ; Multiplication
        CALL    SEARCH_EQ_STR   ; Search for * symbol
        JC      DIVISION_OP     ; Continue if no multiplication
; At this point DS:SI --> * symbol
; Clear operation symbol from equation string
        MOV     BYTE PTR [SI],' '
; Get operands
        CALL    GET_LEFT_OPRND  ; Auxiliary procedure
; Multiplicand now in ST(0)
        CALL    GET_RIGHT_OPRND ; Auxiliary procedure
; At this point:
;     ST(0) = multiplier
;     ST(1) = multiplicand
        FMULP   ST(1),ST        ; Multiply and pop stack
; ST(0) = product
; Place reference in string and store data in CS:CONST_ARRAY
        DEC     SI              ; Point to left element
        CALL    REF_AND_STORE   ; Auxiliary procedure
        JMP     MULTIPLY        ; Test for another multiplication
;***************************|
;          division         |
;***************************|
DIVISION_OP:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
DIVIDE:
        MOV     AL,'/'          ; Division symbol
        CALL    SEARCH_EQ_STR   ; Search for / symbol
        JC      SOLVE_STEP_D    ; Continue if no division
; At this point DS:SI --> / symbol
; Clear operation symbol from equation string
        MOV     BYTE PTR [SI],' '
; Get operands
        CALL    GET_LEFT_OPRND  ; Auxiliary procedure
; Dividend now in ST(0)
        CALL    GET_RIGHT_OPRND ; Auxiliary procedure
; At this point:
;     ST(0) = dividend
;     ST(1) = divisor
; Operation is ST(1)/ST(0)
        FDIV                    ; Divide and pop stack
; ST(0) = quotient
; Place reference in string and store data in CS:CONST_ARRAY
        DEC     SI              ; Point to left element
        CALL    REF_AND_STORE   ; Auxiliary procedure
        JMP     DIVIDE          ; Test for another division
;
;*****************************|
;   solving stage, step D:    |
;  addition and subtraction   |
;*****************************|
; Addition and subtraction operations are performed left-to-right
; as found in the equation string
; First do additions
;****************************|
;         addition           |
;****************************|
SOLVE_STEP_D:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
ADDITION:
        MOV     AL,'+'          ; Addition
        CALL    SEARCH_EQ_STR   ; Search for + symbol
        JC      SUBTRACTION_OP  ; Continue if no multiplication
; At this point DS:SI --> + symbol
; Clear operation symbol from equation string
        MOV     BYTE PTR [SI],' '
; Get operands
        CALL    GET_LEFT_OPRND  ; Auxiliary procedure
; First addend now in ST(0)
        CALL    GET_RIGHT_OPRND ; Auxiliary procedure
; At this point:
;     ST(0) = second addend
;     ST(1) = first addend
        FADD                    ; Add and pop stack
; ST(0) = sum
; Place reference in string and store data in CS:CONST_ARRAY
        DEC     SI              ; Point to left element
        CALL    REF_AND_STORE   ; Auxiliary procedure
        JMP     ADDITION        ; Test for another multiplication
;***************************|
;        subtraction        |
;***************************|
SUBTRACTION_OP:
        MOV     SI,CS:EQ_START  ; Pointer to start of equation
SUBTRACT:
        MOV     AL,'-'          ; Subtraction symbol
        CALL    SEARCH_EQ_STR   ; Search for - symbol
        JC      END_STEP_D      ; Exit if no subtraction
; At this point DS:SI --> - symbol
; Clear operation symbol from equation string
        MOV     BYTE PTR [SI],' '
; Get operands
        CALL    GET_LEFT_OPRND  ; Auxiliary procedure
; Minuend now in ST(0)
        CALL    GET_RIGHT_OPRND ; Auxiliary procedure
; At this point:
;     ST(0) = subtrahend
;     ST(1) = minuend
; Operation is ST(1)-ST(0)
        FSUB                    ; Subtract and pop stack
; ST(0) = quotient
; Place reference in string and store data in CS:CONST_ARRAY
        DEC     SI              ; Point to left element
        CALL    REF_AND_STORE   ; Auxiliary procedure
        JMP     SUBTRACT        ; Test for another subtraction
;***************************|
;  all operations have been |
;         resolved          |
;***************************|
; The result of the last operation performed is stored in C(j-1)
; This logic is valid even if no operations were performed, since
; x is stored at offset 0 of CONST_ARRAY
END_STEP_D:
        MOV     AL,CS:CURRENT_J ; Number of next storage
        DEC     AL              ; Find last operation performed
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage area
        MOV     CL,10           ; 10 bytes per element
        MOV     AH,0            ; Clear high byte
        MUL     CL              ; AX now holds offset
        ADD     DI,AX           ; Pointer plus offset
        FLD     TBYTE PTR CS:[DI]   ; Load result to ST(0)
        RET
EVALUATE        ENDP
;****************************************************************
;              support procedures for EVALUATE
;****************************************************************
NUMERIC_RANGE   PROC    NEAR
; Test if AL holds an ASCII decimal digit, the letters e or E
; or the period
        CMP     AL,'e'          ; Eliminate first exception
        JE      NUMERIC_DIGIT   ; Go if e
        CMP     AL,'E'          ; Eliminate second exception
        JE      NUMERIC_DIGIT   ; Go if e
        CMP     AL,'.'          ; Test for period
        JE      NUMERIC_DIGIT   ; Go if period
; Test for ASCII range
        CMP     AL,'0'          ; Lowest allowed
        JB      NON_NUMERIC     ; Go if less than '0'
        CMP     AL,'9'          ; Highest allowed
        JA      NON_NUMERIC     ; Go if higher than '9'
NUMERIC_DIGIT:
        CLC                     ; Digit if numeric
        RET
NON_NUMERIC:
        STC                     ; Digit is not numeric
        RET
NUMERIC_RANGE   ENDP
;****************************************************************
;
REF_AND_STORE   PROC    NEAR
; Place numeric reference in text string, in Cj format, and store
; numeric data in CS:CONST_ARRAY
; Cj is stored in a single character by adding j to 80H
; Note:
;     The entry point STORE_ONLY skips the string reference
;     operation
;
; On entry:
;         DS:SI --> position in the equation string
;         ST(0) = numeric value to store and reference
;
        MOV     AL,CS:CURRENT_J ; Current j in Cj
        ADD     AL,80H          ; Add in 80H
        MOV     BYTE PTR [SI],AL     ; Store in string
STORE_ONLY:
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage area
        MOV     AL,CS:CURRENT_J ; Number of next storage
        MOV     CL,10           ; 10 bytes per element
        MOV     AH,0            ; Clear high byte
        MUL     CL              ; AX now holds offset
        ADD     DI,AX           ; Pointer plus offset
; ST(0) holds constant to be stored
        FSTP    TBYTE PTR CS:[DI]       ; Store in buffer
        INC     CS:CURRENT_J    ; Bump available storage number
        RET
REF_AND_STORE   ENDP
;****************************************************************
;
GET_LEFT_OPRND  PROC    NEAR
; Auxiliary procedure to load the element to the left of the
; operation symbol into ST(0)
; Clear numeric data from equation string
; On entry:
;         DS:DI points to arithmetic symbol ^, *, /, +, or -
; On exit:
;         ST(0) holds element to the left of the symbol
;
; Representation of left operand in equation string
;         80H+j = numeric data in CONST_STRING, entry number j
;
TEST_LEFT:
        DEC     SI              ; Pointer to previous character
        MOV     AL,[SI]         ; Character to AL
        CMP     AL,80H          ; Test for Cj data
        JAE     CJ_DATA_TYPE    ; Go if > 80H
        CMP     AL,'='          ; Left end of equation
        JNE     TEST_LEFT       ; Continue if not at left end
; At this point no left operand has been found
        STC                     ; Carry is error flag
        RET
;***************************|
;  left operand is Cj type   |
;***************************|
CJ_DATA_TYPE:
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage
        SUB     AL,80H          ; Strip-off excess 80H
        MOV     CL,10           ; 10 bytes per entry
        MOV     AH,0            ; Clear high byte
        MUL     CL              ; AX now holds offset of entry
        ADD     DI,AX           ; Add offset to pointer
        FLD     TBYTE PTR CS:[DI]  ; Entry to ST(0)
LEFT_EXIT:
; Clear numeric data in equation string
        MOV     BYTE PTR [SI],' ' ; Clear item from string
        CLC                     ; No error
        RET
GET_LEFT_OPRND  ENDP
;****************************************************************
;
GET_RIGHT_OPRND  PROC    NEAR
; Auxiliary procedure to load the element to the right of the
; operation symbol into ST(0)
; On entry:
;         DS:DI points to arithmetic symbol ^, *, /, +, or -
; On exit:
;         ST(0) holds operand to the right of the symbol
;
; Representation of left operand in equation string
;         80H+j = numeric data in CONST_STRING, entry number j
;
TEST_RIGHT:
        INC     SI              ; Pointer to next character
        MOV     AL,[SI]         ; Character to AL
        CMP     AL,80H          ; Test for Cj data
        JAE     CJ_DATA         ; Go if > 80H
        CMP     AL,0            ; Right end of equation
        JNE     TEST_RIGHT      ; Continue if not at left end
; At this point no left operand has been found
        STC                     ; Carry is error flag
        RET
;***************************|
;  left operand is Cj type   |
;***************************|
CJ_DATA:
        LEA     DI,CS:CONST_ARRAY       ; Pointer to storage
        SUB     AL,80H          ; Strip-off excess 80H
        MOV     CL,10           ; 10 bytes per entry
        MOV     AH,0            ; Clear high byte
        MUL     CL              ; AX now holds offset of entry
        ADD     DI,AX           ; Add offset to pointer
        FLD     TBYTE PTR CS:[DI]  ; Entry to ST(0)
EXIT_RIGHT:
; Clear numeric data from equation string
        MOV     BYTE PTR [SI],' ' ; Clear data from string
        CLC                     ; No error
        RET
GET_RIGHT_OPRND  ENDP
;****************************************************************
;
SEARCH_EQ_STR   PROC    NEAR
; Search equation string for character in AL
;
; On entry:
;         DS:SI --> current position in equation string
;
; On exit:
;         Carry clear if match found
;           DS_SI --> searched character
;         Carry set if no match in equation string
SEARCH_IT:
        MOV     AH,BYTE PTR [SI]        ; String character to AH
        CMP     AL,AH           ; Compare with input
        JE      CHAR_FOUND      ; Go if a match
        CMP     AH,0            ; Test for end of equation string
        JNE     NOT_EQ_END      ; Go if not at end
; At this point end of equation has been reached with no match
        STC                     ; Carry is no match code
        RET
NOT_EQ_END:
        INC     SI              ; Bump pointer
        JMP     SEARCH_IT       ; and continue search
CHAR_FOUND:
        CLC                     ; No carry is matched
        RET
SEARCH_EQ_STR   ENDP
;****************************************************************
;
SEARCH_4_TOKEN  PROC    NEAR
; Search equation string for token code (value less than 0FH)
;
; On entry:
;         DS:SI --> current position in equation string
;
; On exit:
;         Carry clear if match found
;           DS_SI --> searched character
;           AH = token
;         Carry set if no match in equation string
TOKEN_SEARCH:
        MOV     AH,BYTE PTR [SI]        ; String character to AH
        CMP     AH,0            ; Test for end of equation string
        JNE     NOT_EQU_END     ; Go if not at end
; At this point end of equation has been reached with no match
        STC                     ; Carry is no match code
        RET
NOT_EQU_END:
        CMP     AH,0FH          ; Compare with maximum size
        JBE     TOKEN_FOUND     ; Token found
NOT_TK_END:
        INC     SI              ; Bump pointer
        JMP     TOKEN_SEARCH    ; and continue search
TOKEN_FOUND:
        CLC                     ; No carry is matched
        RET
SEARCH_4_TOKEN  ENDP
;****************************************************************
;
SEARCH_4_NUM    PROC    NEAR
; Search equation string for ASCII-numeric data
;
; On entry:
;         DS:SI --> current position in equation string
;
; On exit:
;         Carry clear if ASCII-numeric data found in equation
;            DS:SI --> first ASCII digit of number
;         Carry set if no numeric data to end of equation string
;
SEARCH_NUM:
        MOV     AH,BYTE PTR [SI]   ; String character to AH
        CMP     AH,0            ; Test for string terminator
        JNE     NOT_NUM_END     ; Go if not at end
; At this point end of string has been reached with no match
        STC             ; Carry is no match flag
        RET
NOT_NUM_END:
        CMP     AH,'.'          ; Test for decimal period
        JNE     NOT_PERIOD      ; Go if not a period
        JMP     NUM_FOUND       ; Go if period
NOT_PERIOD:
        CMP     AH,'0'          ; Test for '0' digit
        JB      NOT_ASCII_DIG   ; Go if less than '0'
        CMP     AH,'9'          ; Test for '9' digit
        JA      NOT_ASCII_DIG   ; Go if greater than '9'
; At this point an ASCII digit has been found in the equation
; string
NUM_FOUND:
        CLC                     ; No carry is digit found
        RET
; Continue searching string
NOT_ASCII_DIG:
        INC     SI              ; Bump text pointer
        JMP     SEARCH_NUM
SEARCH_4_NUM    ENDP

CODE    ENDS
        END

 INC     SI              ; Bump text pointer
        JMP     SEARCH_NUM
SEARCH_4_NUM    ENDP

CODE    ENDS
        EN