;*****************************************************************
;*****************************************************************
;                           UN32_13.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date: March 19/2001  
;       updates:    
;
; Libary name: MATH32.LIB
; This module: UN32_13.ASM
; Operations : Array processing and solving linear systems        
;
;****************************************************************
;                           publics
;*****************************************************************

; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC _ROW_TIMES_SCALAR
PUBLIC _ROW_PLUS_SCALAR
PUBLIC _ROW_DIV_SCALAR
PUBLIC _ROW_MINUS_SCALAR
PUBLIC _MAT_TIMES_SCALAR
PUBLIC _MAT_DIV_SCALAR
PUBLIC _MAT_PLUS_SCALAR
PUBLIC _MAT_MINUS_SCALAR
PUBLIC _ADD_MATRICES
PUBLIC _MUL_MATRICES
PUBLIC _GAUSS_SOLVE

;****************************************************************
;               definitions for 32-bit flat model
;****************************************************************
        .486
        .MODEL flat
        .DATA
;***************************|
;   Data for this module    |
;***************************|
ELEMENT_CNT     DW      0       ; Storage for total number of
                                ; entries in matrix C
;
MAT_A_ROWS      DB      0       ; Rows in matrix A
MAT_A_COLS      DB      0       ; Columns in matrix A
MAT_B_ROWS      DB      0       ; Rows in matrix B
MAT_B_COLS      DB      0       ; Columns in matrix B
MAT_C_ROWS      DB      0       ; Rows in matrix C
MAT_C_COLS      DB      0       ; Columns in matrix C
SKIP_FACTOR     DD      0       ; Element size 
;
; Control variables for matrix multiplication
PROD_COUNT      DB      0       ; Number of product in each
                                ; multiplication iteration
WORK_PRODS      DB      0       ; Working count for number of
                                ; products
WORK_ROWS       DB      0       ; Number of rows in matrices A
                                ; and C
WORK_COLS       DB      0       ; Number of columns in matrices B
                                ; and C
START_BMAT      DD      0       ; Storage for B matrix pointer

;********************************|
; data for GAUSS_SOLVE procedure |
;********************************|
MAT_OFFSET      DD      0       ; Offset of parameter area of
                                ; matrix in the caller's data
M               DB      0       ; Number of rows in matrix
N               DB      0       ; Number of columns in matrix
ROW_SIZE        DW      0       ; N * 10 = bytes per matrix row
                                ; Variable must be less than
                                ; 255 but it is made word size to
                                ; simplify arithmetic operations
;
STATUS_WW       DW      0       ; Storage for FPU status word
CONTROL_WW      DW      0       ; Storage for FPU control word
                                ; as installed on entry
ROUND_WW        DW      0       ; Control word to truncate
;
;**********************|
; counters and control |
;      variables       |
;**********************|
TOP_ROW         DB      1       ; Start row in matrix
TEMP_TOP        DB      2       ; Row temporarily designated as
                                ; top of matrix during solution
BOT_ROW         DB      0       ; Current last row of matrix
ONE_COL         DB      0       ; Column being factorized in
                                ; phase 2
THIS_ROW        DB      0       ; Row being processed
THIS_COL        DB      0       ; Column being processed
PIVOT_ROW       DB      0       ; Number of the row holding the
                                ; pivot entry
;**********************|
;  extended precision  |
;     variables        |
;**********************|
PIV@    DT      0               ; Storage for pivot entry
FAC@    DT      0               ; Storage for multiplication
                                ; factor used in the Gaussian
                                ; algorithm
;**********************|
;  scratchpad buffers  |
;**********************|
ADD_RW  DB      256 DUP (00H)   ; Buffer for matrix row
MATZ    DB      1200 DUP (00H)  ; Matrix for GAUSS_SOLVE
                                ; calculations

;*************************************************************
;                           CODE
;*************************************************************
		.CODE

;*************************************************************
;            low-level procedures for vector arithmetic
;*************************************************************
_ROW_TIMES_SCALAR   PROC    USES esi edi ebx ebp
; Procedure to multiply a matrix row vector by a scalar
; On entry:
;         ST(0) = scalar multiplier
;         ESI --> matrix containing the row vector
;         EAX = number of row vector (0 based)
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor
; On exit:
;         entries of row vector multiplied by ST(0)

; Formula for offset of start of vector is
;              offset = [ ((i-1) * N * s) ]
; AL holds 0-based number of the desired row vector
; CL holds the number of entries per row (N)
; DL holds skip factor
        MOV     AH,0            ; Clear high-order byte
        MUL     CL              ; AX = AL * CL
; Second multiplication assumes that product will be less than
; 65535. This assumption is reasonable since the matrix space
; assigned is 400 s
        PUSH    DX              ; Save before multiply
        MOV     DH,0            ; Clear high-order byte
        MUL     DX              ; AX = AX * DL
        POP     DX              ; Restore DX
        ADD     ESI,EAX         ; Add offset to pointer
        MOV     DH,0            ; Clear high-order byte
; At this point:
;    ESI --> first entry in the matrix row
;    ST(0) holds scalar multiplier
;    ECX = number of entries in row
;    EDX = byte length of each matrix entry (8 for double precision)
ENTRIES:
        CALL    FETCH_ENTRY	
        FMUL    ST,ST(1)        ; Multiply by ST(1)
        CALL    STORE_ENTRY  
        ADD     ESI,EDX         ; Index to next entry
        LOOP    ENTRIES
        RET
_ROW_TIMES_SCALAR      ENDP
;****************************************************************

_ROW_PLUS_SCALAR   PROC    USES esi edi ebx ebp
; Procedure to add a scalar to a matrix row
; On entry:
;         ST(0) = scalar multiplier
;         ESI --> matrix containing the row vector
;         EAX = number of row vector (0 based)
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor
; On exit:
;         entries of row vector multiplied by ST(0)
;
; Formula for offset of start of vector is
;              offset = [ ((i-1) * N * s) ]
; AL holds 0-based number of the desired row vector
; CL holds the number of entries per row (N)
; DL holds skip factor (8 for double precision)
        MOV     AH,0            ; Clear high-order byte
        MUL     CL              ; AX = AL * CL
; Second multiplication assumes that product will be less than
; 65535. This assumption is reasonable since the matrix space
; assigned is 400 s
        PUSH    DX              ; Save before multiply
        MOV     DH,0            ; Clear high-order byte
        MUL     DX              ; AX = AX * DL
        POP     DX              ; Restore DX
        ADD     ESI,EAX         ; Add offset to pointer
        MOV     DH,0            ; Clear high-order byte
; At this point:
;    ESI --> first entry in the matrix row
;    ST(0) holds scalar multiplier
;    ECX = number of entries in row
;    EDX = byte length of each matrix entry (8 for double precision)
ENTRIES_A:
        CALL    FETCH_ENTRY
        FADD    ST,ST(1)        ; Add scalar 
        CALL    STORE_ENTRY 
        ADD     ESI,EDX         ; Index to next entry
        LOOP    ENTRIES_A
        RET
_ROW_PLUS_SCALAR      ENDP
;****************************************************************
;
_ROW_DIV_SCALAR          PROC
; Procedure to divide a matrix row vector by a scalar
; On entry:
;         ST(0) = scalar divisor
;         ESI --> matrix containing the row vector
;         EAX = number of row vector (0 based)
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor

; On exit:
;         Entries of row vector divided by ST(0)
;         ST(0) is preserved
; Algorithm:
;          Division is performed by obtaining the reciprocal of
;          the divisor and using the multiplication routine
;                       |    ST(0)    |    ST(1)   |  ST(2)
                        ;   divisor   |      ?     |    ?
        FLD     ST(0)   ;   divisor   |   divisor  |    ?
        FLD1            ;      1      |   divisor  |  divisor
        FDIV    ST,ST(1);  1/divisor  |      1     |  divisor
        FSTP    ST(1)   ;  1/divisor  |   divisor  |    ?
        CALL    _ROW_TIMES_SCALAR
        FSTP    ST(0)   ;   divisor   |      ?     |    ?
        CLD
        RET
_ROW_DIV_SCALAR          ENDP
;***************************************************************
;
_ROW_MINUS_SCALAR        PROC
; Procedure to subtract a scalar from the entries in a matrix
; row
; On entry:
;         ST(0) = scalar to subtract
;         ESI --> matrix containing the row vector
;         EAX = number of row vector (0 based)
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor
; On exit:
;         Scalar subtracted from entries of the row vector
; Algorithm:
;         Subtraction is performed by changing the sign of the
;         subtrahend and using the addition routine
;                       |    ST(0)    |    ST(1)   |  ST(2)
;                       |      #      |     ?
        FCHS            ;     -#      |     ?
        CALL    _ROW_PLUS_SCALAR
        FCHS            ;      #      |     ?
        CLD
        RET
_ROW_MINUS_SCALAR        ENDP
;
;*****************************************************************
;            procedures for matrix/scalar arithmetic
;*****************************************************************
_MAT_TIMES_SCALAR   PROC    USES esi edi ebx ebp
; Procedure to multiply a matrix by a scalar
; On entry:
;         ST(0) = scalar multiplier
;         ESI --> matrix containing the row vector
;         EAX = number of rows
;         ECX = number of columns
;         EDX = horizontal skip factor
; On exit:
;         entries of matrix multiplied by ST(0)
; Total number of entries is M * N
        MOV     AH,0            ; Clear high-order byte
        MUL     CL              ; AX = AL * CL
        MOV		ECX,EAX         ; Make counter in CX
; At this point:
;    ESI --> first entry in the matrix
;    ST(0) holds scalar multiplier
;    ECX = number of entries in matrix
;    EDX = byte length of each matrix entry (4, 8, or 10 bytes)
MAT_MUL:
        CALL    FETCH_ENTRY
        FMUL    ST,ST(1)        ; Multiply by ST(1)
        CALL    STORE_ENTRY
        ADD     ESI,EDX         ; Index to next entry
        LOOP    MAT_MUL
		CLD
        RET
_MAT_TIMES_SCALAR   ENDP
;****************************************************************

_MAT_PLUS_SCALAR   PROC    USES esi edi ebx ebp
; Procedure to multiply a matrix by a scalar
; On entry:
;         ST(0) = scalar multiplier
;         ESI --> matrix containing the row vector
;         EAX = number of rows in matrix
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor
; On exit:
;         entries of matrix multiplied by ST(0)
; Total number of entries is M * N
        MOV     AH,0            ; Clear high-order byte
        MUL     CL              ; AX = AL * CL
        MOV		CX,AX           ; Make counter in CX
; At this point:
;    ESI --> first entry in the matrix
;    ST(0) holds scalar multiplier
;    ECX = number of entries in matrix
;    EDX = byte length of each matrix entry (4, 8, or 10 bytes)
MAT_ADD:
        CALL    FETCH_ENTRY 
        FADD    ST,ST(1)        ; Multiply by ST(1)
        CALL    STORE_ENTRY
        ADD     ESI,EDX         ; Index to next entry
        LOOP    MAT_ADD
		CLD
        RET
_MAT_PLUS_SCALAR   ENDP

;
;****************************************************************
;
_MAT_DIV_SCALAR          PROC  USES esi edi ebx ebp
; Procedure to divide a matrix by a scalar
; On entry:
;         ST(0) = scalar divisor
;         ESI --> matrix containing the row vector
;         EAX = number of rows in matrix
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor
; On exit:
;         Entries of matrix divided by ST(0)
;         ST(0) is preserved
; Algorithm:
;          Division is performed by obtaining the reciprocal of
;          the divisor and using the multiplication routine
;                       |    ST(0)    |    ST(1)   |  ST(2)
                        ;   divisor   |      ?     |    ?
        FLD     ST(0)   ;   divisor   |   divisor  |    ?
        FLD1            ;      1      |   divisor  |  divisor
        FDIV    ST,ST(1);  1/divisor  |      1     |  divisor
        FSTP    ST(1)   ;  1/divisor  |   divisor  |    ?
        CALL    _MAT_TIMES_SCALAR
        FSTP    ST(0)   ;   divisor   |      ?     |    ?
        CLD
		RET
_MAT_DIV_SCALAR          ENDP
;****************************************************************
;
_MAT_MINUS_SCALAR        PROC  USES esi edi ebx ebp
; Procedure to subtract a scalar from the entries in a matrix
; On entry:
;         ST(0) = scalar to subtract
;         ESI --> matrix containing the row vector
;         EAX = number of rows in matrix
;         ECX = number of columns in matrix
;         EDX = horizontal skip factor
; On exit:
;         Scalar subtracted from entries of the matrix
; Algorithm:
;         Subtraction is performed by changing the sign of the
;         subtrahend and using the addition routine
;                       |    ST(0)    |    ST(1)   |  ST(2)
;                       |      #      |     ?
        FCHS            ;     -#      |     ?
        CALL    _MAT_PLUS_SCALAR
        FCHS            ;      #      |     ?
        CLD
		RET
_MAT_MINUS_SCALAR        ENDP

;***************************************************************
;                 auxiliary procedures for matrix
;                     and vector operations
;***************************************************************
FETCH_ENTRY     PROC
; Procedure to load a matrix entry to ST(0) and push the current
; ST(0) to ST(1)
; On entry:
;         ESI --> pointer to matrix entry
;         EDX = format size of matrix entries as follows:
;             = 4 if data encoded in single precision real
;             = 8 if data encoded in double precision real
;             = 10 if data encoded in extended precision real
; On exit:
;         Stack values are pushed one register
;         ST(0) holds matrix entry
;         ESI is preserved
;
        CMP     DL,4            ; Test for single precision
        JE      LOAD_SP         ; Load single precision
        CMP     DL,8            ; Test for double precision
        JE      LOAD_DP         ; Load double precision
; Data must be in extended precision
        FLD     TBYTE PTR [ESI]  ; Load 10-byte real
        RET
LOAD_SP:
        FLD     DWORD PTR [ESI]  ; Load 4-byte real
        RET
LOAD_DP:
        FLD     QWORD PTR [ESI]  ; Load 8-byte real
        RET
FETCH_ENTRY     ENDP

;****************************************************************
STORE_ENTRY     PROC
; Procedure to store ST(0) in a matrix
; On entry:
;         ESI --> pointer to matrix entry
;         EDX = format size of matrix entries as follows:
;             = 4 if data encoded in single precision real
;             = 8 if data encoded in double precision real
;             = 10 if data encoded in extended precision real
; On exit:
;         ST(0) is stored in matrix
;         Register stack is popped
;         ESI is preserved
;
        CMP     DL,4            ; Test for single precision
        JE      STOR_SP         ; Load single precision
        CMP     DL,8            ; Test for double precision
        JE      STOR_DP         ; Load double precision
; Data must be in extended precision
        FSTP    TBYTE PTR [ESI]  ; Store 10-byte real and pop
        RET
STOR_SP:
        FSTP    DWORD PTR [ESI]  ; Store 4-byte real and pop
        RET
STOR_DP:
        FSTP    QWORD PTR [ESI]  ; Store 8-byte real and pop
        RET
STORE_ENTRY     ENDP

;****************************************************************
;****************************************************************
;         procedures for matrix-by-matrix operations
;****************************************************************
;****************************************************************
;**********************************
;      matrix addition
;**********************************
_ADD_MATRICES    PROC     USES esi edi ebx ebp
; Procedure to add all the corresponding entries of two matrices
; of the same size, as follows:
;
;    A=                 B=              C=(A+B)  
;    A11  A12  A13      B11  B12  B13   A11+B11  A12+B12  A13+B13
;    A21  A22  A23      B21  B22  B23   ....
;    A31  A32  A33      B31  B32  B33                     A33+B33
;
; On entry:
;      ESI --> first matrix (A)
;      EDI --> second matrix (B)
;      EBX --> storage area for addition matrix (C)
;              Code assumes that matrix C is correctly
;              dimensioned
;      EAX = number of rows in matrix
;      ECX = number of columns in matrix
;      EDX = horizontal skip factor
;       
; On exit:
;      AX = 0 if matrices are the same size, then matrix C
;      contains sum of A + B
;
;      AX = 1 if matrices are of different size and the matrix
;      sum is undefined
;
; Note: matrix addition is defined only regarding two matrices of
;       the same size. Matrices must be of type float and of the
;       same format
;
;***************************|
;   test for equal size     |
;***************************|
        CMP     AX,CX           ; Test for matrices of equal size
        JE      GOOD_ESIZE       ; Go if same size
;***************************|
;       DATA ERROR          |
;***************************|
; At this point matrices cannot be added
        MOV     AX,1            ; Error code
		CLD
        RET
;***************************|
;  store matrix parameters |
;***************************|
; Calculate number of entries by multiplying matrix rows times
; matrix columns
GOOD_ESIZE:
       PUSH    EDX                  ; Save register
       MUL     CX                   ; Rows times columns
       MOV     ELEMENT_CNT,AX       ; Store number of entries
	   POP     EDX
; At this point:
;      ESI --> first matrix (A)
;      EDI --> second matrix (B)
;      EBX --> storage area for addition matrix (A+B)
;***************************|
;  perform matrix addition  |
;***************************|
A_PLUS_B:
; ESI --> matrix entry in matrix A
; EDX = entry size (4, 8, or 10 bytes)
        CALL    FETCH_ENTRY     ; ST(0) now holds entry of A
; Fetch entry in matrix B
        XCHG    ESI,EDI         ; ESI --> matrix B entry
        CALL    FETCH_ENTRY     ; ST(0) = matrix B entry
                                ; ST(1) = matrix A entry
        XCHG    ESI,EDI         ; Reset pointer
; Add entries
        FADD                    ; ST(0)   |  ST(1)  |  ST(2)
                                ; eA + eB | ------- |
        XCHG    EBX,ESI         ; ESI --> matrix C entry
; Store sum
        CALL    STORE_ENTRY     ; Store sum in matrix C and pop
                                ; stack
        XCHG    EBX,ESI         ; Restore pointers
; Update entries counter
        DEC     ELEMENT_CNT     ; Counter for matrix s
        JNZ     NEXT_MAT_ELE    ; Continue if not end of matrix
;***************************|
;    end of matrix addition |
;***************************|
        MOV     AX,0            ; No error flag
		CLD
        RET
;***************************|
;   index matrix pointers   |
;***************************|
; Add entry size to each matrix pointer
NEXT_MAT_ELE:
        ADD     ESI,EDX          ; Add size to pointer
        ADD     EDI,EDX
        ADD     EBX,EDX
        JMP     A_PLUS_B
;
_ADD_MATRICES    ENDP
;

;****************************************************************
;                   matrix multiplication
;****************************************************************
_MUL_MATRICES    PROC    USES esi edi ebx ebp
; Procedure to multiply two matrices (A and B) for which a matrix
; product (A * B) is defined. Matrix multiplication requires that
; the number of columns in matrix A be equal to the number of
; rows in matrix B, as follows:
;             A                   B
;           R   C               r   cr
;               |______ = ______|
;
; Example:
;
;    A=(2 by 3)          B=(3 by 4)    
;    A11  A12  A13       B11  B12  B13  B14
;    A21  A22  A23       B21  B22  B23  B24
;                        B31  B32  B33  B34
;
; The product matrix (C) will have 2 rows and 4 columns
;   C=(2 by 4)
;   C11   C12   C13  C14
;   C21   C22   C23  C24
;
; In this case the product matrix is obtained as follows:
;   C11 = (A11*B11)+(A12*B21)+(A13*B31)
;   C12 = (A11*B12)+(A12*B22)+(A13*B32)
;   C13 = (A11*B13)+(A12*B23)+(A13*B33)
;   C14 = (A11*B14)+(A12*B24)+(A13*B34)
;
;   C21 = (A21*B11)+(A22*B21)+(A23*B31)
;   C22 = (A21*B12)+(A22*B22)+(A23*B32)
;   C23 = (A21*B13)+(A22*B23)+(A23*B33)
;   C24 = (A21*B14)+(A22*B24)+(A23*B34)
;
; On entry:
;      ESI --> first matrix (A)
;      EDI --> second matrix (B)
;      EBX --> storage area for products matrix (C)
;      AH = rows in matrix A
;      AL = columns in matrix A
;      CH = rows in matrix B
;      CL = columns in matrix B
;      EDX = number of bytes per entry
; Assumes:
;      Matrix C is dimensioned as follows:
;      Columns of C = columns of B
;      Rows of C = rows of A
; On exit:
;      Matrix C is the products matrix

; Note: the entries of matrices A, B, and C must be of type float
;       and of the same data format
;
; Store number of product in each multiplication iteration
        MOV     PROD_COUNT,AL
; At this point:
;      AH = rows in matrix A
;      AL = columns in matrix A
;      CH = rows in matrix B
;      CL = columns in matrix B
; Store matrix dimensions
       MOV      MAT_A_ROWS,AH
	   MOV      MAT_A_COLS,AL
	   MOV      MAT_B_ROWS,CH
	   MOV      MAT_B_COLS,CL 
; Store skip factor
       MOV      SKIP_FACTOR,EDX
; Calculate total entries in matrix C
; Columns in C = columns in B
;    Rows in C = rows in A
        MOV     MAT_C_COLS,CL
		MOV     MAT_C_ROWS,AH
; Calculate number of products
        MOV     AH,0            ; Clear high byte of product
        MUL     CL              ; Rows times columns
        MOV     ELEMENT_CNT,AX  ; Store  count
; At this point:
;      ESI --> first matrix (A)
;      EDI --> second matrix (B)
;      EBX --> storage area for products matrix (A*B)
       MOV     START_BMAT,EDI        ; Storage for pointer
;***************************|
; initialize row and column |
;         counters          |
;***************************|
; Set up work counter for number of rows in matrix C
; This counter will be used in determining the end of the
; matrix multiplication operation
        MOV     AL,MAT_C_ROWS           ; Rows in matrix C
        MOV     WORK_ROWS,AL            ; To working counter
; Reset counter for number of columns in matrix C
; This counter will be used in resetting the matrix pointers at
; the end of each row in the products matrix
        MOV     AL,MAT_C_COLS           ; Columns in matrix C
        MOV     WORK_COLS,AL            ; To working counter
;***************************|
;  perform multiplication   |
;***************************|
NEW_PRODUCT:
; Save pointers to matrices A and B
        PUSH    ESI             ; Pointer to A
        PUSH    EDI             ; Pointer to B
; Load 0 as first entry in sum of products
        FLDZ
                                ; ST(0)   |  ST(1)  |  ST(2)
;                               |    0    |    ?    |    ?   |
; Store number of products in work counter
        MOV     AL,PROD_COUNT           ; Get count
        MOV     WORK_PRODS,AL           ; Store in work counter
A_TIMES_B:
; Fetch entry in current row of matrix A
        MOV     EDX,SKIP_FACTOR        ;  size to DL
; ESI --> matrix entry in current row of matrix A
        CALL    FETCH_ENTRY     ; ST(0) now holds entry of A
        XCHG    ESI,EDI         ; ESI --> matrix B 
        CALL    FETCH_ENTRY     ; ST(0) = matrix B 
                                ; ST(1) = matrix A 
        XCHG    ESI,EDI         ; Reset pointer
; Multiply s
                                ; ST(0)   |  ST(1)  |  ST(2)
        FMULP   ST(1),ST        ; eA * eB |previous | ------ |
                                ;         |  sum    | ------ |
        FADD                    ; p sum   | ------- |
; Test for last entry in product column
        DEC     WORK_PRODS      ; Is this last product
        JZ      NEXT_PRODUCT    ; Go if at end of products column
;***************************|
;      next product         |
;***************************|
; Index to next column of matrix A
        ADD     ESI,SKIP_FACTOR   ; Add size to pointer
; Index to next row in the same column in matrix B
        MOV     EAX,EDX           ; Horizontal skip factor to AL
		MUL     MAT_B_COLS        ; Times number of columns
        ADD     EDI,EAX           ; Add to pointer
        JMP     A_TIMES_B         ; Continue in same product column
;***************************|
;      store product        |
;***************************|
NEXT_PRODUCT:
; Restore pointers to start of current A row and B column
        POP     EDI              ; B matrix pointer
        POP     ESI              ; A matrix pointer
; At this point ST(0) has sum of products
; Store this sum as entry in products matrix (by DS:BX)
        XCHG    EBX,ESI          ; ESI --> matrix C
; Store sum
        MOV     EDX,SKIP_FACTOR  ; size to DL
        CALL    STORE_ENTRY     ; Store sum in matrix C and pop
                                ; stack
        XCHG    EBX,ESI         ; Restore pointers
; Index to next entry in matrix C
        ADD     EBX,SKIP_FACTOR      ; Add size to pointer
;***************************|
;  test for last column in  |
;        matrix C           |
;***************************|
; WORK_COLS keeps count of current column in matrix C
        DEC     WORK_COLS       ; Is this the last column in C
        JE      NEW_C_ROW       ; Go if last row
; Index to next column in matrix B
        ADD     EDI,SKIP_FACTOR        ; Add size to pointer
        JMP     NEW_PRODUCT
;***************************|
;     index to new row      |
;***************************|
; First test for end of processing
NEW_C_ROW:
        DEC     WORK_ROWS       ; Row counter in matrix C
        JNE     NEXT_C_ROW      ; Go if not last row of C
;***************************|
;       end of matrix       |
;       multiplication      |
;***************************|
        JMP     MULT_M_EXIT

;***************************|
;  next row of matrix C     |
;***************************|
; At the start of every new row in the products matrix, the
; matrix B pointer must be reset to the start of matrix B
; and the matrix A pointer to the start entry of the next
; row of matrix A
NEXT_C_ROW:
        MOV     EDI,START_BMAT  ; EDI --> start of B
        MOV     AH,0            ; Clear high byte of adder
; Pointer for matrix A
        MOV     EAX,SKIP_FACTOR    ; Entry size of A
        MUL     MAT_A_COLS         ; Size times columns
        ADD     ESI,EAX            ; ESI --> next row of A
; Reset counter for number of columns in matrix C
        MOV     AL,MAT_C_COLS      ; Columns in matrix C
        MOV     WORK_COLS,AL       ; To working counter
        JMP     NEW_PRODUCT        ; Continue processing
;***********************|
;         EXIT          | 
;***********************|
MULT_M_EXIT:
        CLD
        RET

_MUL_MATRICES    ENDP

;****************************************************************
;     procedure for solving a system of linear equations by
;            Gauss-Jordan elimination with pivots
;****************************************************************
_GAUSS_SOLVE     PROC
;
; Procedure to solve a system of linear equations by manipulating
; its augmented matrix. The method used is Gauss-Jordan
; elimination with pivoting
; On entry:
;          ESI --> augmented matrix of a system of linear
;                  equations
;          EDI --> storage area for results matrix after
;                  Gaussian elimination
;          AH = rows in matrix S
;          AL = columns in matrix S
; Assumes:
;          Matrix entries are in double precision real format
;          (8 bytes per entry)
; On exit:
;          results matrix moved to caller's buffer
;
; Error considerations:
;   To reduce rounding errors during Gaussian elimination the
;   original matrix entries are expanded to the extended
;   precision format. After the Gaussian reduction, the
;   temporary matrix is again rounded to double precision
;
; Gaussian algorithm:
;
;*********************************************|
;     phase 1 - matrix to row-echelon form    |
;*********************************************|
;
;                          1  X  X  X .....
;                          0  1  X  X .....
;                          0  0  1  X .....
;                          0  0  0  1 .....
;
;
;           1. Designate top row of matrix TOP_ROW
;
;           2. Find largest absolute value to determine PIVOT
;              Proceed down from left-most column to next column
;              Designate column of pivot entry as ONE COL
;              Phase 2 if no non-zero value found
;
;           3. If pivot row is not TOP_ROW, exchange pivot
;              row with TOP ROW
;
;           4. Divide entries in TOP ROW by pivot.
;              Check for a next row.  Exit PHASE 1 if no next
;              row
; LOOP A
;           5. Find the first non-zero value to determine FACTOR
;              Proceed from row below TOP ROW and column ONE COL
;              Search down ONE COL
;              Designate row where factor was found as TEMP_TOP
;              Exit LOOP A if no non-zero found in ONE COL
;
;           6. FACTOR * (-1) = FACTOR.
;              Multiply factor by TOP ROW to create ADD_ROW
;
;           7. Add ADD_ROW to TEMP_TOP
;
;           8. LOOP until all entries in ONE COL below TOP ROW
;              are zero
;
; LOOP A EXIT
;           9. Designate next row after TOP ROW as TOP ROW
;              Continue at step 2
;
;*************************************************|
;   phase 2 - matrix to reduced row-echelon form  |
;*************************************************|
;
;                          1  0  0  0  ......
;                          0  1  0  0  ......
;                          0  0  1  0  ......
;                          0  0  0  1  ......
;
;          13. Designate bottom row of the matrix BOT_ROW
;
;          11. Find the first leading 1 in BOT ROW
;              Designate this column as ONE COL
;              Step 15 if no leading 1 found
;
; LOOP B
;          12. Find the row above BOT ROW that does not have
;              a 0 in ONE COL. Extract this entry as FACTOR
;              Designate this row as TEMP TOP
;              Exit LOOP if no valid FACTOR found
;
;          13. FACTOR * (-1) = FACTOR
;              Multiply BOT ROW by FACTOR to create ADD ROW
;
;          14. Add ADD ROW to TEMP TOP in TEMP TOP
;              LOOP B
;
; LOOP B EXIT
;
;          15. Designate row above BOT ROW as BOT ROW
;              Exit routine if BOT ROW is the first row of
;              the matrix
;              Else, continue at STEP 11
;
;**********************|
;  set rounding mode   |
;**********************|
; For Gaussian elimination the 80x87 rounding control bits are
; set to truncate towards zero
        PUSH    AX              ; Save context
        FINIT                   ; Initialize 80x87
        FSTCW   CONTROL_WW      ; Store control word
        FWAIT                   ; Wait until stored
        MOV     AX,CONTROL_WW   ; Load control word in AX
        AND     AH,11110011B    ; Make sure bits are off
                                ; to round to nearest even
        MOV     ROUND_WW,AX     ; Store in memory
        FLDCW   ROUND_WW        ; Install new control word
		POP     AX              ; Restore
;**********************|
; initialize variables |
;**********************|
        MOV     TOP_ROW,1     ; Code segment variables
        MOV     TEMP_TOP,2
        MOV     THIS_ROW,1
        MOV     THIS_COL,1
;
        MOV     MAT_OFFSET,EDI; Offset of caller's result
                                ; buffer
;**********************|
;  matrix parameters   |
;**********************|
; Matrix parameter area has the following data:
;       register     variable         data
;         AH            M             total number of rows
;         AL            N             total number of columns

        MOV     M,AH         ; Store in scratchpad variables
        MOV     N,AL
;**********************|
; calculate row length |
;**********************|
; Some operations require indexing row-by-row. The byte variable
; ROW_SIZE will hold this value
        MOV     CL,10           ; Temporary matrix is in extended
                                ; precision format
        MOV     AH,0
        MUL     CL              ; AX = AL * CL
        MOV     ROW_SIZE,AX     ; Store. Matrix rows cannot have
                                ; move than 255 bytes
;***********************|
;    copy matrix to     |
;      scratchpad       |
;***********************|
; Caller's matrix, including parameter area, is copied to the
; procedure's scratchpad. Elements are stored in extended
; precision format (see note on error considerations at start
; of procedure)
; Code assumes that the user's matrix is in double precision
; format
; ESI --> caller's matrix
        LEA     EDI,MATZ      ; Pointer to scratchpad
; Move elements from caller's matrix to scratchpad matrix
; Change format to extended precision
; ESI --> first caller's element
; EDI --> first location in scratchpad matrix
; Get count of elements to move
        MOV     ECX,0 
		MOV     EAX,0
        MOV     AL,M            ; AL holds number of rows
        MOV     CL,N            ; CL holds number of columns
		MOV     AH,0
        MUL     CL              ; AX = AL * CL
        XCHG    ECX,EAX         ; Make counter in ECX
ELS_TO_Z:
        FLD     QWORD PTR [ESI]  ; Caller's entry to ST(0)
        FSTP    TBYTE PTR [EDI]  ; To scratchpad matrix
        ADD     ESI,8            ; Index to next caller's element
        ADD     EDI,10           ; and to next storage location
        LOOP    ELS_TO_Z
;****************************************************************
;                           phase 1
;****************************************************************
                                        ;**********************|
                                        ;        step 1        |
                                        ;**********************|
; Designate the absolute top row of the matrix as TOP_ROW
        MOV     TOP_ROW,1     ; Initialize variable
                                        ;**********************|
                                        ;        step 2        |
                                        ;**********************|
; Find the largest absolute entry proceeding column by column
; largest entry constitutes pivot
STEP2:
        CALL    GET_PIVOT       ; Local procedure to find pivot
        CMP     DH,1            ; Return code for pivot found
        JE      OKPIVOT         ; Go if pivot
;**********************|
;     no pivot found   |
;**********************|
        JMP     PHASE2          ; Cannot continue if no pivot
                                        ;**********************|
                                        ;        step 3        |
                                        ;**********************|
; Move row containing pivot to top of matrix, except if it is
; already at top
OKPIVOT:
        MOV     AL,TOP_ROW   ; Top row number to AL
		MOV     BL,PIVOT_ROW
        CMP     AL,BL           ; Compare with pivot row
        JE      PROWOK          ; Go if pivot already at top
        JMP     EXCHANGE        ; Exchange top and pivot rows
;**********************|
;   pivot row at top   |
;**********************|
PROWOK:
        JMP     STEP4
;**********************|
;  exchange pivot and  |
;       top row        |
;**********************|
; Top row is first moved to scratchpad row
EXCHANGE:
        MOV     AL,TOP_ROW      ; Top row number to AL
        CALL    MATX_ROW        ; Set ESI to matrix row in AL
        PUSH    ESI             ; Save pointer to row start
        LEA     EDI,ADD_RW      ; Scratchpad buffer for row
                                ; operations
;**********************|
;     save top row     |
;**********************|
; At this point:
;    ESI --> matrix top row
;    ESI --> row buffer in scratchpad area
        MOV     ECX,0
        MOV     CX,ROW_SIZE    ; Byte counter in CX
SAVE_TOPROW:
        MOV     AL,BYTE PTR[ESI]      ; Byte from top row
        MOV     BYTE PTR[EDI],AL      ; to scratchpad buffer
        INC     ESI              ; Bump pointers
        INC     EDI
        LOOP    SAVE_TOPROW
;**********************|
; pivot row to matrix  |
;      top row         |
;**********************|
        POP     EDI           ; EDI --> offset of top row
        MOV     AL,PIVOT_ROW ; Row number of pivot row
        CALL    MATX_ROW      ; ESI --> offset of pivot row
        PUSH    ESI           ; Save pivot row offset
		MOV     ECX,0
        MOV     CX,ROW_SIZE  ; Get elements per row
MOVE_PIVOT_ROW:
        MOV     AL,BYTE PTR[ESI]      ; Byte from pivot row
        MOV     BYTE PTR[EDI],AL      ; to top row
        INC     ESI              ; Bump pointers
        INC     EDI
        LOOP    MOVE_PIVOT_ROW
;**********************|
; old top row to pivot |
;         row          |
;**********************|
        POP     EDI              ; Offset of pivot row
        LEA     ESI,ADD_RW    ; Pointer to scratchpad buffer
		MOV     ECX,0
        MOV     CX,ROW_SIZE   ; Byte counter in CX
OLD_TOP2_PIV:
        MOV     AL,BYTE PTR[ESI]      ; Byte from scratchpad buffer
        MOV     BYTE PTR[EDI],AL      ; to old pivot row
        INC     ESI              ; Bump pointers
        INC     EDI
        LOOP    OLD_TOP2_PIV
                                        ;**********************|
                                        ;        step 4        |
                                        ;**********************|
;**********************|
;   top row / pivot    |
;**********************|
STEP4:
        CALL    SETTR           ; ESI --> to top row of matrix
; Double precision real divisor is in PIV@
        MOV     ECX,0
        MOV     CL,N         ; Number of elements in row
        CALL    DIV_BY_PIVOT    ; Local row division procedure
;**********************|
;  check for next row  |
;**********************|
; Exit if this is last row of matrix
        MOV     AL,TOP_ROW   ; Row number of this row
        CMP     AL,M         ; Compare with total rows
        JNE     STEP5           ; Continue if not last row
;**********************|
;  phase 1 completed   |
;**********************|
        JMP     PHASE2          ; Continue at phase 2
                                        ;**********************|
                                        ;        step 5        |
                                        ;**********************|
; Find the first non-zero value (factor element) in the rows
; below the current one and in the current column
STEP5:
        CALL    GET_FACTOR      ; Local search procedure
        CMP     DH,0            ; Test factor found code
        JNE     STEP6           ; Go if factor
;**********************|
;  no factor in step 5 |
;**********************|
; At this point all remaining elements in the current column are
; zero. End of loop
        JMP     EXITLUP         ; Exit LOOP A
                                        ;**********************|
                                        ;        step 6        |
                                        ;**********************|
; Change sign of factor. Multiply presently designated top row
; by factor to create the row that must be added in the
; Gaussian reduction
;**********************|
; change sign of factor|
;**********************|
STEP6:
                        ;    ST(0)    |    ST(1)    |   ST(2)
        FLD     TBYTE PTR FAC@
                        ;   factor    |      ?
        FCHS            ; factor * -1 |      ?
;**********************|
;  top row * factor    |
;**********************|
; Multiply elements in the current top row by factor and store
; product in the scratchpad buffer ADD_RW
        CALL    SETTR           ; ESI to first element in top
                                ; row
        LEA     EDI,ADD_RW      ; EDI to destination buffer
		MOV     ECX,0
        MOV     CL,N            ; Counter for number of elements
; At this point:
;   ESI --> first element in top row
;   EDI --> buffer in scratchpad area
;   ECX = number of elements per row
;   ST(0) holds factor (multiplier)
        CALL    MUL_BY_FACTOR   ; Local multiply and store
                                        ;**********************|
                                        ;        step 7        |
                                        ;**********************|
; Add buffer holding products of top row by factor (ADD_RW)
; to the matrix row number in TEMP_TOP
        CALL    SETTT           ; ESI points to first element in
                                ; row number TEMP_TOP
        PUSH    ESI
        POP     EDI              ; Do EDI via stack
        LEA     ESI,ADD_RW    ; Buffer holding products
		MOV     ECX,0
        MOV     CL,N         ; Number of elements per row
        CALL    ROW_ADD
                                        ;**********************|
                                        ;        step 8        |
                                        ;**********************|
; Continue LOOP A until all elements in the factor column
; (ONE_COL) below the row currently designated as top row
; are zero. Exit if factor was found in the last row
        MOV     AL,TEMP_TOP  ; Factor row number
        CMP     AL,M         ; Matrix row count
        JE      EXITLUP         ; Go if last row
        JMP     STEP5           ; Continue if not
                                        ;**********************|
                                        ;        step 9        |
                                        ;**********************|
; End of LOOP A
; Bump the row number currently designated as the top row and
; continue. Note that the check for a last row is performed in
; step 2
EXITLUP:
        MOV     AL,TOP_ROW   ; This top row number to AL
        INC     AL              ; Bump row counter
        MOV     TOP_ROW,AL   ; Store in variable designating
                                ; current top row

        JMP     STEP2
;****************************************************************
;                           phase 2
;****************************************************************
; In step 2 of Gaussian elimination with pivoting the matrix
; is converted to reduced row-echelon form
                                        ;**********************|
                                        ;        step 10       |
                                        ;**********************|
PHASE2:
; The number of the last row of the matrix is stored in the
; variable BOT_ROW
DOPH2:
        MOV     AL,M         ; Matrix row count to AL
        MOV     BOT_ROW,AL   ; Store row number in variable
                                        ;**********************|
                                        ;        step 11       |
                                        ;**********************|
; Find an entry in the bottom row with a value of 1. Store the
; number of the column with a 1 in ONE_COL. Go to step 15
; if no 1 entry
STEP11:
        MOV     THIS_COL,1   ; Initialize column counter
        MOV     AL,BOT_ROW   ; Number of bottom row to AL
        CALL    MATX_ROW        ; ESI to start of bottom row
;**********************|
;   find leading 1     |
;**********************|
CHK41:
        CALL    TEST_1          ; Local procedure to test the
                                ; element at ESI for a value 1
        JNC     ITIS1           ; Go if no carry
; At this point the element is not 1
;**********************|
; check for last column|
;**********************|
        MOV     AL,THIS_COL  ; Present column number
        CMP     AL,N         ; Compare with column count
        JE      NO_1_IN_ROW     ; Go if this is the last column
; Not last element in row. Continue looking for a 1 element
        ADD     ESI,10           ; Index to next element
        INC     THIS_COL     ; Bump column counter
        JMP     CHK41
;**********************|
;  no 1-element in row |
;**********************|
; Continue at step 15 if no leading 1 was found in the row
NO_1_IN_ROW:
        JMP     STEP15          ; Continue processing
;**********************|
;   1-element found    |
;**********************|
; Store number of present column in ONE_COL
ITIS1:
        MOV     AL,THIS_COL  ; Number of present column
        MOV     ONE_COL,AL   ; is stored
                                        ;**********************|
                                        ;        step 12       |
                                        ;**********************|
; Find the row above the currently designated bottom row that
; does not have a value of 0 in the one column. Extract this
; value and use as factor. Store the number of the column in the
; variable CS;TEMP_TOP. Exit LOOP B if no factor is found
STEP12:
        MOV     AL,BOT_ROW   ; Get number of present bottom
                                ; row
        DEC     AL              ; Adjust to zero-base
        MOV     THIS_ROW,AL  ; Save number as present row
;**********************|
;   test for end of    |
;       matrix         |
;**********************|
; If the current bottom row is the top matrix row there can
; be no row above it. Test for this condition
TSTAL:  MOV     AL,THIS_ROW  ; Present row
        CMP     AL,0            ; 0 is illegal
        JE      OUT_OF_MAT      ; Go if out of matrix
;**********************|
;     not at end       |
;**********************|
; At this point present row is a valid matrix row
        CALL    MATX_ROW        ; ESI to current matrix row
        MOV     AL,ONE_COL   ; Current column number
        CALL    MATX_COL        ; ESI to 1 column in current
                                ; row
; ESI --> first element in column number ONE_COL
        CALL    TEST_0          ; Test for zero
; Carry flag set if element by ESI is not zero
        JC      ELEN0           ; Go if not zero
;**********************|
;  index to next entry |
;**********************|
; Element by ESI = 0. Index to row above current row
        DEC     THIS_ROW     ; Decrement current row counter
        JMP     TSTAL           ; Go to last row test
                                ; and continue
;**********************|
;    no row above      |
;**********************|
; Exit loop if there is not row above the current row
OUT_OF_MAT:
        JMP     STEP15
;**********************|
;    valid factor      |
;**********************|
ELEN0:
        MOV     AL,THIS_ROW  ; Current row to AL
        MOV     TEMP_TOP,AL  ; Make current row temporary top
                                ; row
;**********************|
; move factor to buffer|
;**********************|
; Element by ESI is new factor. Move to factor buffer
        LEA     EDI,FAC@      ; Pointer to factor buffer
        MOV     CX,10           ; Counter for 10 bytes
FAC_TO_FBUF:
        MOV     AL,[ESI]      ; Get byte from element
        MOV     [EDI],AL      ; Place byte in factor buffer
        INC     ESI              ; Bump buffer pointers
        INC     EDI
        LOOP    FAC_TO_FBUF
                                        ;**********************|
                                        ;        step 13       |
                                        ;**********************|
; Change sign of factor in FAC@
        FLD     TBYTE PTR FAC@
                        ;   ST(0)     |    ST(1)    |    ST(3)
                        ;   factor    |      ?
        FCHS            ; factor * -1 |      ?
        FLD     ST(0)   ;   factor    |    factor   |     ?
        FSTP    TBYTE PTR FAC@
                        ;   factor    |
; Multiply row by factor and store product in ADD_RW
        MOV     AL,BOT_ROW   ; Number of bottom row to AL
        CALL    MATX_ROW        ; Set ESI to first element in this
                                ; row
        LEA     EDI,ADD_RW    ; EDI to scratchpad buffer
        MOV     CL,N         ; Entries per row
        MOV     CH,0            ; Make CX counter
;   ESI --> first element in bottom row
;   EDI --> buffer in scratchpad area
;   CX = number of elements per row
;   ST(0) holds factor (multiplier)
        CALL    MUL_BY_FACTOR
                                        ;**********************|
                                        ;        step 14       |
                                        ;**********************|
; Add buffer holding products of top row by factor (ADD_RW)
; to the matrix row number in TEMP_TOP
        CALL    SETTT           ; ESI points to first element in
                                ; row number TEMP_TOP
        PUSH    ESI
        POP     EDI              ; Do EDI via stack
        LEA     ESI,ADD_RW    ; Buffer holding products
        MOV     CL,N         ; Number of elements per row
        MOV     CH,0            ; Clear high-order byte
        CALL    ROW_ADD
;**********************|
;    check for end     |
;**********************|
; If the temporary top row is to first row of the matrix go to
; step 15, if not, continue at step 12
        MOV     AL,TEMP_TOP  ; Get row number
        CMP     AL,1            ; Test for first row of matrix
        JE      STEP15          ; Go if first row
        JMP     STEP12          ; Continue at step 12 if not
                                        ;**********************|
                                        ;        step 15       |
                                        ;**********************|
; Designate row above current bottom row as bottom row. If this
; is the last row of the matrix, calculations have concluded
; if not, continue at step 11
STEP15:
        MOV     AL,BOT_ROW   ; Row number to AL
        DEC     AL              ; Previous row
        CMP     AL,1            ; Test for top row of matrix
        JE      GAUSS_EXIT
;**********************|
;  not yet concluded   |
;**********************|
        MOV     BOT_ROW,AL   ; Store new bottom row number
        JMP     STEP11          ; and continue
;*****************************************|
;        exit Gaussian reduction          |
;*****************************************|
GAUSS_EXIT:
; Move matrix in MATZ to DS:MAT_OFFSET (caller's result
; buffer)
; Note that temporary matrix is in double precision and the
; user's matrix is assumed in double precision
        LEA     ESI,MATZ      ; Results matrix in
                                ; scratchpad area
; ESI --> caller's matrix (extended precision format)
        MOV     EDI,MAT_OFFSET        ; User's buffer offset
; EDI --> user's result matrix (double precision format)
; Move elements from scratchpad matrix to the user's buffer
; EDI --> first caller's storage location
; ESI --> first element in scratchpad matrix
; Get count of elements to move
        MOV     EAX,0
		MOV     ECX,0 
        MOV     AL,M            ; AL holds number of rows
        MOV     CL,N            ; CL holds number of columns
        MUL     CL              ; AX = AL * CL
        XCHG    ECX,EAX         ; Make counter in CX
ELS_TO_A:
        FLD     TBYTE PTR [ESI]       ; Entry from scratchpad
                                        ; matrix
        FSTP    QWORD PTR [EDI]  ; To caller's storage area
        ADD     EDI,8            ; Index to next caller's element
        ADD     ESI,10           ; and to next storage location
        LOOP    ELS_TO_A
;**********************|
;    end of routine    |
;**********************|
; Restore caller's 80x87 control word and exit
        FLDCW   CONTROL_WW   ; Caller's control word save
                                ; on entry
        CLD
        RET
_GAUSS_SOLVE     ENDP
;
;****************************************************************
;            auxiliary procedures for GAUSS_SOLVE
;****************************************************************
GET_PIVOT       PROC    
; Procedure to find the larger absolute value searching the
; matrix column by column. Starting at the leftmost column of the
; top row
; Ignore entries with zero value
; Set carry if all entries are zero
; Clear carry if pivot element was found
; Store pivot at PIV@ variable
;
;***********************|
; initialize counters   |
;***********************|
        MOV     THIS_COL,1   ; Column counter
        MOV     AL,TOP_ROW   ; Get top row number
        MOV     THIS_ROW,AL  ; Store current row
;***********************|
; set pointer to first  |
;    matrix element     |
;***********************|
; Search starts at top element of leftmost column
        CALL    SETTR           ; Local procedure to set
                                ; pointer
; Search for non-zero element in this matrix column
CHKCOL:
        CALL    COMPCOL         ; Local search procedure
        CMP     DH,1            ; Code for pivot found
        JE      PFOUND          ; Go if new pivot
;**********************|
; test for last column |
;**********************|
; At this point all elements in the matrix column are zero
; If not at end of matrix, index to next column
        MOV     AL,N            ; Last column to AL
        CMP     AL,THIS_COL     ; Compare last with this one
        JE      MATEND          ; Go if at last column
;**********************|
;    not last column   |
;**********************|
; Index ESI pointer to the first element in the next column
; Note that elements are extended precision (10 bytes long)
        INC     THIS_COL     ; Update control variable
        CALL    SETTR           ; Local procedure to set ESI
                                ; to the top row
        MOV     EAX,0
		MOV     ECX,0
        MOV     AL,THIS_COL  ; Update column counter
        DEC     AL              ; Adjust to zero base
        MOV     CL,10           ; 10 bytes per entry
        MUL     CL              ; AX = AL * CL
        ADD     ESI,EAX         ; Add offset to pointer
        JMP     CHKCOL          ; Continue
;**********************|
;  all matrix tested   |
;    no pivot found    |
;**********************|
MATEND:
        STC                     ; Return code for no pivot
        RET
;**********************|
;     pivot found      |
;**********************|
; Store column number where new pivot is located
PFOUND:
        MOV     AL,THIS_COL  ; Update control variables
        MOV     ONE_COL,AL
        CLC                     ; Return code for pivot found
        RET
GET_PIVOT       ENDP
;****************************************************************
;
COMPCOL         PROC    
; Procedure to compare elements in a matrix column
; On entry:
;         ESI --> first element in matrix column
; On exit:
;         DH = 0 if no non-zero element was found
;            = 1 if pivot element stored in PIV@
;         PIVOT_ROW holds row number of pivot element
;         DL = row counter from matrix top
;**********************|
;  initialize counters |
;**********************|
; Clear stored pivot element. Assume double precision real
        LEA     EDI,PIV@      ; Pointer to storage area
        PUSH    EDI           ; Save pointer
        MOV     ECX,10        ; Counter for 10 bytes
CLEAR_PIVOT:
        MOV     BYTE PTR [EDI],0      ; Store zero
        INC     EDI              ; Bump pointer
        LOOP    CLEAR_PIVOT
        POP     EDI
;
        MOV     AL,TOP_ROW   ; Row number to AL
        MOV     THIS_ROW,AL  ; Store value in variable
; Compare this element with value stored in PIV@
; On entry PIV@ = 0
; Skip to next entry if entry equal or smaller than PIV@
; Replace PIV@ with entry if entry > PIV@
        MOV     DH,0            ; Clear return register
        MOV     AL,TOP_ROW   ; Row counter to AL
        MOV     PIVOT_ROW,AL ; Store counter in variable
        MOV     DL,AL           ; Set row counter
; Elements are encoded in extended precision real format
; Compare operation is performed bit by bit. The sign bit is
; masked out
; At this point:
; ESI --> matrix entry in double precision format
; EDI --> stored pivot element in double precision format
;**********************|
;      compare         |
;**********************|
NEWELE:
        MOV     ECX,10           ; Bytes to test
        PUSH    ESI              ; Save pointers
        PUSH    EDI
; Index to high-order byte of double precision reals
; Note that elements are stored with the most significant byte
; at the higher numbered address
        ADD     ESI,9            ; To high-order bytes
        ADD     EDI,9
        MOV     AL,[EDI]      ; Get byte of old pivot
        AND     AL,01111111B    ; Mask off sign bit
        MOV     AH,[ESI]      ; Get byte of matrix entry
        AND     AH,01111111B    ; Mask off sign bit
NEWBY:  CMP     AL,AH           ; Old pivot compare to entry
        JC      REPLACE         ; Replace old pivot if entry
                                ; is larger
        JZ      NEXTBY          ; If both bytes are equal, test
                                ; next byte
;**********************|
;  old pivot > entry   |
;**********************|
; At this point the stored element is found larger than the
; matrix entry. Testing continues with next entry
        POP     EDI              ; Pointer to start of elements
        POP     ESI
        JMP     CONTINUE
;**********************|
;   compare next byte  |
;**********************|
NEXTBY:
        DEC     ESI              ; Bump pointers to next byte
        DEC     EDI
        MOV     AL,[EDI]      ; Get elements
        MOV     AH,[ESI]
        LOOP    NEWBY
; At this point both elements have the same absolute value
; Old pivot is preserved
        POP     EDI              ; Restore pointer to start
        POP     ESI              ; of elements
        JMP     CONTINUE
;**********************|
;  replace old pivot   |
;**********************|
; At this point a non-zero element has been found that is larger
; than the stored pivot
REPLACE:
        POP     EDI              ; Recover start addresses
        POP     ESI
        PUSH    ESI              ; Save again in stack
        PUSH    EDI
        MOV     ECX,10           ; Counter for 10 bytes
; ESI --> matrix entry in double precision format
; EDI --> stored pivot element in double precision format
REPLACE_PIVOT:
        MOV     AL,[ESI]      ; Get byte from matrix entry
        MOV     [EDI],AL      ; and store in pivot storage
        INC     ESI              ; Bump pointers
        INC     EDI
        LOOP    REPLACE_PIVOT
        POP     EDI              ; Restore pointer to start
        POP     ESI
        MOV     DH,1            ; Code for non-zero element found
; Save row number in memory variable
        MOV     PIVOT_ROW,DL ; DL holds current row
;**********************|
;   continue search    |
;**********************|
; Test for end of column. Continue search if not at end
CONTINUE:
        MOV     AL,M         ; Is this the last row?
        CMP     AL,THIS_ROW
        JE      COLEND          ; Go if last row in column
        INC     THIS_ROW     ; If not, bump pointer and
                                ; continue search
; Index to next column
        MOV     EAX,0
        MOV     AL,N         ; Columns in matrix
        MOV     CL,10           ; 10 bytes per entry
        MUL     CL              ; AX = AL * CL
        ADD     ESI,EAX
        INC     DL              ; Bump row counter
        JMP     NEWELE
;
COLEND: RET
;
COMPCOL ENDP
;****************************************************************
;
SETTR   PROC    
; Set ESI to first element in the row number currently stored
; in TOP_ROW
        LEA     ESI,MATZ      ; Start of parameter area
; Number of bytes per row is stored in the variable ROW_SIZE
        MOV     EAX,0
        MOV     AX,ROW_SIZE   ; From variable storage
; The number of rows to index depends on the number of the
; current top row
        MOV     CL,TOP_ROW    ; Top row number to CL
        DEC     CL            ; Adjust to zero base
        MUL     CL            ; AX = AL * CL
; Calculation assumes that the length of the matrix row does
; not exceed 255 bytes
        ADD     ESI,EAX
        RET
SETTR   ENDP
;****************************************************************
;
SETTT   PROC    
; Set ESI to first element in the row number currently stored
; in TEMP_TOP
;
        LEA     ESI,MATZ      ; Start of parameter area
; Number of bytes per row is stored in the variable ROW_SIZE
        MOV     EAX,0
        MOV     AX,ROW_SIZE   ; From variable storage
; The number of rows to index depends on the number of the
; current top row
        MOV     CL,TEMP_TOP   ; Top row number to CL
        DEC     CL            ; Adjust to zero base
        MUL     CL            ; AX = AL * CL
; Calculation assumes that the length of the matrix row does
; not exceed 255 bytes
        ADD     ESI,EAX
        RET
SETTT   ENDP
;****************************************************************
;
MATX_ROW        PROC    
; Procedure to set the ESI register to point to a matrix row
; On entry:
;         AL = matrix row desired
; On exit:
;         ESI --> first element in matrix row
;
        DEC     AL              ; Reduce AL to zero base
        MOV     CL,AL           ; Move value to CL
		MOV     EAX,0
        MOV     AX,ROW_SIZE     ; From memory variable
        MUL     CL              ; AX = AL * CL
        LEA     ESI,MATZ        ; Pointer to matrix start
        ADD     ESI,EAX           ; Add row offset
        RET
MATX_ROW        ENDP
;****************************************************************
;
MATX_COL        PROC    
; Procedure to set the ESI register to point to a matrix column
; On entry:
;         AL = matrix column desired
; On exit:
;         ESI --> first element in matrix column
;
        PUSH    AX
		MOV     EAX,0
		POP     AX
  
        DEC     AL              ; To zero base
        MOV     CL,10           ; 10-byte elements
        MUL     CL              ; AX = AL * CL
        ADD     ESI,EAX         ; Add offset to ESI
        RET
MATX_COL        ENDP
;****************************************************************
;
DIV_BY_PIVOT    PROC    
; Divide row elements by pivot
; On entry:
;         ESI --> first element in row to be divided
;         CX = number of entries in row
;         PIV@ holds divisor
; On exit:
;         Row elements divided by pivot
        PUSH    ESI              ; Save entry pointer
                        ;    ST(0)    |    ST(1)    |    ST(2)
        FLD     PIV@ ;   pivot     |      ?
ROW_BY_PIV:
        FLD     TBYTE PTR [ESI]
                        ;     ele     |    pivot    |      ?
        FDIV    ST,ST(1);  ele/pivot  |    pivot    |      ?
        FSTP    TBYTE PTR [ESI]
                        ;    pivot    |      ?
        ADD     ESI,10   ; Bump pointer to next element
        LOOP    ROW_BY_PIV
        FSTP    ST(0)   ;    EMPTY    |
        POP     ESI              ; Restore entry pointer
        RET
DIV_BY_PIVOT    ENDP
;****************************************************************
;
GET_FACTOR      PROC    
; Procedure to search the rows below the one presently designated
; as top row for a non-zero element to use as a factor. Search
; proceeds vertically along the column where the pivot element
; is located (ONE_COL)
; On exit:
;         DX = 1 if a factor was found
;
;**********************|
;  clear factor buffer |
;**********************|
        LEA     EDI,FAC@      ; EDI --> to factor buffer
        MOV     ECX,10           ; Buffer capacity
CLEAR_FACBUF:
        MOV     BYTE PTR [EDI],0      ; Place zero in buffer
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_FACBUF
;**********************|
;   set row counter    |
;**********************|
        MOV     AL,TOP_ROW   ; Get number of present top row
        MOV     THIS_ROW,AL  ; and store it as temporary top
                                ; row
;**********************|
;  test if top row is  |
;       last row       |
;**********************|
; AL holds the row number of the present top row
        CMP     AL,M         ; Compare with total row count
        JNE     ROW_BELOW       ; Go if there is another row
; This is a last row in the matrix
        JMP     COL_1_END       ; Go to exit procedure
;**********************|
;    get next row      |
;**********************|
ROW_BELOW:
        INC     THIS_ROW     ; Bump temporary top row counter
        INC     AL              ; and present row number
        CALL    MATX_ROW        ; Get offset of row below top
                                ; row
;**********************|
;  search for factor   |
;**********************|
; Set pointer to presently designated column (ONE_COL)
        MOV     AL,ONE_COL   ; Column number
        CALL    MATX_COL        ; Index to desired column
GETFAC:
        CALL    TEST_0          ; Test if element by ESI = 0
        JNC     NOFAC           ; No carry if element is 0
;**********************|
;     factor found     |
;**********************|
; Move factor to FAC@ and store factor row
        MOV     AL,THIS_ROW  ; Current row
        MOV     TEMP_TOP,AL  ; Store row number as temporary
                                ; top row
;**********************|
;   store factor in    |
;       buffer         |
;**********************|
; Element to be used as factor (by ESI) is moved to the factor
; buffer FAC@
        LEA     EDI,FAC@      ; Pointer to factor buffer
        MOV     ECX,10        ; Byte count to move
SAVE_FACTOR:
        MOV     AL,[ESI]      ; Get byte from matrix entry
        MOV     [EDI],AL      ; and store in factor buffer
        INC     ESI              ; Bump pointers
        INC     EDI
        LOOP    SAVE_FACTOR
;**********************|
;   exit with factor   |
;**********************|
        MOV     DH,1            ; Return code for factor found
        RET
;**********************|
;   element is zero    |
;**********************|
; Test for end of column. If not at end, index to next element
NOFAC:
        MOV     AL,THIS_ROW   ; Current row to AL
        CMP     AL,M          ; Compare with total count
        JE      COL_1_END       ; Go if last element
; Not last element in column, index to next element
        INC     THIS_ROW      ; Bump row counter
        MOV     EBX,0
		MOV     BX,ROW_SIZE
        ADD     ESI,EBX       ; Add length of row to index
                              ; to next element in column
        JMP     GETFAC
;**********************|
;   no factor found    |
;**********************|
COL_1_END:
        MOV     DH,0
        RET
GET_FACTOR      ENDP
;****************************************************************
;
MUL_BY_FACTOR   PROC    
; Multiply all elements in the currently designated top row
; by factor and store product in scratchpad buffer
; On entry:
;         ESI --> multiplicand elements
;         EDI --> scratchpad buffer for products
;         CX = number of elements per row
;         ST(0) holds multiplier (factor)
;
MUL_STORE:
                        ;    ST(0)    |    ST(1)    |    ST(2)
                        ;    factor   |      ?
        FLD     TBYTE PTR [ESI]
                        ;      #      |    factor   |     ?
        FMUL    ST,ST(1);  factor * # |    factor   |     ?
        FSTP    TBYTE PTR [EDI]
                        ;   factor    |      ?
;**********************|
; index to next element|
;**********************|
        ADD     ESI,10           ; Each element is 10 bytes
        ADD     EDI,10
        LOOP    MUL_STORE
        FSTP    ST(0)           ; Discard factor from stack
        RET
MUL_BY_FACTOR   ENDP
;****************************************************************
;
ROW_ADD         PROC    
; Add row holding products of top row by factor (ADD_RW) to
; to temporary top row (in TEMP_TOP). Sum in TEMP_TOP
; On entry:
;         ESI --> buffer row holding previous products
;         EDI --> buffer row holding temporary top row
;         CX = number of elements per row
;
ADD_BY_ROW:
                        ;    ST(0)    |    ST(1)    |    ST(2)
                        ;      ?
        FLD     TBYTE PTR [ESI]
                        ;  ele[ESI]   |     ?
        FLD     TBYTE PTR [EDI]
                        ;  ele[EDI]   |   ele[ESI]   |    ?
        FADD    ST,ST(1); [ESI]+[EDI] |   ele[ESI]   |     ?
        FSTP    TBYTE PTR [EDI]
                        ;   ele[ESI]  |      ?
        FSTP    ST(0)   ;     ?
;**********************|
; index to next element|
;**********************|
        ADD     ESI,10           ; Each element is 10 bytes
        ADD     EDI,10
        LOOP    ADD_BY_ROW
        RET
ROW_ADD         ENDP
;****************************************************************
;
TEST_0          PROC    
; Test double precision real pointed at by ESI for a zero value
; On entry:
;          ESI --> element to be tested
; On exit:
;          Carry set if element is not zero
;
                        ;    ST(0)    |    ST(1)    |    ST(2)
        FLDZ            ;      0      |      ?
        FLD     TBYTE PTR [ESI]
                        ;    ele      |      0      |     ?
        FCOMPP          ;     ?       |
        FSTSW   STATUS_WW
        FWAIT
; Relevant condition codes after FCOMPP:
; C3     C2    C1    C0
;  0      0    --     0   if  ST(0) > ST(1)
;  0      0           1   if  ST(0) < ST(1)
;  1      0           0   if  ST(0) = ST(1)
        MOV     AH,BYTE PTR STATUS_WW+1
        AND     AH,01000011B    ; Mask off all unused bits
        CMP     AH,01000000B    ; Test C3 bit
        JE      IS_ZERO         ; Go if C3 bit set
; At this point element by ESI is not zero
        STC                     ; Carry flag is signal to caller
        RET
IS_ZERO:
        CLC                     ; Clear carry for zero
        RET
TEST_0          ENDP
;****************************************************************
;
TEST_1  PROC    
; Test double precision real pointed at by ESI for a value of 1
; On entry:
;          ESI --> element to be tested
; On exit:
;          Carry set if element is not one
;
                        ;    ST(0)    |    ST(1)    |    ST(2)
        FLD1            ;      1      |      ?
        FLD     TBYTE PTR [ESI]
                        ;    ele      |      1      |     ?
        FCOMPP          ;     ?       |
        FSTSW   STATUS_WW
        FWAIT
; Relevant condition codes after FCOMPP:
; C3     C2    C1    C0
;  0      0    --     0   if  ST(0) > ST(1)
;  0      0           1   if  ST(0) < ST(1)
;  1      0           0   if  ST(0) = ST(1)
        MOV     AH,BYTE PTR STATUS_WW+1
        AND     AH,01000011B    ; Mask off all unused bits
        CMP     AH,01000000B    ; Test C3 bit
        JE      IS_ONE          ; Go if C3 bit set
; At this point element by ESI is not 1
        STC                     ; Carry flag is signal to caller
        RET
IS_ONE:
        CLC                     ; Clear carry for 1
        RET
TEST_1  ENDP


        end


