;*****************************************************************
;*****************************************************************
;                           Un32_3.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS
; by Sanchez and Canton
; Creation date:  January 8, 2001
; Last update:    April 10, 2001
;
; Libary name: MATH32.LIB
; This module: Un32_3.asm
;
; Description:
;             BCD20 arithmetic
;
;****************************************************************
;                           publics
;*****************************************************************
PUBLIC _SIGN_DIV_BCD20
PUBLIC _SIGN_MUL_BCD20
PUBLIC _SIGN_ADD_BCD20
PUBLIC _SIGN_SUB_BCD20

;****************************************************************
;               definitions for 32-bit flat model
;****************************************************************
        .486
        .MODEL flat
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATH32 library
EXTRN   _BIN_TO_BCD:  PROC   ; In Un32_1 module    

        .DATA
;***************************|
;   Data for this module    |
;***************************|
; Scratch pad area for routine work variables
ENTRY_ESI        DD      0       ; Storage for entry pointer ESI
ENTRY_EDI        DD      0       ; Storage for entry pointer EDI
ENTRY_EBX        DD      0       ; Storage for entry pointer BX
; Biased exponents
BIAS_X           DW      0       ; Biased exponent for x
BIAS_Y           DW      0       ; Biased exponent for y
BIAS_Z           DW      0       ; Biased exponent for sum
; Work areas for significand operations
SIG_L   DB      68 DUP (00H)    ; Larger significand area
SIG_S   DB      68 DUP (00H)    ; Smaller significand area
SIG_R   DB      69 DUP (00H)    ; Significand of result
                                ; Extra byte holds possible carry
; Others
LARGER  DB      0               ; 'x' or 'y', whichever is larger
CASE_12 DB      0               ; Switch for CASE 1 or CASE 2
QDIG    DD      0               ; Digits counter
; Partial products buffer for multiplication operations
PART_PROD       DB      69 DUP (00H)
                                ; Buffer for partial products
; Buffers for BCD division
DIVISOR         DB      34 DUP (00H)    ; Buffer for divisor
                DB      0FFH            ; Buffer terminator
DIVIDEND        DB      35 DUP (00H)    ; Buffer for dividend
                DB      0FFH            ; Buffer terminator
NEW_DIVIDEND    DB      35 DUP (00H)    ; Buffer for trial dividend
                DB      0FFH            ; Buffer terminator
TRIAL_PROD      DB      35 DUP (00H)    ; Buffer for trial product
                DB      0FFH            ; Buffer terminator
QUOTIENT        DB      69 DUP (00H)    ; Buffer for quotient
                DB      0FFH            ; Buffer terminator
BCD_BUF         DB      5 DUP (0H)
                DB      0H
ASC_TEMP_BUF    DB      5 DUP (0H)
                DB      0H

        .CODE


;************************************|
; signed floating point BCD addition |
;************************************|
_SIGN_ADD_BCD20     PROC
; Addition of two signed BCD numbers stored in BCD12 floating point
; format
; Operation:
;           z = x + y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          bcd1[] =  Addend (element x in z = x + y)
;          bcd2[] =  Augend (element y in z = x + y)
;          result[] is 12-byte storage area for result in BCD12 format
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       result = Sum (element z in z = x + y)
;
; This routine operates on two numbers encoded in BCD12 format (see
; Figure 2.10) as follows:
;
;          Ss e e e e m . m m m m m m m m m m m m m m m m m m
;                       |
;                       |____ implicit decimal point
;
;    S = sign of number (1 BCD digit)
;    s = sign of exponent (1 BCD digit)
;    e = exponent (4 BCD digits)
;    m = normalized significand (18 BCD digits)
;        (first significand digit must be non-zero)
;    . = implicit decimal point between the first and second
;        significand digits
;*******************************************************************
; BCD signed addition algorithm:
; CASE 1:
;    If x and y have the same sign, the absolute values are
;    added and the result has the common sign
; CASE 2:
;    If x and y have different signs, the smaller value is
;    subtracted from the larger value and the result has the sign
;    of the larger
;*******************************************************************
; Routine operations
; CASE 1 and 2:
; A. The input elements are tested for zero values. If one element
;    is zero the result is the value of the other element
; B. The packed siginificands in BCD12 format are unpacked and moved
;    into work buffers located in the code segment
; C. The upacked significands are aligned in the work buffers SIG_L
;    (for the significand of the number with the larger absolute
;     value) and SIG_S (for the significand of the number with the
;     smaller absolute value)
; CASE 1 (x and y have the same sign)
;     SIG_R will hold the significand of the sum
;     Addition operation:
;      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;  +   SIG_S =    0 0 0 0 d d d . . . d d d 0 0 0 0 0 0 0 0 0 0 0
;               -------------------------------------------------
;      SIG_R =  C s s s s s s s s s s s s s 0 0 0 0 0 0 0 0 0 0 0
;          leyend:
;             D = digits in the larger significand
;             d = digits in the smaller sginificand
;             s = digits in the sum
;             C = possible carry digit in the sum significand
;                 (in this case the exponent must be adjusted)
; CASE 2 (x and y have different signs)
;     SIG_R will hold the significand of the difference
;     Subtraction operation:
;      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;  -   SIG_S =    0 0 0 0 d d d . . . d d d 0 0 0 0 0 0 0 0 0 0 0
;               -------------------------------------------------
;      SIG_R =  B s s s s s s s s s s s s s 0 0 0 0 0 0 0 0 0 0 0
;          leyend:
;             B = possible borrow digit in the difference
; TRIVIAL CASE:
;    If the difference between exponents is larger than the number
;    of significand digits, then the aligned significands will be
;    as follows:
;      SIG_L =    D D D . . . D D D 0 0 0 0 0 0 0 0 0 0
;      SIG_S =    0 0 0 0 0 0 0 0 0 0 d d d . . . d d d
;              ----------------------------------------
;      SIG_R =  0 D D D . . . D D D 0 ? ? ? . . . ? ? ?
;    This means that the result (rounded to the format's significand
;    size) will equal the significand of the larger number. Therefore
;    the addition or subtraction operation would be trivial
; D. The exponent of the sum is the exponent of the element with
;    the larger absolute value, adjusted according to the operations
;    performed on the significands
;*******************************************************************
	     CALL    STORE_AND_CLEAR     ; Local procedure
;**********************|
;    test x or y = 0   |
;**********************|
; If x or y = 0 then the sum is the non-zero element
        CALL    TEST_4_ZERO     ; Test routine
        JNC     X_NOT_ZERO      ; Go if no carry
; At this point x is zero
        MOV     ESI,EDI         ; Set SI to point to y
MOVE_2Z:
        MOV     ECX,20          ; Count for digits to move
MOVE_20_2Z:
        MOV     AL,BYTE PTR[ESI] ; Get byte from source
        MOV     BYTE PTR[EBX],AL ; and store by BX pointer
        INC     ESI              ; Bump pointers
        INC     EBX
        LOOP    MOVE_20_2Z
        JMP     SIGN_ADD_EXIT
; End of processing
;**********************|
;    test for y = 0    |
;**********************|
X_NOT_ZERO:
        PUSH    ESI             ; Save x pointer
        MOV     ESI,EDI         ; Pointer to SI
        CALL    TEST_4_ZERO     ; Test routine
        POP     ESI             ; Restore x pointer
        JNC     XY_NOT_ZERO     ; Continue if no carry
        JMP     MOVE_2Z         ; Exit routine if y = 0
;**********************|
;  neither x nor y = 0 |
;**********************|
; Convert BCD signed exponent to biased binary exponent
; Binary bias = 10000
; 0, 1, 2 ............... 10000 ...............9997, 9998, 9999
;  negative exponents  <-- zero -->  positive exponents
;**********************|
;  exponent to biased  |
;      10000 form      |
;**********************|
; First convert exponent of x
; SI points to x
XY_NOT_ZERO:
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     WORD PTR BIAS_X,DX      ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    ESI,EDI         ; Save x and set y
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     WORD PTR BIAS_Y,DX      ; Store in scratch pad area
        XCHG    ESI,EDI         ; Restore x and y pointers
;**********************|
;   determine which    |
;   number is larger   |
;**********************|
; DX = biased exponent for y
        CMP     BIAS_X,DX      ; Compare exponent of x and y
        JA      X_BIGGER        ; x exponent is larger
        JB      Y_BIGGER        ; y exponent is larger
;**********************|
; compare significands |
;**********************|
; At this point x and y have the same exponent
; To determine which number is larger or if they are equal it is
; necessary to look at the digits in the significands
; SI --> x and DI --> y
        ADD     ESI,3           ; Pointer to significand digits
        ADD     EDI,3
        MOV     ECX,17          ; 17 packed BCD digits to compare
COMPARE_SIGS:
        MOV     AL,[ESI]        ; Get two significand digits of x
        CMP     AL,BYTE PTR[EDI]        ; Compare to two y digits
        JA      X_BIGGER        ; Go if x > y
        JB      Y_BIGGER        ; or if y > x
; At this point the compared digits are equal
        INC     ESI             ; Bump pointers
        INC     EDI
        LOOP    COMPARE_SIGS    ; Continue comparing digits
;**********************|
;       x = y          |
;**********************|
; At this point x and y are identical numbers
        MOV     BYTE PTR LARGER,'e'      ; Code for x = y
        JMP     X_EQUAL_Y       ; Go to entry point
;**********************|
; y exponent is larger |
;**********************|
Y_BIGGER:
        MOV     BYTE PTR LARGER,'y'      ; y is larger than x
; Move 34 packed BCD digits from BCD20 encoding to SIG_L buffer
X_EQUAL_Y:
        MOV     ESI,ENTRY_EDI   ; Offet of y value pointer
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_L       ; Offset of CS buffer
        MOV     ECX,17          ; 17 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
; Get exponents
        MOV     DX,WORD PTR BIAS_Y      ; Load y exponent
        MOV     WORD PTR BIAS_Z,DX      ; and store as exponent of sum
        MOV     AX,BIAS_X       ; Smaller exponent to AX
        JMP     X_OR_Y          ; Go to next routine
;**********************|
; x exponent is larger |
;**********************|
X_BIGGER:
        MOV     LARGER,'x'       ; x is larger than y
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
        MOV     ESI,ENTRY_ESI    ; Offet of x value pointer
        ADD     ESI,3            ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_L        ; Offset of buffer
        MOV     ECX,17           ; 17 packed digits to move
        CALL    UNPACK_MOVE      ; Unpack and move digits routine
; Get larger exponent
        MOV     ESI,ENTRY_EDI
		MOV     DX,BIAS_X       ; Load x exponent
        MOV     BIAS_Z,DX       ; and store as exponent of sum
        MOV     AX,BIAS_Y       ; Smaller exponent to AX
X_OR_Y:
; At this point DX holds the larger exponent and AX the smaller one
; Test for a difference that is larger than the number of digits
; in the significand (34 digits in the BCD20 format)
        SUB     DX,AX           ; Larger minus smaller
        CMP     DX,34           ; Test for limit
        JBE     NOT_TRIVIAL     ; Not trivial case if <= 18
;**********************|
;  trivial operation   |
;**********************|
; If one element is too small to affect the sum, then the result is
; the larger element
; Assume that x > y
        MOV     ESI,ENTRY_ESI   ; x value to SI
        MOV     EBX,ENTRY_EBX
        CMP     LARGER,'x'      ; Test if x > y (assumed case)
        JNE     NOT_LARGER_X    ; Go is x is not larger
        JMP     MOVE_2Z         ; Exit routine for x > y
NOT_LARGER_X:
        MOV     ESI,ENTRY_EDI   ; y entry value to SI
        JMP     MOVE_2Z
;**********************|
;    offset smaller    |
;     significand      |
;**********************|
; DX holds the difference between exponents if this difference is 18
; or less. This value is used to offset the second significand
; SI --> BCD12 encoding of smaller number
NOT_TRIVIAL:
        LEA     EDI,SIG_S       ; Buffer for smaller number
        AND     EDX,0FFFFH
        ADD     EDI,EDX         ; Add offset to destination
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        MOV	    ECX,17          ; 17 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
;    test for unequal  |
;        signs         |
;**********************|
; The sign of the number is the high nibble of the first BCD12 digit
        MOV     ESI,ENTRY_ESI         ; Set pointers to entry values
        MOV     EDI,ENTRY_EDI
; SI --> x and DI --> y
        MOV     AL,BYTE PTR [ESI]        ; x's first digit to AL
        MOV     AH,BYTE PTR [EDI]        ; y's first digit to AH
        AND     AL,11110000B    ; Clear x's low nibble
        AND     AH,11110000B    ; and y's low nibble
        CMP     AL,AH           ; Test for equal sign
        JE      SAME_SIGN
        JMP     DIFFERENT_SIGN
;************************************|
; CASE 1: x and y have the same sign |
;************************************|
SAME_SIGN:
        MOV     CASE_12,1       ; Case 1 processing switch
;**********************|
;   add significands   |
;**********************|
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S, ready for addition into SIG_R
        LEA     ESI,SIG_L     ; ESI --> to larger significand
        LEA     EDI,SIG_S     ; EDI --> to smaller significand
        LEA     EBX,SIG_R     ; EBX --> buffer for sum
        ADD     ESI,67        ; Point SI to last digit in buffer
        ADD     EDI,67        ; also DI
        ADD     EBX,68        ; and BX
        MOV     ECX,68        ; 68 digits to add
        CLC                   ; Clear carry flag for first sum
ADD_68_DIGITS:
        MOV     AL,BYTE PTR [ESI] ; One digit to AL
        MOV     AH,BYTE PTR [EDI]
        ADC     AL,AH           ; Add other digit
        AAA                     ; Convert sum to unpacked BCD
        MOV     BYTE PTR [EBX],AL     ; Store digit in buffer
        DEC     ESI             ; Bump pointers to next digit
        DEC     EDI
        DEC     EBX
        LOOP    ADD_68_DIGITS   ; Loop for 36 digits
; At this point carry is set if sum overflows 36 digits
        JNC     NO_OVF          ; Go if no overflow
        MOV     BYTE PTR [EBX],01H    ; Enter 1 if overlfow
        INC     BIAS_Z       ; Adjust exponent
NO_OVF:
; SIG_R buffer now holds sum of significands
        JMP     FORM_BCD        ; Construct BCD12 result
;************************************|
; CASE 2: x and y have different sign|
;************************************|
DIFFERENT_SIGN:
        MOV     CASE_12,2      ; Case 2 processing switch
;**********************|
;    discard x = y     |
;**********************|
; Signed addition of numbers with different signs and the same values
; result in zero
        CMP     LARGER,'e'      ; Encoding for x = y
        JNE     X_NOTEQ_Y       ; Continue if not equal
; Clear result buffer to encode zero and exit
OUTPUT_ZERO:
        MOV     ESI,ENTRY_EBX   ; Result buffer for BCD12 number
        MOV     ECX,12          ; 12 bytes in BCD12 format
ZERO_RESULT:
        MOV     BYTE PTR [ESI],00H       ; Place zero byte in buffer
        INC     ESI             ; Bump buffer pointer
        LOOP    ZERO_RESULT     ; Loop until buffer cleared
        JMP     SIGN_ADD_EXIT
;**********************|
; subtract significands|
;**********************|
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S. The number with the larger exponent is in SIG_L and the
; number with the smaller exponent in SIG_S
X_NOTEQ_Y:
        LEA     ESI,SIG_L     ; SI --> to larger significand
        LEA     EDI,SIG_S     ; DI --> to smaller significand
        LEA     EBX,SIG_R     ; BX --> buffer for sum
        ADD     ESI,35           ; Point SI to last digit in buffer
        ADD     EDI,35           ; also DI
        ADD     EBX,36           ; and BX
        MOV     ECX,36           ; 36 digits to subtract
        CLC                      ; Clear carry flag for first sum
SUB_68_DIGITS:
        MOV     AL,BYTE PTR [ESI]     ; x digit to AL
        SBB     AL,BYTE PTR [EDI]     ; Subtract y digit
        AAS                     ; Convert difference to unpacked BCD
        MOV     BYTE PTR [EBX],AL     ; Store digit in buffer
        DEC     ESI             ; Bump pointers to next digit
        DEC     EDI
        DEC     EBX
        LOOP    SUB_68_DIGITS   ; Loop for 36 digits
; At this point carry is set if there is a carry out of the last
; digit
        JNC     FORM_BCD        ; Go if no overflow
        MOV     BYTE PTR [EBX],01H    ; Enter 1 if overlfow
; SIG_R buffer now holds difference of significands
;*************************************|
;   Construct BCD20 number and exit   |
;*************************************|
FORM_BCD:
        CALL    ROUND_SIGF      ; Round off significand
        MOV     EDI,ENTRY_EBX   ; Set pointer to result buffer
                                ; for BCD12 output
        ADD     EDI,3           ; Point to significand area
;**********************|
;   store significand  |
;**********************|
; Significand of result is unpacked BCD form in the first 18 digits
; of the SIG_R buffer
; DI --> significand field of the BCD12 result buffer
        MOV     ECX,17          ; Counter for 34 bcd digits
        LEA     ESI,SIG_R       ; Pointer to unpacked digits
; Skip possible leftmost zero in buffer
        CMP     BYTE PTR [ESI],00H    ; Test for zero
        JNE     PACK_AND_MOVE   ; Go if no leading zero
        INC     ESI              ; Skip zero
; Other leading zeros in buffer require adjustment of exponent
LEADING_ZEROS:
        CMP     BYTE PTR [ESI],00H    ; Test for leading zero
        JNE     PACK_AND_MOVE    ; Go if no leading zero
        INC     ESI              ; Skip leading zero
        DEC     BIAS_Z           ; Adjust exponent
        LOOP    LEADING_ZEROS
; At this point all significand is zeros, therefore the number is
; zero
        JMP     OUTPUT_ZERO     ; Zero result exit
PACK_AND_MOVE:
        MOV     AH,BYTE PTR [ESI]     ; Get one digit
        INC     ESI             ; Bump pointer
        MOV     AL,BYTE PTR [ESI]     ; Get next digit
        INC     ESI             ; Bump pointer to next digit
        SHL     AH,4            ; Shift high digit to high nibble
        OR      AL,AH           ; Merge high nibble in AH with low
                                ; nibble in AH (pack digits)
        MOV     BYTE PTR [EDI],AL        ; Store packed digits
        INC     EDI
        LOOP    PACK_AND_MOVE
; All digits moved to the destination BCD12 buffer
        MOV     EDI,ENTRY_EBX   ; Set pointer to result buffer
                                ; sign area
; Test case switch to determine sign of result
        CMP     CASE_12,1       ; See if CASE 1
        JE      CASE1_SIGN      ; Go to sign routine for CASE 1
;**********************|
; CASE 2 sign routine  |
;**********************|
; In CASE 2 the sign of the result is the sign of the larger element
; The variable CS:LARGER holds 'x' if x > y
; Note that the possibily x = y is discarded previously
        CMP     LARGER,'x'      ; Test for x > y
        JE      CASE1_SIGN      ; Use sign of x for result
; At this point the sign of y must be used for the result
; This sign is obtained from the entry value of DI
        MOV     ESI,ENTRY_EDI           ; Set pointer the y element
        MOV     AL,BYTE PTR [ESI]       ; Get sign byte
        AND     AL,11110000B            ; Isolate sign of number
        MOV     BYTE PTR [EDI],AL       ; Store in result buffer
        JMP     EXPONENT_SIGN           ; Continue processing
;**********************|
; CASE 1 sign routine  |
;**********************|
; In CASE 1 both numbers have the same sign. The sign of the sum is
; the sign common to both elements
CASE1_SIGN:
        MOV     ESI,ENTRY_ESI           ; Set pointer the x element
        MOV     AL,BYTE PTR [ESI]       ; Get sign byte
        AND     AL,11110000B            ; Isolate sign of number
        MOV     BYTE PTR [EDI],AL       ; Store in result buffer
;**********************|
; get sign of exponent |
;**********************|
EXPONENT_SIGN:
; Exponent of result is stored in BIAS_Z in bias 10000 form
; Values < 10000 indicate a negative exponent
        MOV     DX,WORD PTR BIAS_Z    ; Exponent of result to DX
        CMP     DX,10000        ; Compare with bias
        JAE     GET_POS_VALUE   ; Go if positive exponent
; Positive exponent is already encoded in result buffer
; Encode negative exponent (0001B) into BCD12 number
; DI --> sign field
        MOV     AL,[EDI]         ; Get sign digit into AL
        MOV     AH,00000001B    ; Code for negative exponent
        OR      AL,AH           ; OR exponent sign into byte
        MOV     BYTE PTR [EDI],AL        ; Store in buffer
; Calculate absolute value of negative exponent
        MOV     AX,10000        ; Bias to AX
        SUB     AX,DX           ; Bias minus exponent
        XCHG    AX,DX           ; Move exponent to DX
        JMP     CALC_EXPONENT   ; Go to conversion routine
;**********************|
;  calculate value of  |
;  positive exponent   |
;**********************|
; DX = exponent in bias 10000 form
GET_POS_VALUE:
        MOV     AX,10000        ; Bias to AX
        SUB     DX,AX           ; Exponent minus bias
CALC_EXPONENT:
; DX = absolute value of exponent (in binary)
        MOV     AX,DX           ; For conversion routine
        PUSH    EDI             ; Save pointer for result
        CALL    _BIN_TO_BCD     ; Conversion routine
        POP     EDI             ; Restore pointer
; DH and DL holds 4 packed BCD digits
        INC     EDI             ; Bump pointer to exponent filed
        MOV     BYTE PTR [EDI],DH        ; Store two digits
        INC     EDI
        MOV     BYTE PTR [EDI],DL        ; Store two other digits
;**********************************|
;           exit routine           |
;**********************************|
SIGN_ADD_EXIT:
        MOV     ESI,ENTRY_ESI
		MOV     EDI,ENTRY_EDI
		MOV     EBX,ENTRY_EBX
		CLD
		RET
_SIGN_ADD_BCD20 ENDP


;**************************************|
; signed floating point BCD subtraction|
;**************************************|
_SIGN_SUB_BCD20  PROC
; Subtraction of two signed BCD numbers stored in BCD12 floating
; point format
; Operation:
;           z = x - y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          ESI --> Minuend (element x in z = x - y)
;          EDI --> Subtraend (element y in z = x - y)
;          EBX --> 12-byte storage area for difference (element z)
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       EBX --> difference (element z in z = x - y)
;       Entry pointers are preserved
; Operation:
;   Processing is based on the algebraic priciple of changing the
;   sign of the subtraend and proceeding as in addition
;
; change subtraend sign
        CALL    REVERSE_SIGN    ; Procedure to reverse the sign of
                                ; a BCD12 number
; Call signed addition routine
        CALL    _SIGN_ADD_BCD20
; Restore sign of subtraend to entry value
        CALL    REVERSE_SIGN
		CLD
        RET
_SIGN_SUB_BCD20  ENDP


REVERSE_SIGN    PROC
; Invert sign of a number encoded in BCD12 format
; On entry:
;         EDI --> number
; On exit: 
;         EDI --> number with sign reversed
;
        TEST    BYTE PTR [EDI],00010000B ; Bit 4 set if sign is -
        JNZ     SIGN_MINUS      ; Go if bit set
;**********************|
;    change + to -     |
;**********************|
        OR      BYTE PTR [EDI],00010000B ; Make negative
        RET
;**********************|
;    change - to +     |
;**********************|
SIGN_MINUS:
        AND     BYTE PTR [EDI],00001111B ; Make postive
        RET
REVERSE_SIGN    ENDP

;************************************|
; signed floating point BCD multiply |
;************************************|
_SIGN_MUL_BCD20  PROC
; Multiplication of two signed BCD numbers stored in BCD20 format
; Operation:
;           z = x * y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          ESI --> Muliplicand (element x in z = x * y)
;          EDI --> Multiplier (element y in z = x * y)
;          EBX --> 20-byte storage area for product in BCD20 format
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       EBX --> Product (element z in z = x * y)
;       Entry pointers are preserved
;*******************************************************************
; BCD signed multiplication algorithm
; A. If one of the factors is zero, the product is zero
;    If neither factor is zero then:
; B. If the factors have equal signs the product is positive
;    If the factors have unequal sings the product is negative
; C. The significand of the product is the product of the
;    significands of the factors
; D. The exponent of the product is the sum of the exponents of
;    the factors
;*******************************************************************
; Store entry variables and clear buffers

        CALL    STORE_AND_CLEAR         ; Local procedure
;**********************|
;    test x or y = 0   |
;**********************|
; If x or y = 0 then the product is zero
        CALL    TEST_4_ZERO     ; Test routine
        JNC     NOT_ZERO_X      ; Go if no carry
; At this point x is zero
MOVE_00:
        MOV     ECX,20           ; Count for digits to move
MOVE_20_00:
        MOV     BYTE PTR[EBX],00 ; Store zero by BX pointer
        INC     EBX              ; Bump pointer
        LOOP    MOVE_20_00
;**********************************|
;    exit from multiplication      |
;**********************************|
SIGN_MUL_EXIT:
        MOV     ESI,ENTRY_ESI  ;Restore entry pointers
        MOV     EDI,ENTRY_EDI
        MOV     EBX,ENTRY_EBX
        RET                     ; End of processing
;**********************|
;    test for y = 0    |
;**********************|
NOT_ZERO_X:
        PUSH    ESI              ; Save x pointer
        MOV     ESI,EDI           ; Pointer to SI
        CALL    TEST_4_ZERO     ; Test routine
        POP     ESI              ; Restore x pointer
        JNC     NOT_ZERO_XY     ; Continue if no carry
        JMP     MOVE_00         ; Exit routine if y = 0
;**********************|
;  neither x nor y = 0 |
;     add exponents    |
;**********************|
NOT_ZERO_XY:
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     BIAS_X,DX    ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    ESI,EDI           ; Save x and set y
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     BIAS_Y,DX    ; Store in scratch pad area
        XCHG    ESI,EDI           ; Restore x and y pointers
;**********************|
;    add exponents     |
;**********************|
        MOV     AX,BIAS_X    ; Exponent of x to AX
        ADD     AX,DX           ; Add exponents of x and y
        SUB     AX,10000        ; Subtract bias from sum
                                ; to re-scale exponent
        INC     AX              ; Normalize siginifcand by
                                ; incrementing the exponent
        MOV     BIAS_Z,AX    ; Store sum in scratch pad area
;**********************|
;    unpack and move   |
;      significands    |
;**********************|
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
        MOV     ESI,ENTRY_ESI   ; Set pointer to x
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_L       ; Offset of CS buffer
        MOV     ECX,17          ; 17 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
; Move significand of y to SIG_S buffer
        MOV     ESI,ENTRY_EDI   ; Set pointer to y
        ADD     ESI,3           ; Mantissa digits in BCD12 encoding
        LEA     EDI,SIG_S       ; Offset of CS buffer
        MOV     ECX,17          ; 17 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
; multiply significands|
;**********************|
;*******************************************************************
; Structures for multidigit unpacked BCD multiplication
; 1. PART_PROD is a 37-digit buffer to hold the product or the
;    multiplicand times each digit in the multiplier. The operation
;    can be visualized as follows:
;       COLUMN NUMBERS
;       3 3 3 3 3 3 3 3 2 2 ..... 1 1 1 1 1 1 1 1 1
;       7 6 5 4 3 2 1 0 9 8 ..... 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1
;                                 x x x x x x x x x x x x x x x x x x
;                                 y y y y y y y y y y y y y y y y y y
; ROWS --------------------------------------------------------------
; 1     0 0 0 0 0 0 0 0 0 0 ....C z z z z z z z z z z z z z z z z z z
; 2     0 0 0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0
; 3     0 0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0 0
; 4     0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0 0 0
; 17    0 C z z z z ... z z z z z z 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
; 18    C z z z z ... z z z z z z 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;       -------------------------------------------------------------
;       C p p p p ... p p p p p p p p p p p p p p p p p p p p p p p p
;
; In the actual routine the PART_PROD buffer contains a single row
; of digits. This product row is added to the previous contents of
; the SIG_R buffer before the next partial product is calculated
;*******************************************************************
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S
        LEA     EDI,SIG_S       ; EDI --> to multiplier
        ADD     EDI,33          ; Point ESI to last multiplier digit
        MOV     QDIG,1          ; QDIG is counter for 18 rows of digits
;**********************|
; clear product buffer |
;**********************|
ONE_ROW_OP:
        LEA     ESI,SIG_L       ; SI --> to multiplicand
        ADD     ESI,33          ; Point SI to end of multiplicand
        LEA     EBX,PART_PROD   ; BX --> buffer for partial products
        MOV     ECX,69          ; Bytes to clear
        PUSH    EBX             ; Save buffer start address
CLEAR_PP:
        MOV     BYTE PTR [EBX],0H     ; Store a zero digit
        INC     EBX              ; Bump buffer pointer
        LOOP    CLEAR_PP
; Offset of rightmost digit in PART_PROD is 69 minus the row
; count. Row count is in QDIG. The offset of the last digit in the
; first partial product is EBX + (69-1) = EBX + 68
        POP     EBX              ; BX --> start of products buffer
        MOV     EAX,69           ; Offset of last buffer digit
        SUB     EAX,QDIG         ; Minus current row
        ADD     EBX,EAX          ; Add offset to pointer
;**********************|
;    multiply row      |
;**********************|
MULT_DIGIT:
        MOV     DL,[EDI]         ; Multiplier digit
        MOV     DH,0             ; Register for carry
        MOV     ECX,34           ; Counter for 34 digits
; At this point:
; ESI --> rightmost digit in the multiplicand
; DL = multiplier digit
; EBX --> last digit offset into partial products buffer
MUL_34_DIGITS:
        MOV     AH,0            ; Clear high digit for product
        MOV     AL,[ESI]     ; Multiplicand digit to AL
        MUL     DL              ; AX = AL * DL
        ADD     AL,DH           ; Add in carry from last product
        AAM                     ; Adjust product to unpacked BCD
        MOV     [EBX],AL     ; Store digit in buffer
        MOV     DH,AH           ; Carry into next digit
        DEC     EBX              ; Bump to next storage position
        DEC     ESI              ; and to next multiplicand digit
        LOOP    MUL_34_DIGITS
; 18 digits multiplied. Store possible carry
        MOV     [EBX],DH     ; Digit to buffer
; At this point PART_PROD holds product of multiplicand times
; current multiplier digit
;**********************|
;      add row         |
;**********************|
; PART_PROD is now added to the accumulated sum stored in SIG_R
; Both buffers have 69 digits
        PUSH    EDI             ; Save pointer to multiplier
        LEA     ESI,PART_PROD   ; Pointer to first digit row
        LEA     EDI,SIG_R       ; Pointer to destination of sum
        ADD     ESI,68          ; Point SI to last digit in buffer
        ADD     EDI,68          ; also DI
        MOV     ECX,69          ; 69 digits to add
        CLC                     ; Clear carry for first addition
ADD_ROW:
        MOV     AL,[ESI]        ; One digit to AL
        ADC     AL,[EDI]        ; Add other digit
        AAA                     ; Convert sum to unpacked BCD
        MOV     [EDI],AL        ; Store digit in same buffer
        DEC     ESI             ; Bump pointers to next digit
        DEC     EDI
        LOOP    ADD_ROW         ; Loop for 69 digits
; Since the buffers contain 69 digits there can be no carry out of
; the last digit
        POP     EDI             ; Restore multiplier pointer
;**********************|
;   multiply next row  |
;**********************|
        DEC     EDI             ; Index to next multiplier
        INC     QDIG            ; Increment row counter
        CMP     QDIG,35         ; Test for all rows multiplied
        JE      DONE_34         ; Go if at end of operation
        JMP     ONE_ROW_OP
; At this point the SIG_R buffer contains the product of the
; significands
;*************************************|
;   Construct BCD20 number and exit   |
;*************************************|
DONE_34:
        CALL    RESTORE_BCD     ; Local procedure used by multiply
                                ; and divide rouitnes
        JMP     SIGN_MUL_EXIT
_SIGN_MUL_BCD20  ENDP


;************************************|
; signed floating point BCD division |
;************************************|
_SIGN_DIV_BCD20  PROC
; Division of two signed BCD numbers stored in BCD20 format
; Operation:
;           z = x / y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          ESI --> Dividend (element x in z = x / y)
;          EDI --> Divisor (element y in z = x / y)
;          EBX --> 20-byte storage area for quotient
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       EBX --> Quotient (element z in z = x / y)
;       carry set if divisor equals zero
;       carry clear if divisor not equal zero
;       Entry pointers are preserved
;*******************************************************************
; BCD signed division algorithm
; A. If the dividend is zero the quotient is zero
;    If the divisor is zero the operation is undefined
;    If neither element is zero then:
; B. If the factors have equal signs the quotient is positive
;    If the factors have unequal sings the quotient is negative
; C. The significand of the quotient is the quotient resulting from
;    dividing the significand of the dividend by the siginificand of
;    the divisor
; D. The exponent of the quotient is the difference between the
;    exponent of the dividend minus the exponent of the divisor
;*******************************************************************
;**********************|
;    clear division    |
;        buffers       |
;**********************|
        PUSH    EBX              ; Save pointer passed by caller
        LEA     EBX,DIVIDEND     ; First of 5 buffers
        MOV     ECX,35           ; bytes to clear
        CALL    CLEAR_BUF        ; Local procedure to clear buffer
        LEA     EBX,DIVISOR      ; Next buffer
        MOV     ECX,34           ; bytes to clear
        CALL    CLEAR_BUF        ; Local procedure to clear buffer
        LEA     EBX,TRIAL_PROD   ; Next buffer
        MOV     ECX,35           ; bytes to clear
        CALL    CLEAR_BUF        ; Local procedure to clear buffer
        LEA     EBX,NEW_DIVIDEND ; Next buffer
        MOV     ECX,35           ; bytes to clear
        CALL    CLEAR_BUF        ; Local procedure to clear buffer
        LEA     EBX,QUOTIENT     ; Last buffer
        MOV     ECX,69           ; bytes to clear
        CALL    CLEAR_BUF        ; Local procedure to clear buffer
        POP     EBX              ; Restore caller's pointer
;**********************|
; save entry variables |
;**********************|
        CALL    STORE_AND_CLEAR         ; Local procedure
;**********************|
; test for dividend = 0|
;**********************|
; If the dividiend = 0 then the quotient = 0
        CALL    TEST_4_ZERO     ; Test routine
        JNC     NOT_ZERO_ID    ; Go if no carry
; At this point x is zero
MOVE_0_QUOT:
        MOV     ECX,20           ; Count for digits to move
MOVE_20_QUOT:
        MOV     BYTE PTR[EBX],00 ; Store zero by BX pointer
        INC     EBX              ; Bump pointer
        LOOP    MOVE_20_QUOT
;**********************************|
;   normal exit from division      |
;**********************************|
SIGN_DIV_EXIT:
        MOV     ESI,ENTRY_ESI  ;Restore entry pointers
        MOV     EDI,ENTRY_EDI
        LEA     EBX,QUOTIENT
		MOV     EBX,ENTRY_EBX
        CLC                     ; Divisor is not zero
		CLD
        RET                     ; End of processing
;**********************|
; test for divisor = 0 |
;**********************|
; Division is undefined if the divisor = 0
; This is signaled to the caller with the carry flag set
NOT_ZERO_ID:
        PUSH    ESI              ; Save x pointer
        MOV     ESI,EDI           ; Pointer to ESI
        CALL    TEST_4_ZERO     ; Test routine
        POP     ESI              ; Restore x pointer
        JNC     NOT_ZERO_ELMS   ; Continue if no carry
;**********************************|
;    error exit from division      |
;**********************************|
; Divisor = zero
        MOV     ESI,ENTRY_ESI  ;Restore entry pointers
        MOV     EDI,ENTRY_EDI
        MOV     EBX,ENTRY_EBX
        STC                     ; Zero divisor. Operation undefined
        RET                     ; End of processing
;**********************|
;  neither x nor y = 0 |
;  subtract exponents  |
;**********************|
NOT_ZERO_ELMS:
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     BIAS_X,DX    ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    ESI,EDI           ; Save x and set y
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     BIAS_Y,DX    ; Store in scratch pad area
        XCHG    ESI,EDI           ; Restore x and y pointers
;**********************|
;  subtract exponents  |
;**********************|
        MOV     AX,BIAS_X    ; Exponent of x to AX
        SUB     AX,DX           ; Subtract exponent of x - y
        ADD     AX,10000        ; Add bias to difference
                                ; to re-scale exponent
        MOV     BIAS_Z,AX    ; Store sum in scratch pad area
;*******************************************************************
; Buffers used in the division of significands:
;        DIVIDEND     (--> EDI)       35 bytes
;        NEW_DIVIDEND                 35 bytes
;        DIVISOR      (--> ESI)       34 bytes
;        TRIAL_PROD                   35 bytes
;        QUOTIENT                     69 bytes
;*******************************************************************
;**********************|
;    unpack and move   |
;      significands    |
;**********************|
; Move 34 packed BCD digits from BCD20 encoding to DIVIDEND buffer
        MOV     ESI,ENTRY_ESI    ; Set pointer to x
        ADD     ESI,3            ; Mantissa digits in BCD12 encoding
        LEA     EDI,DIVIDEND     ; Offset of buffer
        INC     EDI              ; Right-justify dividend in buffer
        MOV     ECX,17           ; 17 packed digits to move
        CALL    UNPACK_MOVE      ; Unpack and move digits routine
; Move significand of y to DIVISOR buffer
        MOV     ESI,ENTRY_EDI    ; Set pointer to y
        ADD     ESI,3            ; Mantissa digits in BCD12 encoding
        LEA     EDI,DIVISOR      ; Offset of CS buffer
        MOV     ECX,17           ; 17 packed digits to move
        CALL    UNPACK_MOVE      ; Unpack and move digits routine
;**********************|
; init quotiend digit  |
;      counter         |
;**********************|
        MOV     QDIG,0           ; Counter for digits
                                 ; in the quotient buffer
; Go to entry point for initial undersize dividend
        JMP     EXPU
;**********************|
; obtain trial divisor |
;**********************|
TEST_EDIVIDE:
        CALL    TRIAL_QUOT       ; Local routine to find a trial
                                 ; quotient digit
        MOV     DL,AL            ; Trial quotient digit to DL
TEST0:
        LEA     ESI,DIVISOR      ; Set pointer to divisor
        LEA     EBX,TRIAL_PROD   ; Pointer to trial product
        ADD     ESI,33           ; To last digit in buffer
        ADD     EBX,34           ; also this pointer
        MOV     DH,0             ; Clear carry register
        MOV     ECX,34           ; Set counter for 34 digits
;**********************|
;  multiply dividend   |
; times trial quotient |
;**********************|
TRI0:
        MOV     AH,0             ; Clear high byte
        MOV     AL,[ESI]         ; Get high order divisor digit
        MUL     DL               ; Divisor times trial quotient
        ADD     AL,DH            ; Add possible carry to product
        AAM                      ; Convert product to BCD
        MOV     DH,AH            ; Save carry in DH
        MOV     [EBX],AL         ; Store digit as trial product
        DEC     EBX              ; Bump pointers
        DEC     ESI
        LOOP    TRI0             ; Continue for 18 digits
; At this point there can be a carry into the 19th digit
; Place possible carry or zero in buffer
        MOV     [EBX],DH         ; Store extra digit
;**********************|
;   subtract product   |
;    from dividend     |
;**********************|
RESTA:
        PUSH    EDX              ; Save registers and pointers
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        LEA     EBX,NEW_DIVIDEND ; Set context pointers
        LEA     ESI,TRIAL_PROD
        LEA     EDI,DIVIDEND
        ADD     EBX,34           ; Index pointers to low order byte
        ADD     ESI,34
        ADD     EDI,34
        MOV     ECX,35           ; Counter for 35 digits
        CLC                      ; Clear carry flag
; Perform subtraction operation
REST0:
        MOV     AL,[EDI]         ; Dividend digit to AL
        MOV     DL,[ESI]         ; Trial product digit to DL
        SBB     AL,DL            ; Dividend minus trial product
        AAS                      ; Convert to BCD
        MOV     [EBX],AL         ; Store result in buffer
        DEC     EBX              ; Bump buffer pointers
        DEC     ESI
        DEC     EDI
        LOOP    REST0           ; Continue for 19 bytes
        POP     EBX              ; Restore entry pointers
        POP     EDI
        POP     ESI
        POP     EDX
; If there is a carry into the buffer's 35th digit then the trial
; quotient is too large
        JNC     PUTQ            ; Go if no carry (quotient digit OK)
;**********************|
;   invalid quotient   |
;        digit         |
;**********************|
; At this point the trial quotient must be adjusted and retried
; If the trial digit goes to zero, then it is stored as a quotient
; digit and processing continues
        DEC     DL              ; Decrement quotient digit
        JZ      PUTQ            ; Store zero
        JMP     TEST0           ; Retry new trial quotiend digit
;**********************|
; valid quotient digit |
;**********************|
; Processing:
; 1. Store quotient digit in buffer (DIGIT_TO_QUOT procedure)
; 2. Make the trial dividend (in NEW_DIVIDEND) the dividend
PUTQ:
        CALL    DIGIT_TO_QUOT   ; Store digit in buffer
        JC      SIG_DONE        ; Go if buffer full returned
; Move NEW_DIVIDEND to DIVIDEND multiplying by 10
        LEA     EDI,DIVIDEND  ; Pointer to dividend buffer
        LEA     ESI,NEW_DIVIDEND      ; Pointer to new dividend
        MOV     BYTE PTR [ESI],0H     ; Clear high byte
        INC     ESI
        MOV     ECX,34          ; Bytes to move from ESI to EDI
        CALL    TRANSFER_CS_CX  ; Local procedure
; New dividend is now in NEW_DIVIDEND buffer
; Chech size of dividend. Place zeros in buffer is the dividend
; is smaller than the divisor
EXPU:
        CALL    COMPARE_EDI_DR   ; Compare divisor and dividend
        JC      EXPU1           ; Go if divisor > dividend
        JMP     TEST_EDIVIDE     ; Continue processing
;**********************|
;   quotient = 0       |
;**********************|
EXPU1:
        MOV     DL,0             ; Digit to store
        CALL    DIGIT_TO_QUOT    ; Local procedure
        JC      SIG_DONE         ; Exit routine if carry set
; Multiply dividend by 10 by offsetting into buffer
        LEA     EDI,DIVIDEND     ; EDI points to dividend area
        MOV     ESI,EDI          ; and so does ESI
        INC     ESI              ; Skip first digit (0) in source
                                 ; effectively multiplying by 10
        MOV     ECX,34           ; Bytes to move from ESI to EDI
        CALL    TRANSFER_CS_CX   ; Local procedure
        JMP     EXPU
;**********************|
;  move significand    |
;**********************|
; Significand digits from QUOTIENT buffer to SIG_R buffer
SIG_DONE:
        LEA     ESI,QUOTIENT     ; Source buffer
        LEA     EDI,SIG_R        ; Destination buffer
        INC     EDI              ; Skip first digit
        MOV     CX,68            ; Digits to move
        CALL    TRANSFER_CS_CX   ; Local procedure
        CALL    RESTORE_BCD      ; Local procedure used by multiply
                                 ; and divide rouitnes
        JMP     SIGN_DIV_EXIT
_SIGN_DIV_BCD20  ENDP

;
;*******************************************************************
;                     auxiliary procedures
;*******************************************************************{
; BIN_TO_BCD
; Convert a 16-bit binary number to 5 BCD digits
; On entry:
;         AX = binary number
; On exit:
;         5 packed BCD digits in the following registers:
;                CL = digit number 1 (high order digit)
;                DH = digits number 2 and 3
;                DL = digits number 4 and 5 (low order BCD digit)
;
; Operation:
; Binary in AX is first converted into 5 unpacked BCD digits and
; stored in a temporary buffer (BCD_BUF). Digits are then moved
; from this buffer and packed into DX and CL
;
;
BIN_TO_BCD      PROC 
;**********************|
;    clear buffer      |
;**********************|
        LEA     EDI,BCD_BUF     ; Set pointer to buffer
        PUSH    EDI             ; Save buffer start
        MOV     ECX,5           ; Five digits to clear
CLEAR_5BCD:
        MOV     BYTE PTR [EDI],0         ; Clear digit
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_5BCD      ; Repeat for 5 digits
        DEC     EDI             ; Adjust buffer pointer to last
                                ; digit
        MOV     ECX,10          ; Decimal divisor to CX
GET_BIN:
        MOV     DX,0            ; Clear for word division
; Divide binary by 10
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
        MOV     [EDI],DL     ; Store digit in buffer
        DEC     EDI             ; Bump destination pointer
; Note: step 5 is performed automatically since the binary
; quotient is left in AX
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_BIN         ; Continue if not end
; Last digit converted
;**********************|
;   BCD to registers   |
;      and pack        |
;**********************|
        POP     EDI             ; Restore pointer to start of
                                ; buffer
        ADD     EDI,4           ; Point to last digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     BL,AL           ; To results register low nibble
        DEC     EDI             ; Point to fourth BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BL,AL           ; OR into result register high
                                ; nibble
        DEC     EDI             ; Pointer to third BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     BH,AL           ; To results register low nibble
        DEC     EDI             ; Point to second BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BH,AL           ; OR into result register high
                                ; nibble
; Move 4 low order BCD digits to DX
        MOV     DX,BX
; Move high order digit to CL
        DEC     EDI             ; Pointer to first BCD digit
        MOV     AL,[EDI]        ; BCD digit to AL
        MOV     CL,AL           ; To results register low nibble
        RET
BIN_TO_BCD      ENDP


; Convert a 5-digit BCD number into a binary in DX
; On entry:
;         AL-BH-BL-CH-CL hold 5 BCD digits
;                        holding a value < 65536
; On exit:
;         DX = binary number (range 0 to 65535)
;
BCD_INTO_REG    PROC
; Set segment for conversion procedure
        LEA     ESI,ASC_TEMP_BUF ; Address of buffer off DS
        CALL    BCD_TO_ASCII  ; Convert BCD digits to ASCII
                                ; string
        CALL    ASCII_TO_DX     ; Conversion routine
        RET
BCD_INTO_REG    ENDP

BCD_TO_ASCII    PROC
; Convert 5 unpacked BCD digits in machine registers into to 5
; ASCII digits
; On entry:
;         AL-BH-BL-CH-CL hold 5 unpacked BCD digits
;                        holding a value < 65536
;         ESI --> 5 byte buffer to hold ASCII string
; On exit:
;         ESI --> buffer holding ASCII digits
;
;
; Clear buffer for ASCII digits
        PUSH    ESI              ; Save start of buffer
        PUSH    ESI              ; Twice
        PUSH    ECX              ; Save BCD high digit
        MOV     ECX,5            ; Counter for 5 digits
CLEAR_BUF_5:
        MOV     BYTE PTR [ESI],' '    ; Set blank in buffer
        INC     ESI              ; Bump pointer
        LOOP    CLEAR_BUF_5
        POP     ECX              ; Restore BCD digit
;**********************|
; convert to ASCII and |
;    move to buffer    |
;**********************|
; Convert BCD digits to ASCII
        ADD     AL,30H          ; Add 30H to each digit
        ADD     BH,30H
        ADD     BL,30H
        ADD     CH,30H
        ADD     CL,30H
        POP     ESI              ; Restore buffer start
        MOV     [ESI],AL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     [ESI],BH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     [ESI],BL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     [ESI],CH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],CL     ; Digit to buffer
        POP     ESI              ; Restore start of buffer
        RET
BCD_TO_ASCII    ENDP


ASCII_TO_DX     PROC
; Convert an ASCII number to binary and store in DX register
; On entry:
;         ESI --> start of 5-digit ASCII buffer holding decimal
;                string in the range 0 to 65,535
; On exit:
;         DX = binary number
;
;
;***********************|
;  initialize registers |
;***********************|
        MOV     DX,0            ; Clear accumulator register
        MOV     ECX,10           ; Load multiplier into CX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC16:
        MOV     AL,[ESI]         ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC16      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC16      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    AX              ; Save digit in stack
        MOV     AX,DX           ; Previous product to AX
        MUL     CX              ; Perform AX = AX * 10
        MOV     DX,AX           ; Move product to accumulator
        POP     AX              ; Restore decimal digit
        ADD     DX,AX           ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     ESI              ; Bump pointer
        JMP     DIGIT_TO_ACC16  ; Continue
EXIT_ASC16:
        RET
ASCII_TO_DX     ENDP

RESTORE_BCD     PROC    
; Reconstruct BCD20 number after multiplication and division operation
; Note that the sign rules for significand and exponents are the
; same
; On entry:
;         BIAS_Z holds biased exponent of result
;         SIG_R buffer holds significand
; On exit:
;         Buffer at ENTRY_EBX contains results in BCD12 format
;
        CALL    ROUND_SIGF      ; Round off significand
        MOV     EDI,ENTRY_EBX  ; Set pointer to result buffer
                                ; for BCD12 output
        MOV     BYTE PTR [EDI],00H       ; Set default positive sign
        ADD     EDI,3            ; Point to significand area
;**********************|
;   store significand  |
;**********************|
; Significand of result is unpacked BCD form in the first 18 digits
; of the SIG_R buffer
; EDI --> significand field of the BCD12 result buffer
        MOV     ECX,17         ; Counter for 34 digits
        LEA     ESI,SIG_R      ; Pointer to unpacked digits
        INC     ESI            ; Skip leading zero
; Skip possible leftmost zero in buffer
        CMP     BYTE PTR [ESI],00H    ; Test for zero
        JNE     PACK_EDIGITS   ; Go if no leading zero
        INC     ESI            ; Skip zero
        DEC     BIAS_Z         ; Adjust exponent
; In multiplication there can be only one leading zero in the buffer
; because the input significands are in normalized form
PACK_EDIGITS:
        MOV     AH,BYTE PTR [ESI]     ; Get one digit
        INC     ESI            ; Bump pointer
        MOV     AL,BYTE PTR [ESI]     ; Get next digit
        INC     ESI            ; Bump pointer to next digit
        SHL     AH,4          ; Shift high digit to high nibble
        OR      AL,AH          ; Merge high nibble in AH with low
                               ; nibble in AH (pack digits)
        MOV     BYTE PTR [EDI],AL        ; Store packed digits
        INC     EDI
        LOOP    PACK_EDIGITS
; All digits moved to the destination BCD12 buffer
        MOV     EBX,ENTRY_EBX  ; Set pointer to result buffer
                                ; sign area
;**********************|
;   sign of number     |
;**********************|
; Rule: if the factors have equal signs the product is positive
;       If the factors have unequal sings the product is negative
; The sign of the number is the high nibble of the first BCD12 digit
        MOV     ESI,ENTRY_ESI  ; Set pointers to entry values
        MOV     EDI,ENTRY_EDI
; ESI --> x and EDI --> y
        MOV     AL,BYTE PTR [ESI]        ; x's first digit to AL
        MOV     AH,BYTE PTR [EDI]        ; y's first digit to AH
        AND     AL,11110000B    ; Clear x's low nibble
        AND     AH,11110000B    ; and y's low nibble
        CMP     AL,AH           ; Test for equal sign
        JE      EXPN_SIGN       ; Numbers have equal sign
                                ; This encoding already in buffer
; At this point the factors have unequal sign. Product is negative
        MOV     [EBX],10H       ; Negative sign for number
;**********************|
; get sign of exponent |
;**********************|
EXPN_SIGN:
; Exponent of result is stored in BIAS_Z in bias 10000 form
; Values < 10000 indicate a negative exponent
        MOV     DX,BIAS_Z    ; Exponent of result to DX
        CMP     DX,10000        ; Compare with bias
        JAE     POS_VALUE       ; Go if positive exponent
; Positive exponent is already encoded in result buffer
; Encode negative exponent (0001B) into BCD12 number
; EDI --> sign field
        MOV     AL,[EBX]        ; Get sign digit into AL
        MOV     AH,00000001B    ; Code for negative exponent
        OR      AL,AH           ; OR exponent sign into byte
        MOV     BYTE PTR [EBX],AL        ; Store in buffer
; Calculate absolute value of negative exponent
        MOV     AX,10000        ; Bias to AX
        SUB     AX,DX           ; Bias minus exponent
        XCHG    AX,DX           ; Move exponent to DX
        JMP     CALCULATE_VALUE ; Go to conversion routine
;**********************|
;  calculate value of  |
;  positive exponent   |
;**********************|
; DX = exponent in bias 10000 form
POS_VALUE:
        MOV     AX,10000        ; Bias to AX
        SUB     DX,AX           ; Exponent minus bias
CALCULATE_VALUE:
; DX = absolute value of exponent (in binary)
        MOV     AX,DX           ; For conversion routine
        PUSH    EBX             ; Save pointer for result
        CALL    BIN_TO_BCD      ; Conversion routine
        POP     EBX             ; Restore pointer
; DH and DL holds 4 packed BCD digits
        INC     EBX              ; Bump pointer to exponent filed
        MOV     [EBX],DH        ; Store two digits
        INC     EBX
        MOV     [EBX],DL        ; Store two other digits
        RET
RESTORE_BCD     ENDP
;*******************************************************************
;
TRIAL_QUOT      PROC    
; Obtain largers possible trial quotient for division
; On entry:
;          nothing
; On exit:
;          AL = trial quotient
;
        LEA     EBX,DIVIDEND  ; Pointer to dividend buffer
        MOV     AL,[EBX]      ; First dividend digit to AL
        INC     EBX              ; Pointer to next dividend digit
        MOV     CH,[EBX]      ; Second dividend digit to CH
        MOV     DL,DIVISOR   ; High order divisor digit to DL
; Convert dividend digits (in AL-CH) to binary by multiplying the
; first digit by 10 and adding the second digit
        MOV     CL,10           ; Multiplier
        MUL     CL              ; First digit (in AL) times 10
        ADD     AL,CH           ; Add in second digit (in CH)
; AL now holds the binary dividend
; Trial quotient is obtained by dividing by the high order divisor
; digit (in DL)
        MOV     AH,0            ; Clear high nibble
        DIV     DL              ; Dividend / divisor = quotient
; The quotient must be a single digit. Test for a trial quotient
; larger than 9 and adjust
        CMP     AL,9            ; Test for limit value
        JA      ADJUST_EDIGIT    ; Go is larger than 9
; At this point there is a valid trial quotient digit in AL
        RET
ADJUST_EDIGIT:
        MOV     AL,9            ; Use largest single digit
        RET
TRIAL_QUOT      ENDP
;*******************************************************************
;
TRANSFER_CS_CX  PROC    
; Transfer CX byte from a code segment buffer (by ESI) to another
; one (by EDI)
; On entry:
;         ESI --> source buffer for move
;         EDI --> destination buffer for move
;         ECX = number of bytes to move
; On exit:
;         EDI --> moved string
;         all registers preserved
;
        PUSH    ESI              ; Save entry registers
        PUSH    EDI
        PUSH    ECX
MOVE_TO_CS:
        MOV     AL,[ESI]      ; Get byte from source buffer
        MOV     [EDI],AL      ; and place it in the destination
        INC     ESI              ; Bump buffer pointers
        INC     EDI
        LOOP    MOVE_TO_CS
        POP     ECX              ; Restore caller's registers
        POP     EDI
        POP     ESI
        RET
TRANSFER_CS_CX  ENDP
;*******************************************************************
;
DIGIT_TO_QUOT   PROC    
; Place digit (in DL) in QUOTIENT buffer (if not full)
; On entry:
;         DL = quotient digit to store
; On exit:
;         carry set if QUOTIENT buffer full
;
        PUSH    EBX               ; Save entry registers
        PUSH    EAX
        LEA     EBX,QUOTIENT      ; Pointer to destination buffer
        ADD     EBX,QDIG          ; Digit counter
; Test for end of QUOTIENT buffer
        CMP     BYTE PTR [EBX],0FFH   ; Marker for end of buffer
        JE      QUOT_BUF_END      ; Go if at end
; Place digit in buffer and index pointer
        MOV     [EBX],DL          ; Store digit
        INC     QDIG              ; Bump offset counter
;**********************|
;    normal exit       |
;**********************|
        POP     EAX              ; Restore registers
        POP     EBX
        CLC                     ; Signal buffer not full
        RET
;**********************|
;  buffer full exit    |
;**********************|
QUOT_BUF_END:
        POP     EAX              ; Restore entry registers
        POP     EBX
        STC                     ; Signal buffer full to caller
        RET
DIGIT_TO_QUOT   ENDP
;*******************************************************************
;
COMPARE_EDI_DR   PROC    
; Compare dividend to divisor
; On entry:
;          nothing
; On exit:
;          carry set if divisor > dividend
;          carry clear if divisor < dividend
;
        LEA     EDI,DIVIDEND  ; EDI --> dividend
        LEA     ESI,DIVISOR   ; ESI --> divisor
; If first digit of divided is not zero then it must be larger than
; the divisor
        CMP     BYTE PTR [EDI],00     ; Test for zero digit
        JNE     LARGER_DIVIDEND         ; Dividend is larger
        INC     EDI              ; Bump pointer to next digit
COMP_EDI_DR:
        MOV     AL,[ESI]      ; Divisor digit
        CMP     AL,0FFH         ; Test for end of divisor
        JE      LARGER_DIVIDEND ; Dividend is larger exit
        MOV     AH,[EDI]      ; Get dividend digit to AH
; Compare dividend and divisor digits
        CMP     AH,AL           ; Operation is dividend - divisor
        JB      LARGER_DIVISOR  ; Go if divisor > dividend
        JA      LARGER_DIVIDEND ; or if divisor < dividend
; At this point divisor = dividend. Continue comparing digits
        INC     ESI              ; Bump pointers
        INC     EDI
        JMP     COMP_EDI_DR
;**********************|
;  dividend > divisor  |
;**********************|
LARGER_DIVIDEND:
        CLC                     ; Signal to caller
        RET
;**********************|
;  divisor > dividend  |
; or divisor = dividend|
;**********************|
LARGER_DIVISOR:
        STC
        RET
COMPARE_EDI_DR   ENDP
;*******************************************************************
;
CLEAR_BUF    PROC    
; Clear CX number of bytes in code segment buffer
; On entry:
;         EBX --> buffer area
;         ECX = number of bytes to clear
CLEAR_BYCS:
        MOV     BYTE PTR [EBX],00H    ; Clear one byte
        INC     EBX                      ; Bump pointer
        LOOP    CLEAR_BYCS
        RET
CLEAR_BUF    ENDP
;*******************************************************************
;
STORE_AND_CLEAR         PROC    
; Store entry pointers in scratchpad area and clear work buffers
;**********************|
; store entry pointers |
;**********************|
        MOV     ENTRY_ESI,ESI     ; Store ESI pointer
        MOV     ENTRY_EDI,EDI     ; Store EDI pointer
        MOV     ENTRY_EBX,EBX     ; Store BX pointer
;**********************|
;  clear work buffers  |
;**********************|
        PUSH    ESI               ; Save x value entro pointer
        LEA     ESI,SIG_L         ; Pointer to buffer
        MOV     ECX,205           ; Clear three buffers
CLEAR_205:
        MOV     BYTE PTR [ESI],00H    ; Clear one byte in buffer
        INC     ESI               ; Bump pointer
        LOOP    CLEAR_205
        POP     ESI               ; Restore x pointer
        RET
STORE_AND_CLEAR         ENDP
;*******************************************************************
;

ROUND_SIGF      PROC    
; Rounding of significand digits stored in SIG_R buffer
; Rationale:
; SIG_R holds 69 digits in unpacked BCD format. The BCD20 format
; encodes 34 siginificand digits
; Rounding proceeds with the 35th digit in SIG_R, as follows:
; 1. If the 35th digit is < 5 then no rounding is required
; 2. If the 35th digit is > 5 then the first 18 digits are rounded
; 3. If the 35th digit = 5 then the digits to the right are examined.
;    If the digits to the right are all zeros (5000..00) then the
;    18th digit is incremented only if it is odd. This rounding
;    method is known as "round to nearest even".
; 4. If the result of incrementing the 18th digit is 10, the digit
;    to the left is also incremented, until no carry is generated.
;
        LEA     ESI,SIG_R     ; Set pointer to buffer in CS
        ADD     ESI,34        ; Index to the 19th digit
        MOV     AL,[ESI]      ; 19th digit to AL
        CMP     AL,05H        ; Test for 5 or greater
        JB      EXIT_ROUNDING ; No rounding necessary
; 19th digit is 5 or greater
        JE      MIDWAY_CASE   ; Go to "round to even" routine
; At this point the 19th digit is > 5
;**********************|
;  bump digit and test |
;**********************|
ROUND_OFF:
        DEC     ESI           ; Point to 18th digit
ROUND_EVEN:
        MOV     AL,[ESI]      ; Get BCD digit into AL
        ADD     AL,01H        ; Add 1 to digit
        AAA                   ; Adjust result
        JNC     EXIT_ROUNDING ; Go if no carry generated
        JMP     ROUND_OFF
; Note that no testing for the end of buffer is required since the
; leftmost digit in SIG_R cannot be greater than 8
;**********************|
; test for midway case |
;**********************|
; ESI --> to the 35th digit
MIDWAY_CASE:
        PUSH    ESI              ; Save pointer to 35th digit
        MOV     ECX,33           ; Set counter for 33 digits
TEST_FOR_0H:
        INC     ESI              ; Bump pointer to next digit
        CMP     BYTE PTR [ESI],00H    ; Test for a zero digit
        JNE     NOT_MIDWAY      ; Exit if not midway case
        LOOP    TEST_FOR_0H     ; Continue testing
; At this point the undisplayed digits are in the pattern 5000..00
        POP     ESI              ; Restore pointer to 19th digit
        DEC     ESI              ; Point to 18th digit
        MOV     AL,[ESI]      ; Get 18th digit into AL
;**********************|
;  test for odd 36th   |
;        digit         |
;**********************|
; Binary encodings of odd numbers have the LOB set
        TEST    AL,00000001B    ; Test LOB to determine if odd
        JNZ     ROUND_EVEN      ; Go to rounding routine if odd
        JMP     EXIT_ROUNDING   ; Do not round if already even
;**********************|
;    not midway case   |
;**********************|
; The number is rounded if the undiplayed significand is not in the
; form 50000..00
NOT_MIDWAY:
        POP     ESI              ; Restore pointer
        JMP     ROUND_OFF       ; Go to rounding routine
;**********************|
;   exit procedure     |
;**********************|
EXIT_ROUNDING:
        CLD
        RET
ROUND_SIGF      ENDP
;*******************************************************************
;
UNPACK_MOVE     PROC    
; Move the number of packed digits in CX from ESI to EDI 
; and unpack
; On entry:
;         ECX = number of packed BCD digits to be moved
;         ESI --> packed BCD string to move
;         EDI --> destination of unpacked string
; On exit:
;         EDI = unpacked string of BCD digits
;
        PUSH    ESI
        PUSH    EDI
; Move CX digits
MOVE_EDIGS:
        MOV     AL,[ESI]        ; Get one byte from source
; AL has two packed BCD digits
        MOV     AH,AL           ; Copy packed digits to AH
        AND     AL,00001111B    ; Clear high nibble to isolate low
                                ; BCD digit
        PUSH    ECX             ; Save digits counter
        MOV     CL,4            ; Counter for 4 bits
        SHR     AH,CL           ; Shift right to isolate high BCD
                                ; digit
        POP     ECX             ; Restore counter
; AH holds the unpacked high digit and AL holds the low digit
        MOV     [EDI],AH        ; High BCD digit to buffer
        INC     EDI             ; Bump buffer pointer
        MOV     [EDI],AL        ; Low BCD digit to buffer
        INC     EDI             ; Bump pointer again
        INC     ESI             ; Bump source pointer
        LOOP    MOVE_EDIGS
        POP     EDI
        POP     ESI
        RET
UNPACK_MOVE     ENDP
;*******************************************************************
;
TEST_4_ZERO     PROC    
; Test a number encoded in BCD20 format for a zero value
; Note: a number in BCD20 format is zero if all numbers in the
;       exponent and significand are zero
; On entry:
;         ESI --> Number stored in BCD20 format
; On exit:
;         Carry set if number is zero
;         Carry clear if number is not zero
;
        PUSH    ESI              ; Save entry pointer
        INC     ESI              ; Point to first exponent digit
        MOV     CX,19            ; Counter for 19 bytes to test
TEST_11_ZERO:
        CMP     BYTE PTR[ESI],0  ; Is the byte zero
        JNE     NOT_11_ZERO     ; Go if not zero
        INC     ESI              ; Bump pointer
        LOOP    TEST_11_ZERO    ; Test all digits
; At this point the number is zero
        POP     ESI              ; Restore entry pointer
        STC                     ; Set zero return signal
        RET
NOT_11_ZERO:
        POP     ESI              ; Restore entry pointer
        CLC                     ; Set non-zero return signal
        RET
TEST_4_ZERO     ENDP
;*******************************************************************
;
EXP_2_BIAS      PROC    
; Convert exponent of BCD20 number into binary and add bias
; On entry:
;         ESI --> Number encoded in BCD12 format
; On exit:
;         DX = binary value of exponent
;
        PUSH    EBX              ; Save pointers
        PUSH    ESI
        PUSH    EDI
        INC     ESI              ; Pointer to exponent field
        INC     ESI              ; Second exponent digit
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [ESI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        MOV     CX,4            ; For a count of 4
        SHR     DH,CL           ; Shift digit to low position
        DEC     ESI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [ESI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        MOV     CX,4            ; For a count of 4
        SHR     BH,CL           ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        CALL    BCD_INTO_REG    ; BCD to binary conversion
        POP     EDI             ; Restore entry pointers
        POP     ESI
        POP     EBX
; At this point DX hold binary value for exponent
; ESI --> sign byte
; Test for negative exponent. Add bias if positive
        TEST    BYTE PTR[ESI],01H        ; Test for negative exponent
        JNZ     EXP_IS_NEG
; At this point exponent is positive
        ADD     DX,10000        ; Add bias to exponent value
        RET                     ; End of processing
; Subtract exponent from bias
EXP_IS_NEG:
        MOV     AX,10000        ; Bias
        SUB     AX,DX           ; Value minus bias
        MOV     DX,AX           ; Result to DX
        RET                     ; End of processing
EXP_2_BIAS      ENDP


	end
