;*****************************************************************
;*****************************************************************
;                           UN32_6.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 14, 2001
; Last update:    
;
; Libary name: MATH32.LIB
; This module: UN32_6.ASM
; Operations : Range scaling and calculation of trigonometric
;              functions
;
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC  _DEG_TO_RAD      ; .. degrees to radian conversion
PUBLIC  _RAD_TO_DEG      ; .. radians to degrees conversion
PUBLIC  _CIRCLE_SCALE    ; .. scale angle to unit circle
PUBLIC  _OCTANT_SCALE    ; .. scale angle to first octant
PUBLIC  _TANGENT         ; .. calculate tangent function
PUBLIC  _SINE            ; .. idem for sine
PUBLIC  _COSINE          ; .. idem for cosine
PUBLIC  _COTANGENT       ; .. idem for cotangent
PUBLIC  _SECANT          ; .. idem for secant
PUBLIC  _COSECANT        ; .. idem for cosecant
PUBLIC  _ARC_SINE        ; .. calculate arcsine function
PUBLIC  _ARC_COSINE      ; .. idem for arccosine
PUBLIC  _ARC_TANGENT     ; .. idem for arctangent
PUBLIC  _SINH            ; .. hyperbolic sine
PUBLIC  _COSH            ; .. hyperbolic cosine
PUBLIC  _TANH            ; .. hyperbolic tangent
PUBLIC  _ARC_SINH        ; .. hyperbolic arcsine
PUBLIC  _ARC_COSH        ; .. hyperbolic arcosin
PUBLIC  _ARC_TANH        ; .. hyperbolic arctangent
;
;****************************************************************
;               definitions for 32-bit flat model
;****************************************************************
        .486
        .MODEL flat
        .DATA

;***************************|
;   Data for this module    |
;***************************|
SIGN            DB      0       ; Storage for sign of input
STATUS_WW       DW      0       ; Storage for status word
REMAINDER       DB      1       ; Storage for zero remainder
                                ; of r MOD(pi/4)
; Constants are defined in extended precision format
ONE_80          DT      180.0   ; 180
; Definition of constant e
E               DT      4000ADF85458A2BB4A9BH
; Constant defined in single precision format
ONE_HALF        DD      0.5     ; 1/2 in single precision
; Storage for math unit controls
CONTROL_WW      DW      0       ; Control word
ROUND_DOWN      DW      177FH   ; Control word for round down
                                ; operations

;****************************************************************
;                           code
;****************************************************************
       .CODE

;************************************
;        degrees to radian
;************************************
_DEG_TO_RAD      PROC
; Convert entry value in ST(0) to radian measure
;
; On entry:
;         ST(0) = value in degrees
; On exit:
;         ST(0) = value converted to radians
;
; Formula:
;         r = d * (pi/180)
;
; where r = angle in radian, d = angle in degrees
;

;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      d      |  --------  |  -------- |
        FLD     ONE_80
                        ;     180     |     d      |
        FLDPI           ;    pi       |    180     |     d     |
        FDIV    ST,ST(1);  pi/180     |    180     |     d     |
        FSTP    ST(1)   ;  pi/180     |     d      |
        FMUL    ST,ST(1);     r       |     d      |
        FSTP    ST(1)   ;     r       |  --------- |
; ST(0) now holds entry value in radian
        CLD 
		RET
_DEG_TO_RAD      ENDP


;************************************
;        radian to degrees
;************************************
_RAD_TO_DEG      PROC
; Convert entry value in ST(0) to degrees
;
; On entry:
;         ST(0) = value in radian
; On exit:
;         ST(0) = value converted to degrees
;
; Formula:
;         d = r * (180/pi)
;
; where r = angle in radian, d = angle in degrees
;
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      d      | ---------- | --------- |
        FLDPI           ;      r      |    pi      | --------- |
        FLD     ONE_80
                        ;     180     |     pi     |     r     |
        FDIV    ST,ST(1);  180/pi     |     pi     |     r     |
        FSTP    ST(1)   ;  180/pi     |     r      | --------- |
        FMUL    ST,ST(1);     d       |     r      | --------- |
        FSTP    ST(1)   ;     d       |  --------  |
; ST(0) now holds entry value in degrees
        CLD
		RET
_RAD_TO_DEG      ENDP


_CIRCLE_SCALE      PROC
; Reduce angle in ST(0) (r) to a value less than 2pi radian
; by calculating  n = r (MOD 2pi)
; The code assumes that a word-size variable, named STATUS_WW
; has been previously defined in the code segment.
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      |  --------- | --------- |
        FLDPI           ;     pi      |     r      | --------- |
        FADD    ST,ST(0);     2pi     |     r      | --------- |
        FXCH            ;      r      |     2pi    | --------- |
UNIT_CIRC:
        FPREM           ; REM(r/2pi)  |     2pi    | --------- |
        FSTSW   STATUS_WW    ; Store status bits in memory
        FWAIT                   ; Delay until end of store
        MOV     AH,BYTE PTR STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,00000100B    ; Mask off all except C2
        JNZ     UNIT_CIRC       ; Repeat if C2 not clear
;                       ;    ST(0)    |    ST(1)   |    ST(2)
        FSTP    ST(1)   ;      n      | ---------- | --------- |
; ST(0) now holds an angle in the unit circle (n)
        CLD
		RET
_CIRCLE_SCALE    ENDP


_OCTANT_SCALE    PROC
; Scale an input angle (in radian measure) to the first octant
; and report the octant of the original angle
; On entry:
;         ST(0) = positive angle in the unit circle
; On exit:
;         ST(0) = input scaled to pi/4 radian (first octant)
;         ST(1) = pi/4
;         AH = condition codes C3, C1, and C0 in their original
;              bit positions
;
;***************************|
;   reduce to unit circle   |
;***************************|
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- | --------- |
        FLDPI           ;     pi      |     r      | --------- |
        FADD    ST,ST(0);     2pi     |     r      | --------- |
        FXCH            ;      r      |     2pi    | --------- |
SCALE_TO_UNIT:
        FPREM           ; r = MOD(2pi)|     2pi    | --------- |
        FSTSW   STATUS_WW    ; Store status bits in memory
        FWAIT                   ; Delay until end of store
        MOV     AH,BYTE PTR STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,00000100B    ; Mask off all except C2
        JNZ     SCALE_TO_UNIT   ; Repeat if C2 not clear
;                       ;    ST(0)    |    ST(1)   |    ST(2)
        FSTP    ST(1)   ;  r (scaled  | ---------- | --------- |
;                       |   to 2pi)   |            |           |
; ST(0) now holds an angle in the unit circle
;***************************|
;      reduce to pi/4       |
;***************************|
        FCLEX
        FLD1                    ;     1    |     r    | ------- |
        FADD    ST(0),ST        ;     2    |     r    | ------- |
        FCHS                    ;    -2    |     r    | ------- |
        FLDPI                   ;    pi    |    -2    |     r   |
        FSCALE                  ;   pi/4   |    -2    |     r   |
        FSTP    ST(1)           ;   pi/4   |     r    | ------- |
        FXCH                    ;     r    |    pi/4  | ------- |
        FPREM                   ; MOD(pi/4)|   pi/4   | ------- |
        FSTSW   STATUS_WW       ; storage for status word
; At this point ST(0) contains an angle in the first octant
; The FPREM instruction will produce a valid remainder in one
; execution if the input is < 2pi radian
        MOV     AH,BYTE PTR STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,01000011B    ; Mask off unused bits
;                   |    ||_______ C0
;                   |    |________ C1
;                   |_____________ C3
; FPU stack on exit
;                               |r (scaled |   pi/4   | ------- |
;                               | to pi/4) |          |
;
        CLD
        RET
_OCTANT_SCALE    ENDP

_TANGENT     PROC
; Compute tangent of an angle in radians in the 80387 or the math
; unit of the 486 and Pentium. (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = tangent function
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    _CIRCLE_SCALE
        FPTAN                   ;    1    | tan r    | -------- |
        FSTP    ST(0)           ; tan r   | -------- |
        CLD
		RET
_TANGENT     ENDP


_SINE       PROC
; Compute sine of an angle in radians in the 80387 or the math
; unit of the 486 and Pentium. (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = sine function
;
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    _CIRCLE_SCALE
        FSIN                    ;  sin r  | -------- | -------- |
        CLD
		RET
_SINE        ENDP

_COSINE     PROC
; Compute sine of an angle in radians in the 80387 or the math
; unit of the 486 and Pentium. (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = sine function
;
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    _CIRCLE_SCALE
        FCOS                    ;  cos r  | -------- | -------- |
        CLD
		RET
_COSINE      ENDP



_COTANGENT       PROC
; Compute cotangent of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = cotangent function
;
; Processing is based on the identity
;                 cot (r) = 1 / tan(r)
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        FPTAN                   ;    1    | tan r    | -------- |
        FDIV    ST,ST(1)        ;  1/tan  |   tan r  | -------- |
        FSTP    ST(1)           ;  cot    | -------- |
        CLD
		RET
;
_COTANGENT       ENDP

_SECANT          PROC
; Compute secant of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = secant function
;
; Processing is based on the identity
;                 sec (r) = 1 / sin(r)
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        FSIN                    ;   sin r | -------- |----------|
		FLD1                    ;    1    |   sinr   |----------|
        FDIV    ST,ST(1)        ;  1/sin  |   sin r  | -------- |
        FSTP    ST(1)           ;  sec    | -------- |
        CLD
		RET
;
_SECANT          ENDP

_COSECANT        PROC
; Compute secant of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = secant function
;
; Processing is based on the identity
;                cosc (r) = 1 / cos(r)
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        FCOS                    ;  cos r  | -------- | ---------|
		FLD1                    ;    1    |   cos r  | ---------|
        FDIV    ST,ST(1)        ;  1/cos  |   cos r  | -------- |
        FSTP    ST(1)           ;  cosc   | -------- |
        CLD
		RET
;
_COSECANT        ENDP


;*********************************|
;           arc sine              |
;*********************************|
_ARC_SINE        PROC
; Find the angle (in radian) whose sine is in ST(0)
; On entry:
;         ST(0) = sine of an angle (in radian)
;                 Must be a normal number in the range 0 < ST < 1
;                 this corresponds to an angle in the first
;                 quadrant
; On exit:
;         ST(0) = arc sine of entry value in ST(0)
;                 expressed in radian
; Formula:
;         arcsin (r) = arctangent (r / SQR(1 - r^2))
;
; Legend: x2 = x**2
;         y2 =  y**2
;         SQR = square root
;
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- | --------- |
        FLD     ST(0)   ;      r      |     r      | --------- |
        FMUL    ST,ST(1);     r^2     |     r      | --------- |
        FLD1            ;      1      |    r^2     |     r     |
        FSUB    ST,ST(1);    1-r^2    |    r^2     |     r     |
        FSQRT           ; SQR(1-r^2)  |    r^2     |     r     |
        FSTP    ST(1)   ; SQR(1-r^2)  |     r      | --------- |
; FPATAN calculates arctangent ST(1)/ST
        FPATAN          ; arcsin (r)  | ---------- |
        CLD
		RET
_ARC_SINE        ENDP


;*********************************|
;           arc cosine            |
;*********************************|
_ARC_COSINE      PROC
; Find the angle (in radian) whose cosine is in ST(0)
; On entry:
;         ST(0) = cosine of an angle (in radian)
;                 Must be a normal number in the range 0 < ST < 1
;                 this corresponds to an angle in the first
;                 quadrant (0 < angle < pi/2)
; On exit:
;         ST(0) = arc cosine of entry value in ST(0)
; Formula:
;         arccos (r) = arctangent (SQR (1 - r^2) / r)
;
; Legend: x2 = x**2
;         y2 =  y**2
;         SQR = square root
;
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- | --------- |
        FLD     ST(0)   ;      r      |     r      | --------- |
        FMUL    ST,ST(1);     r^2     |     r      | --------- |
        FLD1            ;      1      |    r^2     |     r     |
        FSUB    ST,ST(1);    1-r^2    |    r^2     |     r     |
        FSQRT           ; SQR(1-r^2)  |    r^2     |     r     |
        FSTP    ST(1)   ; SQR(1-r^2)  |     r      | --------- |
        FXCH            ;     r       | SQR(1-r^2) | --------- |
; FPATAN calculates arctangent ST(1)/ST
        FPATAN          ; arcsin (r)  | ---------- |
        CLD
		RET
_ARC_COSINE      ENDP

;
;*********************************|
;          arc tangent            |
;*********************************|
_ARC_TANGENT     PROC
; Find the angle (in radian) whose tangent is in ST(0)
; On entry:
;         ST(0) = tangent of an angle (in radian)
;                 Must be a normal number
; On exit:
;         ST(0) = arc tangent of entry value in ST(0)
;
; Entry values and identities:
; 1. If ST(0) is positive normal proceed as follows:
;    1a. If ST(0) = 1 then arctan (1) = pi/4
;    1b. If ST(0) > 1 then ATAN (x) = pi/2 - ATAN (1/x)
;    1c. If ST(0) < 1 then in range for FPATAN instruction
; 2. If ST(0) is negative then arctan(-x) = - arctan(x)
; 3. If ST(0) = 0 then arctan (0) = 0
;
; The BH register is used to store the sign of the original
; value. If BH = 0 original input was positive. If BH = 1
; original input was negative

        MOV     BX,0            ; Clear BX. Default sign is +
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      |   empty    |   empty
        FXAM            ; Examine stack top register
        FSTSW   STATUS_WW
        FWAIT
; Relevant condition codes:
; C3     C2    C1    C0
;  0      1     0     0   if  ST(0) = positive normal
;  0      1     1     0   if  ST(0) = negative normal
;  1      0     ?     0   if  ST(0) = zero
;
        MOV     AH,BYTE PTR STATUS_WW+1
        AND     AH,01000111B    ; Mask off all unused bits
        CMP     AH,00000110B    ; Bit code for negative
        JNE     NON_NEGATIVE    ; Argument is not negative
;**********************|
;   negative argument  |
;**********************|
; Change sign of ST(0) and store original negative sign
        MOV     BH,1            ; Code for negative input
        FCHS            ; Change sign of ST(0)
        JMP     CASE_1  ; Continue as if input was positive
NON_NEGATIVE:
        AND     AH,01000011B    ; Mask off C3 bit
        CMP     AH,01000000B    ; Test for zero
        JNE     CASE_1          ; Argument is not zero
;**********************|
;     zero argument    |
;**********************|
; The arc tangent of zero is zero
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- |
        FSTP    ST(0)   ; ----------- |
        FLDZ            ;     0.0     | ---------- |
        JMP     EXIT_ATAN
;**********************|
;   positive normal    |
;      argument        |
;**********************|
CASE_1:
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- |
        FLD1            ;     1.0     |      r     | ---------- |
        FCOM            ; Compare angle with 1 to determine if
                        ; r < 1, r > 1, or r = 1
        FSTSW   STATUS_WW
        FWAIT
; Condition codes after FCOM:
; C3     C2     C0
;  0      0      0   if  1 > r
;  0      0      1   if  1 < r
;  1      0      0   if  1 = r
;
        MOV     AH,BYTE PTR STATUS_WW+1
        AND     AH,01000011B    ; Mask off all unused bits
        JZ      R_LESS_1        ; Go if r < 1
        CMP     AH,00000001B    ; Test for r > 1
        JE      R_GREATER_1     ; Go if r > 1
;**********************|
; case 1A              |
;           ST(0) = 1  |
;**********************|
; If ST(0) = 1 the arc tangent ST(0) = pi/4
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;     1.0     |      r     | --------- |
        FADD    ST(0),ST;     2.0     |      r     | --------- |
        FADD    ST(0),ST;     4.0     |      r     | --------- |
        FLDPI           ;     pi      |     4.0    |     r     |
        FDIV    ST,ST(1);    pi/4     |     4.0    |     r     |
        FSTP    ST(1)   ;    pi/4     |      r     | --------- |
        JMP     SIGN_RTN
;**********************|
; case 1B              |
;           ST(0) > 1  |
;**********************|
R_GREATER_1:
;                       ;   ST(0)    |    ST(1)   |    ST(2)
                        ;    1.0     |      r     | --------- |
        FXCH            ;     r      |     1.0    | --------- |
        FPATAN          ; arctan(r)  |     1.0    | --------- |
        FLD1            ;     1      |  arctan(r) |    1.0    |
        FADD    ST,ST(0);     2      |  arctan(r) |    1.0    |
        FLDPI           ;    pi      |     2.0    | arctan(r) |..
        FDIV    ST,ST(1);   pi/2     |     2.0    | arctan(r) |..
        FSTP    ST(1)   ;   pi/2     | arctan(r)  |    1.0    |
        FSUBRP  ST(1),ST;  angle     |     1.0    |   empty
        JMP     SIGN_RTN
;**********************|
; case 1C              |
;           ST(0) < 1  |
;**********************|
R_LESS_1:
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;     1.0     |      r     | --------- |
        FPATAN          ;     angle   |      r     | --------- |
;**********************|
;     sign routine     |
;**********************|
; Change sign of angle if original tangent was negative
SIGN_RTN:
        CMP     BH,1            ; BH = 1 if negative input
        JNE     EXIT_ATAN     ; Go if not negative
        FCHS
EXIT_ATAN:
        CLD
		RET
_ARC_TANGENT     ENDP


;*********************************************
;       support for hyperbolic functions
;*********************************************
E_TO_Y          PROC 
; Procedure for obtaining e**y in support of hyperbolic function
; calculations
;
; On entry:
;          ST(0) = y
; On exit:
;          ST(0) = e**y
;
        FLD     E                 ;     e     |    y     | -------- |
        FXCH                      ;     y     |    e     | -------- |
        CALL    _X_TO_Y_BYLOG     ;   e**y    | -------- |
        RET
E_TO_Y          ENDP

LOG_E           PROC
; Calculate the natural log (base e)
; On entry:
;         ST(0) = positive number
; On exit:
;         ST(0) = ln ST(0) (Natural log)
;
;                       |    ST(0)    |    ST(1)   |    ST(2)
                        ;      x      | ---------- | ---------- |
        FLDLN2          ;   LOGe(2)   |     x      | ---------- |
        FXCH            ;      x      |   LOGe(2)  | ---------- |
        FYL2X           ;  LOGe(2)*   | 
                        ;   LOG2(x)   | ---------- |
        CLD
        RET
LOG_E           ENDP


_X_TO_Y_BYLOG    PROC  USES edi esi ebx ebp
; Raise a nymber (y) to an arbitrary power (x)
; On entry:
;         ST(0) = exponent
;         ST(1) = base
; On exit:
;         ST(0) = x**y
; 
; The FPU exponential function x**y is implemented by the following
; identity:
;            x**y = 2**(y*Log2(x))

; Algorithm to raise 2 to an arbitrary power::
;            2**x = (2**i)*(2**f)
;            where:
;                  i = the nearest integer smaller than x
;                  f = the fraction x - i
; In calculations:
;            if f > 1/2 then
;               2**f = 2**((f-1/2) + 1/2)
;                    = 2**(f-1/2) * 2**(1/2)
;                    = 2**(f-1/2) * (square root of 2)
; This manipulation allows reducing f to the range of the instruction
; F2XM1. 2**i is calculated using FSCALE
;
        FXCH                    ; Invert x and y
;                       |    ST(0)    |    ST(1)    |   ST(2)
                        ;      y      |      x      | --------- |
        FYL2X           ;   y*LOG2(x) | ----------- | --------- |
;**********************|
;   set to round down  |
;**********************|
        FSTCW   CONTROL_WW   ; Store control word in memory
        FLDCW   ROUND_DOWN   ; Install new control word
; At this point the 80x87 is set to round down
; This ensures that i is smaller than x
;                       |    ST(0)    |    ST(1)   |    ST(2)
                        ;      x      | ---------- | --------- |
        FLD     ST(0)   ;      x      |     x      | --------- |
        FRNDINT         ;      i      |     x      | --------- |
        FLDCW   CONTROL_WW   ; Restore original control word
        FSUB    ST(1),ST;      i      |     f      | --------- |
; f = x - i, that is, the fractional part of x
        FXCH            ;      f      |     i      | --------- |
        FLD     ONE_HALF
                        ;     1/2     |     f      |     i     |
        FXCH            ;      f      |    1/2     |     i     |
; Scale f to the range of F2XM1, which is 0 <= x => 1/2
        FPREM           ; REM(f/(1/2) |    1/2     |     i
; If the value of f => 0.5 then the quotient of FPREM is 1
; and bit C1 is set. Otherwise x is < 1/2
        FSTSW   STATUS_WW    ; Store status word
        FWAIT
;                       |      f      |     1/2    |     i    |
        FSTP    ST(1)   ;      f      |     i      | -------- |
        F2XM1           ;   2**(f)-1  |     i      | -------- |
; Since F2XM1 calculates 2^(f)-1, 1 must be added to the
; result
        FLD1            ;      1      | 2**(f)-1   |     i   |
        FADDP   ST(1),ST;    2**f     |     i      | ------- |
        MOV     AH,BYTE PTR STATUS_WW +1
                        ; Move status bits into AH
        TEST    AH,00000010B    ; Test bit C1
        JZ      CASE_2
; If f was => 1/2 then multiply by square root of 2
        FLD1            ;     1       |   2**f     |     i   |
        FADD    ST,ST(0);     2       |   2**f     |     i   |
        FSQRT           ;  2**(1/2)   |   2**f     |     i   |
        FMULP   ST(1),ST;2**f*2**(1/2)|     i      | ------- |
CASE_2:
;                       |    2**f     |      i     | ------- |
        FSCALE          ;  2(f)*2(i)  |      i     | ------- |
        FSTP    ST(1)   ;    2**x     | ---------- |
EXTI_XTOYL:
        CLD 
        RET
_X_TO_Y_BYLOG    ENDP


;*******************************************
;     calculating hyperbolic functions
;*******************************************
;
_SINH    PROC
; Calculate hyperbolic sine using formula
;        sinh(x) = (e**x - e**-x)/2
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = sinh(x)
;
;                               ;    x     | -------- |
; First obtain e**-x
        FLD     ST(0)           ;    x     |     x    | -------- |
        FCHS                    ;   -x     |     x    | -------- |
        CALL    E_TO_Y          ;  e**-x   |     x    | -------- |
        FXCH                    ;    x     |   e**-x  | -------- |
        CALL    E_TO_Y          ;   e**x   |   e**-x  | -------- |
        FSUB    ST,ST(1)        ;  (e**x - |
                                ;   e**-x) |   e**-x  | -------- |
        FSTP    ST(1)           ;(e**x ..) | -------- | -------- |
        FLD1                    ;    1     |(e**x ..) | -------- |
        FADD    ST,ST(0)        ;    2     |(e**x ..) | -------- |
        FDIV    ST(1),ST        ;    2     | sinh (x) | -------- |
        FSTP    ST(0)           ; sinh(x)  | -------- |
        CLD
		RET
_SINH    ENDP
;
_COSH    PROC
; Calculate hyperbolic cosine using formula
;        cosh(x) = (e**x + e**-x)/2
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = cosh(x)
;
;                               ;    x     | -------- |
; First obtain e**-x
        FLD     ST(0)           ;    x     |     x    | -------- |
        FCHS                    ;   -x     |     x    | -------- |
        CALL    E_TO_Y          ;  e**-x   |     x    | -------- |
        FXCH                    ;    x     |   e**-x  | -------- |
        CALL    E_TO_Y          ;   e**x   |   e**-x  | -------- |
        FADD    ST,ST(1)        ;  (e**x - |
                                ;   e**-x) |   e**-x  | -------- |
        FSTP    ST(1)           ;(e**x ..) | -------- | -------- |
        FLD1                    ;    1     |(e**x ..) | -------- |
        FADD    ST,ST(0)        ;    2     |(e**x ..) | -------- |
        FDIV    ST(1),ST        ;    2     | sinh (x) | -------- |
        FSTP    ST(0)           ; sinh(x)  | -------- |
        CLD
		RET
_COSH    ENDP
;
_TANH    PROC
; Calculate hyperbolic tangent using formula
;        tanh(x) = sinh(x) / cosh(x)
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = tanh(x)
;
;                               ;    x     | -------- |
        FLD     ST(0)           ;    x     |     x    | -------- |
        CALL    _SINH           ;  sinh(x) |     x    | -------- |
        FXCH                    ;    x     |  sinh(x) | -------- |
        CALL    _COSH           ;  cosh(x) |  sinh(x) | -------- |
        FDIV    ST(1),ST        ;  cosh(x) |  tanh(x) | -------- |
        FSTP    ST(0)           ;  tanh(x) | -------- |
        CLD
		RET
_TANH    ENDP

;*************************************
;        hyperbolic arcfunctions
;*************************************
_ARC_SINH        PROC
; Calculate hyperbolic arcsine using the formula:
;
;       arcsinh(x) = log(e)[x + SQR(x**2 + 1)]
;
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = arcsin(x)
; Calculate x**2
        FLD     ST(0)           ;    x     |     x    | -------- |
        FMUL    ST(0),ST        ;   x**2   |     x    | -------- |
        FLD1                    ;    1     |   x**2   |    x     |..
        FADDP   ST(1),ST        ;(x**2 + 1)|    x     | -------- |
        FSQRT                   ;SQR(x2+1) |    x     |
        FADDP   ST(1),ST        ; x + SQR( |
                                ; x**2 + 1)| -------- |
        CALL    LOG_E           ;arcsinh(x)| -------- |
        CLD
        RET
_ARC_SINH        ENDP
;

_ARC_COSH        PROC
; Calculate hyperbolic arcosine using the formula:
;
;       arcosh(x) = log(e)[x + SQR(x**2 - 1)]
;
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = arcosh(x)
; Calculate x**2
        FLD     ST(0)           ;    x     |     x    | -------- |
        FMUL    ST(0),ST        ;   x**2   |     x    | -------- |
        FLD1                    ;    1     |   x**2   |    x     |..
        FSUBP   ST(1),ST        ;(x**2 + 1)|    x     | -------- |
        FSQRT                   ;SQR(x2+1) |    x     |
        FADDP   ST(1),ST        ; x + SQR( |
                                ; x**2 + 1)| -------- |
        CALL    LOG_E           ;arcsinh(x)| -------- |
        CLD
        RET
_ARC_COSH        ENDP
;
_ARC_TANH        PROC
; Calculate hyperbolic arctangent using the formula:
;
;       arctanh(x) = log(e) SQR [(1 + x)/(1 - x)]
;
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = arctanh(x)
;
;***************************|
;      test for negative    |
;        argument           |
;***************************|
        MOV     SIGN,0          ; Assume positive
        FTST                    ; Test ST(0)
        FSTSW   STATUS_WW       ; Store status word
        FWAIT
        MOV     AH,BYTE PTR STATUS_WW + 1
        TEST    AH,00000001B    ; Is bit C0 set?
        JZ      POSITIVE_VALUE  ; Go if positive
; At this point input angle is negative
        MOV     SIGN,1          ; Store sign of input angle
        FABS                    ; Change sign
;***************************|
;  calculate arctanh(x)     |
;***************************|
; Calculate 1 + x 
POSITIVE_VALUE:
        FLD     ST(0)           ;    x     |     x    | -------- |
        FLD1                    ;    1     |     x    |     x    |
        FADDP   ST(1),ST        ;  1 + x   |     x    | -------- |
; Calculate 1 - x
        FXCH                    ;    x     |   1 + x  | -------- |
        FLD1                    ;    1     |     x    |   1 + x  |
        FXCH                    ;    x     |     1    |   1 + x  |
        FSUBP   ST(1),ST        ;  1 - x   |   1 + x  | -------- |
        FDIVRP  ST(1),ST        ; 1+x/1-x  | -------- |
        FSQRT                   ;SQR(1+x/  |
                                ;  1-x)    | -------- |
        CALL    LOG_E           ;arctanh(x)| -------- |
        FABS                    ; Change to positive
;***************************|
;      determine sign       |
;      of arcfunction       |
;***************************|
; Test for sign change in input angle
        CMP     SIGN,0
        JE      POS_ATANH
        FCHS                    ; Change sign of tangent
POS_ATANH:
        CLD
		RET
_ARC_TANH        ENDP
;
        end
