;*****************************************************************
;*****************************************************************
;                           UN32_8.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 25, 2001
; Last update:    
;
; Libary name: MATH32.LIB
; This module: UN32_8.ASM
; Operations : Finnancial calculations
;*****************************************************************
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
; Amortization and annuities
PUBLIC  _PAYMENTS        ; .. periodic payments on a loan
PUBLIC  _ANNUITY_PV      ; .. present value of an annuity
PUBLIC  _ANNUITY_FV      ; .. future value of an investment
PUBLIC  _ANNUITY_DUE     ; .. Annuity due at start of period
PUBLIC  _COMP_PERIODS    ; .. number of compounding periods
PUBLIC  _SINKING_FUND    ; .. periodic payment on annuity
;
; Interest calculations
PUBLIC  _SIMPLE_INTEREST ; .. simple interest
PUBLIC  _FUTURE_VALUE    ; .. future value at compound interest
PUBLIC  _PRESENT_VALUE   ; .. present value at compound interest
PUBLIC  _EFFECTIVE_YIELD ; .. effective annual yield
PUBLIC  _APR             ; .. Annual Percentage Rate
PUBLIC  _AOI_PAYMENT     ; .. Add-on interest payment
;
; 
PUBLIC  _SL_DEPRECIATION ; .. straight-line depreciation
;****************************************************************
;               definitions for 32-bit flat model
;****************************************************************
        .486
        .MODEL flat
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATH32 library
EXTRN   _LOG_E:        PROC   ;-----  Un32_5 module  
EXTRN   _X_TO_Y_MIXED: PROC   ;-----  Un32_5 module

        .DATA
;***************************|
;   Data for this module    |
;***************************|
; Data for financial procedures
PRINCIPAL       DT      0       ; Storage for principal
INTEREST        DT      0       ; Storage for interest rate
TERM            DT      0       ; Storage for term of loan
PAYMENT         DT      0       ; Storage for periodic payment
PRESENT_VAL     DT      0       ; Storage for present value
FUTURE_VAL      DT      0       ; Storage for future value
ANNUITY         DT      0       ; Storage for annuity
;****************************************************************
;                           code
;****************************************************************
        .CODE  
;******************************************
;            financial procedures
;******************************************

;*****************************************|
;               interest                  |
;*****************************************|
_SIMPLE_INTEREST    PROC
; Simple interest calculation according to the formula:
;
;          I = Prt
; where:
;          P = principal amount (present value)
;          r = Annual interest rate
;          t = Time in years
;          I = Amount of interest
; On entry:
;          ST(0) = r (interest rate)
;          ST(1) = P (principal)
;          ST(2) = t (time in years)
; On exit:
;          ST(0) = I (interest amount)
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   r   |   P   |   t   | ----- |
        FMULP      ST(1),ST     ;  rP   |   t   | ----- |
        FMULP      ST(1),ST     ;   I   | ----- |
        CLD
        RET
_SIMPLE_INTEREST   ENDP
;

_FUTURE_VALUE     PROC
; Financial procedure to calculate the future value of an investment
; at compound interest, according to the formula:
;
;         A = P(i + 1)^n
;
; where
;         P = present value
;         i = interest rate per compounding period
;         n = number of compunding periods
;         A = Value of investment at end of period
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = P (present value)
; On exit:
;          ST(0) = A (Future value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time, for example: to calculate monthly compounding periods
;       on a 3-year loan, at 12% yearly interest rate
;                   i = 0.12/12 = 0.01
;                   n = 3*12 = 36;     
;       Preliminary code must perform these adjustments
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   P   | ----- |
        FSTP    TERM            ;   i   |   P   | ----- |
        FSTP    INTEREST        ;   P   | ----- |
        FSTP    PRESENT_VAL     ; ----- |
;***************************|
;   calculate (i + 1)**n    |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^n| ----- |
;***************************|
;  calculate P*(i + 1)^n    |
;***************************|
        FLD     PRESENT_VAL     ;  P    |(i+1)^n| ----- |
        FMULP   ST(1),ST        ;  P    | ----- |
        CLD
        RET		 
_FUTURE_VALUE   ENDP
;
;         
_PRESENT_VALUE     PROC
; Financial procedure to calculate the present value of an investment
; at compound interest, according to the formula:
;
;         P = A(i + 1)^-n
;
; where
;         P = present value
;         i = interest rate per compounding period
;         n = number of compunding periods
;         A = Value of investment at end of period
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = A (future value)
; On exit:
;          ST(0) = A (Present value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time, for example: to calculate monthly compounding periods
;       on a 3-year loan, at 12% yearly interest rate
;                   i = 0.12/12 = 0.01
;                   n = 3*12 = 36;     
;       Preliminary code must perform these adjustments
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   A   | ----- |
        FSTP    TERM            ;   i   |   A   | ----- |
        FSTP    INTEREST        ;   A   | ----- |
        FSTP    FUTURE_VAL      ; ----- |
;***************************|
;   calculate (i + 1)**n    |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^n| ----- |
; x^-n = 1/x^n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;(i+1)^-n|
;***************************|
;  calculate P*(i + 1)^n    |
;***************************|
        FLD     FUTURE_VAL      ;  A    |(i+1)^n| ----- |
        FMULP   ST(1),ST        ;  P    | ----- |
        CLD
        RET		 
_PRESENT_VALUE  ENDP

_EFFECTIVE_YIELD     PROC
; Finnacial procedure to calculate the effective annual yield
; an investment at compound interest, according to the formula:
;
;         Y = ((1 + i/n)^n) - 1
;
; where
;         i = interest rate per compounding period
;         n = number of compounding periods
;         Y = effective annual yield of investment
; On entry:
;         ST(0) = i (interest rate)
;         ST(1) = n (term)
; On exit:
;         ST(0) = Y (effective annual yield of investment)
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   i   |   n   | ----- | ----- |
;***************************|
; calculate (1+i/n)^n       |
;***************************|
        FDIV    ST,ST(1)        ;  i/n  |   n   | ----- |
        FLD1                    ;   1   |  i/n  |   n   | ----- |
        FADD                    ; 1+i/n |   n   | ----- | ----- |
        FXCH                    ;   n   | 1+i/n | ----- |
        CALL    _X_TO_Y_MIXED   ;(1+i/n)
		                        ; ^n    | ----- |
;***************************|
; calculate ((1+i/n)^n) - 1 |
;***************************|
; (1+i/n)^n = w
        FLD1                    ;   1   |   w   | ----- |
		FXCH                    ;   w   |   1   | ----- |
        FSUB    ST,ST(1)        ;   Y   |   1   | ----- |
		FSTP    ST(1)           ;   Y   |
        CLD
        RET		 
_EFFECTIVE_YIELD  ENDP
;
;
_APR              PROC
; Finnacial procedure to calculate the annual percentage rate
; APR, according to the formula:
;
;              2rN
;      APR = -------
;             N + 1
;
; where
;         N = number of payments
;         r = add-on interest rate
;       APR = annual percentage rate
; On entry:
;         ST(0) = r (add-on interest)
;         ST(1) = N (number of payments)
; On exit:
;         ST(0) = APR (annual percentage rate)
;
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   r   |   N   | ----- | ----- |
;***************************|
;     calculate 2Nr         |
;***************************|
        FMUL    ST,ST(1)        ;   rN  |   N   | ----- |
		FADD    ST,ST(0)        ;  2rN  |   N   | ----- |
;***************************|
;  calculate 2rN/N + 1      |
;***************************|
        FLD1                    ;   1   |  2rN  |   N   | ----- |
		FADD    ST,ST(2)		; 1 + N |  2rN  |   N   | ----- |
		FSTP    ST(2)           ; 1 + N |  2rN  | ----- |
		FDIV    ST,ST(1)        ;  APR  |  2rN  |
		FSTP    ST(1)           ;  APR  | ----- |
        CLD
        RET		 
_APR    ENDP


_AOI_PAYMENT    PROC
; Finnacial procedure to calculate the individual payments
; on a add-on interest loan, according to the formula:
;
;              Prt
;      AOR =  -------
;               N
;
; where
;         N = number of payments
;         r = add-on interest rate
;         t = loan time in years
;         P = principal
;       AOR = amount of each add-on intrest payment
; On entry:
;         ST(0) = r (add-on interest)
;         ST(1) = P (principal)
;         ST(2) = t (number of years)
;         ST(3) = N (number of payments)
; On exit:
;         ST(0) = AOR (AOR payment)
;
;
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   P   |   r   |   t   |   N   |
;***************************|
;     calculate Prt         |
;***************************|
        FSTP     PRINCIPAL      ; Save principal
		FLD      PRINCIPAL      ; Reload  
        CALL    _SIMPLE_INTEREST
                                ;  Prt  |   N   | ----- |
        FLD     PRINCIPAL       ;   P   |  Prt  |   N   | ----- |  
        FADD                    ; P+Prt |   N   | ----- |
        FDIV    ST,ST(1)        ; Prt/N |   N   | ----- |
		FSTP    ST(1)           ;  AOR  | ----- |
        CLD
        RET		 
_AOI_PAYMENT    ENDP


;****************************************|
;     annuities and ammortization        |
;****************************************|
;
_ANNUITY_PV     PROC
; Financial procedure to calculate the present value of an
; annuity, using the formula:
;
;                 1-(i + 1)^-n
;      PV =  pm * ---------------
;                       i
;
; where:
;     PV = present value
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = PV (present value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    TERM            ;   i   |   pm  | ----- |
        FSTP    INTEREST        ;   pm  | ----- |
        FSTP    PAYMENT         ; ----- |
;***************************|
; calculate 1-(i + 1)**-n   |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^n| ----- |
; x^-n = 1/x^n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;  1/   |
                                ;(i+1)^n| ----- |
        FLD1                    ;  1    |1/(i+..| ----- |
        FSUBR                   ;1-(1/..| ----- |
;***************************|
; calculate:                |
;           1-(i + 1)^-n    |
;   pm *  ---------------   |
;                i          |
;***************************|
; num = 1-(i + 1)**-n
;                               ; num   | ----- |
        FLD     INTEREST        ;   i   |  num  | ----- |
        FDIV                    ; num/i | ----- |
        FLD     PAYMENT         ;  pm   | i/num | ----- |
        FMULP   ST(1),ST        ;  PV   | ----- |
        CLD
        RET
_ANNUITY_PV     ENDP
;
;
_ANNUITY_FV     PROC
; Financial procedure to calculate the future value of an
; ordinary annuity, using the formula:
;
;                 (i + 1)^n - 1
;      FV =  pm * --------------
;                       i
;
; where:
;     FV = future value of investment
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = FV (future value of investment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    TERM            ;   i   |   pm  | ----- |
        FSTP    INTEREST        ;   pm  | ----- |
        FSTP    PAYMENT         ; ----- |
;***************************|
; calculate (i + 1)^n - 1   |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^n| ----- |
        FLD1                    ;   1   |(i+1)^n| ----- |
        FSUB                    ;(i+1)^n|
                                ;  -1   | ----- |
;***************************|
; calculate:                |
;          (i + 1)^(n - 1)  |
;   pm *  ---------------   |
;                i          |
;***************************|
; num = (i + 1)^(n - 1)
;                               ; num   | ----- |
        FLD     INTEREST        ;   i   |  num  | ----- |
        FDIV                    ; num/i | ----- |
        FLD     PAYMENT         ;  pm   | i/num | ----- |
        FMULP   ST(1),ST        ;  PV   | ----- |
        CLD
        RET
_ANNUITY_FV     ENDP
;
;
_ANNUITY_DUE    PROC
; Financial procedure to calculate the future value of an
; annuity due, using the formula:
;
;                 (i + 1)^(n+1) - 1
;      A =  pm * ------------------ - 1
;                         i
;
; where:
;     A = annuity due at the beginning of each period
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = pm (periodic payment)
; On exit:
;          ST(0) = A (annuity due)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   pm  | ----- |
        FSTP    TERM            ;   i   |   pm  | ----- |
        FSTP    INTEREST        ;   pm  | ----- |
        FSTP    PAYMENT         ; ----- |
;***************************|
; calculate (i + 1)^(n + 1) |
;***************************|
        FLD     TERM            ;   n   | ----- |
	    FLD1                    ;   1   |	n   | ----- |
        FADD                    ;  n+1  | ----- |
        FLD     INTEREST        ;   i   |  n+1  | ----- |
        FLD1                    ;   1   |   i   |  n+1  | ----- |
        FADD                    ;  i+1  |  n+1  | ----- |
        FXCH                    ;  n+1  |  i+1  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^
		                        ; (n+1) | ----- |
; (i+1)^(n+1) = f1
;***************************|
;   calculate f1 - 1        |
;***************************|
        FLD1                    ;   1   |   f1  | ----- |
        FSUBP   ST(1),ST        ; 1-f1  | ----- |
;***************************|
; calculate:                |
;            f1             |
;   pm *  ------- - 1       |
;            i              |
;***************************|
;                               ;   f1  | ----- |
        FLD     INTEREST        ;   i   |   f1  | ----- |
        FDIVP   ST(1),ST        ;  f1/i | ----- |
        FLD1                    ;   1   | f1/i  | ----- |
        FSUBP   ST(1),ST        ;1-f1/i | ----- |
        FLD     PAYMENT         ;  pm   |1-f1/i | ----- |
        FMULP   ST(1),ST        ;   A   | ----- |
        CLD
        RET
_ANNUITY_DUE    ENDP
;
;
_SINKING_FUND   PROC
; Financial procedure to calculate the periodic payment
; (sinking fund) required in an annuity, using the formula:
;
;                Ai   
;     pm =  --------------
;           (i + 1)^n - 1
;
; where:
;     pm = periodic payment
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;     A = annuity
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = A (annuity due)
; On exit:
;          ST(0) = pm (periodic payment required)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time (see note in the PAYMENTS procedure)
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   A   | ----- |
        FSTP    TERM            ;   i   |   A   | ----- |
        FSTP    INTEREST        ;   A   | ----- |
        FSTP    ANNUITY         ; ----- |
;***************************|
;   calcuate Ai             |
;***************************|
        FLD     INTEREST        ;   i   | ----- |
		FLD     ANNUITY         ;   A   |   i   | ----- |
		FMUL    ST,ST(1)        ;  Ai   |   i   | ----- |
	    FXCH                    ;   i   |  Ai   | ----- |
;***************************|
; calculate (i + 1)^n - 1   |
;***************************|
        FLD     TERM            ;   n   |   i   |   Ai  | ----- |
		FXCH                    ;   i   |   n   |   Ai  |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   |   Ai  | ----- |
        FXCH                    ;   n   |  i+1  |   Ai  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^n|  Ai   | ----- |
        FLD1                    ;   1   |(i+1)^n|   Ai  | ----- |
        FSUB                    ;(i+1)^n|
                                ;  -1   |  Ai   | ----- |
;***************************|
; calculate:                |
;                Ai         |
;   pm = ----------------   |
;          (i + 1)^n - 1    |
;                           |
;***************************|
; (i + 1)^n - 1 = f1
;                               ;   f1  |   Ai  | ----- |
        FDIVP   ST(1),ST        ; Ai/f1 | ----- |
        CLD
        RET
_SINKING_FUND   ENDP
;
;
_PAYMENTS        PROC
; Financial procedure to calculate the ammount of each payment on
; a loan using the formula:
;
;                     i
;      P =  p * ---------------
;               1-(i + 1)^-n
; where
;     P = ammount of periodic payment
;     p = principal
;     i = periodic interest rate (in fractional form)
;     n = term of loan (number of payment periods)
;
; On entry:
;          ST(0) = n (term)
;          ST(1) = i (interest rate)
;          ST(2) = p (principal)
; On exit:
;          ST(0) = P (ammount of periodic payment)
;
; Note: interest rate (i) and term (n) must be in the same units
;       of time, for example: to calculate monthly payments on a
;       3-year loan of $4000.00, at 0.14% yearly interest rate
;                   p = 4000
;                   i = 0.14/12 = 0.0116666...
;                   n = 3*12 = 36
;
;***************************|
;   store input data        |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   i   |   p   | ----- |
        FSTP    TERM            ;   i   |   p   | ----- |
        FSTP    INTEREST        ;   p   | ----- |
        FSTP    PRINCIPAL       ; ----- |
;***************************|
;  calculate 1-(i + 1)^-n   |
;***************************|
        FLD     TERM            ;   n   | ----- |
        FLD     INTEREST        ;   i   |   n   | ----- |
        FLD1                    ;   1   |   i   |   n   | ----- |
        FADD                    ; i + 1 |   n   | ----- |
        FXCH                    ;   n   |  i+1  | ----- |
        CALL    _X_TO_Y_MIXED   ;(i+1)^n| ----- |
; x^-n = 1/x^n
        FLD1                    ;   1   |(i+1)^n| ----- |
        FDIVR                   ;  1/   |
                                ;(i+1)^n| ----- |
        FLD1                    ;  1    |1/(i+..| ----- |
        FSUBR                   ;1-(1/..| ----- |
;***************************|
; calculate:                |
;     p * i/1-(i + 1)^-n    |
;***************************|
; den = 1-(i + 1)**-n
;                               ; den   | ----- |
        FLD     INTEREST        ;   i   |  den  | ----- |
        FDIVR                   ; i/den | ----- |
        FLD     PRINCIPAL       ;  p    | i/den | ----- |
        FMULP   ST(1),ST        ;  P    | ----- |
        CLD
        RET
_PAYMENTS        ENDP
;
;
_COMP_PERIODS    PROC
; Financial procedure to calculate the number of compounding
; periods necessary for an investment to grow to a predetermined
; value, by means of the formula:
;
;           ln(fv/sv)
;      CP = ---------
;           ln(1 + i)
;
; where:
;      CP = number of compounding periods
;      fv = final value
;      sv = starting value
;       i = interest rate
;      ln = log base e (natural logarithm)
; On entry:
;          ST(0) = i (interest rate)
;          ST(1) = fv (final value of investment)
;          ST(2) = sv (starting value of investment)
; On exit:
;          ST(0) = CP (number of compounding periods for growth)
;
; Note: the number of compounding periods (CP) will be in the same
;       units of time as the interest rate (i). For example: to
;       calculate compounding periods (in months) for an investment
;       at a yearly interest of 14 percent
;                 i = 0.14/12 = 0.0116666...
;
;***************************|
;   calculate ln(1 + i)     |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   i   |   fv  |   sv  | ----- |
        FLD1                    ;   1   |   i   |   fv  |   sv  |
        FADD                    ; 1 + i |   fv  |   sv  | ----- |
        CALL    _LOG_E          ;ln(1+i)|   fv  |   sv  | ----- |
;***************************|
;   calculate:
;      ln(fv/sv)/ln(1 + i)  |
;***************************|
        FXCH    ST(2)           ;  sv   |   fv  |ln(1+i)| ----- |
        FDIV                    ; fv/sv |ln(1+i)| ----- |
        CALL    _LOG_E          ;ln(fv/ |
                                ;  sv)  |ln(1+i)| ----- |
        FDIVR                   ;  CP   | ----- |
        CLD 
        RET
_COMP_PERIODS    ENDP
;
;
;**************************|
;      accounting          |
;**************************|
_SL_DEPRECIATION PROC
; Financial procedure to calculate the straight-line depreciation
; of an asset during 1 period, using the formula:
;
;              (c - s)
;      SLD =  --------
;                 ul
; where:
;     SLD = straight-line depreciation allowance
;       c = initial cost of the asset
;       s = salvage value of the asset at the end of the
;           depreciation period
;      ul = useful life of the asset
;
; On entry:
;          ST(0) = ul (useful life)
;          ST(1) = s (salvage value)
;          ST(2) = c (initial cost)
; On exit:
;          ST(0) = SLD allowance per period
;
; Note: the depreciation period (SLD) and the useful life (ul) of
;       asset must be in the same units of time
;
;***************************|
;   calculate (c-s)/ul      |
;***************************|
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   ul  |   s   |   c   | ----- |
        FXCH    ST(2)           ;   c   |   s   |   ul  | ----- |
        FSUB    ST,ST(1)        ;  c-s  |   s   |   ul  | ----- |
        FSTP    ST(1)           ;  c-s  |   ul  | ----- |
        FDIV    ST,ST(1)        ; c-s/ul|   ul  | ----- |
        FSTP    ST(1)           ;  SLD  | ----- |
        CLD
        RET
_SL_DEPRECIATION ENDP
;
;


        end


    